/* FILE: simp.c              Copyright (c), Symmetric Research, 2003-2008
 *
 * Acquire PARxCH and PARGPS data with a simple text only display and disk
 * save.
 *
 * For an overview of the program flow, see the main function located
 * at the end of this file after all the helper functions.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "parxch.h"
#include "srhelper.h"   // for SrFileExist, SrGetKeyWait, SrStringCopy, etc
#include "srdat.h"
#include "simp.h"       // Needs PARXCH_TOTAL_CHANNELS define from parxch.h



/* GPS IFDEFS:

This code can be compiled with or without GPS support.  When the
symbol GPS_AVAILABLE is defined, GPS support is included and the PARGPS
software must be installed in order to compile.

*/

#ifdef GPS_AVAILABLE  // GPS support included, PARGPS software required

#include "pargps.h"
#define HDR_PRODUCT_STRING   "SR PARxCH with GPS"


#else // NO GPS support, provide some dummy defines instead

#define HDR_PRODUCT_STRING   "SR PARxCH"
#define PARGPS_0             "None"
#define GPSMODEL_NONE         0
#define SERIAL_PORT_1         1
typedef int PPSTDATA;
typedef int SERIALDATA;

#endif



// Prototypes needed to satisfy forward references

void CleanExit( int value );





// Global variables

DEVHANDLE ParXchHandle;
DEVHANDLE GpsHandle;
struct INI_data INI;     // <<< see simp.h for data layout ...

#define DIM_STR 256
char msg[DIM_STR];
char FullFileName[DIM_STR];

int GpsMarkIndex;
int DigitalIndex;
int CounterIndex;

unsigned long NpointsFromStart;



/* STRING HELPERS:
 *
 *
 *
 */

static char ChopBuffer[256];

char *Chop( char *s ) {

        char *p;


        // Chop spaces ' ' off the end of a string and add '\n'.

        strncpy( ChopBuffer, s, 255 );

        p = ChopBuffer;

        while ( *p )

                p++;


        if ( *--p == '\n' ) {

                while ( *--p == ' ' )

                         ; // back up


                *++p = '\n';
                *++p = '\0';

                }


        return( ChopBuffer );

}
void DisplayInfo( char *info ) {

        if (INI.DisplayInfo == ON)
                printf( info );
}
void DisplayInfoFlush( char *info ) {

        if (INI.DisplayInfo == ON) {
                printf( info );
                SrFlushPrint();
                }

}





/* REV MESSAGE:
 *
 * 06/05/08 Add GPS counter frequency check message
 * 12/05/07 Add Garmin GPS 18 LVC support
 * 08/20/07 Add IRQ number check when using PARGPS
 * 01/15/07 Make sequential file naming work under 16-bit DOS
 * 01/03/07 Allow restart after sanity error and SrParXch0 driver name
 * 03/20/06 Removed fileno call
 * 02/20/06 Added digital edge triggering, sequential decimal file naming,
 *          and SRDAT_REV 106 changes
 * 06/20/05 Added ChopBuffer so Linux won't complain about fwritable-strings
 * 03/25/03 Previous rev
 *
 */

void Rev( void ) {

        // Print the rev info.

        sprintf( msg, "PARxCH simp [/?]: (c) SR 2008 : Rev {06/05/08}\n" );
        DisplayInfo( msg );
}






/* ERRORS:
 *
 * All errors except for .ini parsing errors go through these functions.
 *
 *
 */

char *UsageMsg[] = {
        "USAGE: simp file.ini [/?]                         ",
        "                                                  ",
        "       file.ini     is a required .ini file       ",
        "       /?           show this usage message       ",
        "                                                  ",
        "EXAMPLE: simp simp4.ini                           ",
        NULL,
        };

void Usage( char *errmsg ) {

        int i;

        // Print rev message

        INI.DisplayInfo = ON;
        Rev();
        

        // Print the command line usage message and quit.

        // Select first string depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( errmsg == NULL )          
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( errmsg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", errmsg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", Chop( UsageMsg[i] ) );

        exit( 1 );

}

// Runtime error handling ...

#define ERROR_NOT_FATAL         0
#define ERROR_FATAL             1

#define MSGLEVEL_MSG            0
#define MSGLEVEL_WARN           1
#define MSGLEVEL_ERR            2


#define ErrorFatal( msg )       ErrorRuntime( msg, MSGLEVEL_ERR,  ERROR_FATAL )
#define ErrorWarning( msg )     ErrorRuntime( msg, MSGLEVEL_WARN, INI.ExitOnError )
#define LogMessage( msg )       ErrorRuntime( msg, MSGLEVEL_MSG,  ERROR_NOT_FATAL )


void ErrorRuntime( char *msg, int msglevel, int fatal ) {

        FILE *ErrFp;

            char *datestr, dbuffer[16];
            char *timestr, tbuffer[16];
            char *levelstr;

            time_t     current_time;
            struct tm *local_time;



        // Get the date and time.

        time( &current_time );
        local_time = localtime( &current_time );

        strftime( dbuffer, 16, "%m/%d/%y", local_time );
        strftime( tbuffer, 16, "%H:%M:%S", local_time );

        datestr = dbuffer;
        timestr = tbuffer;


        // Set message level string.
        
        if (msglevel == MSGLEVEL_MSG)
                levelstr = "Message";
        else if (msglevel == MSGLEVEL_WARN)        
                levelstr = "Warning";
        else if (msglevel == MSGLEVEL_ERR)        
                levelstr = "Error";
        else        
                levelstr = "Note";
                   

        // Append the error to simp.err.

        sprintf( FullFileName, "%sSIMP.ERR", INI.OutputDirectory );

        if ( ErrFp = fopen( FullFileName, "a" ) ) {

                fprintf( ErrFp, "%s %s %-7s = %s\n", datestr, timestr, levelstr, msg );

                   fclose( ErrFp );

                }



        // Display the error.

        printf( "Simp %-7s: %s\n", levelstr, msg );



        // Quit if a fatal error.

        if ( fatal ) CleanExit( 1 );

}



/* INI FILE PARSING AND OUTPUT HELPER FUNCTIONS:
 *
 *
 */

int ProcessCommandLine( int argc, char **argv ) {

        int i, IniArg;


        // Check the number of command line arguments.

        if ( argc < 2 || argc > 3 )

                Usage( "wrong number of command line arguments" );

        // Process args

        IniArg = 0;
        for ( i = 1; i < argc ; i++ ) {
                if ( strcmp( argv[i], "/?" ) == 0 )         // help requested
                        Usage( "?" );
                else if ( IniArg == 0 )                     // ini name found
                        IniArg = i;
                else                                        // dup ini name
                        Usage( "Too many .ini filenames" );
                }
        if ( IniArg == 0 )
                Usage( "No ini filename given" );

        return( IniArg );

}








/* INI FILE PARSING AND OUTPUT HELPER FUNCTIONS:
 *
 * These helper functions are used in parsing the .ini input file and
 * displaying the .ini values.  The parsing functions are more
 * complicated than simple fscanf to allow for free format syntax and
 * better error reporting.
 *
 * The INI structure is declared as a global variable at the top of the
 * program and the major INI functions are below.
 *
 *
 */

FILE *IniFp;

int fline = 1;


#define MAXTOKEN        256             // token info

char TOKEN[MAXTOKEN], *token;

#define SYMBOL          1               // token types
#define EQUALS          2
#define HEX             3
#define INTEGER         4
#define FLOAT           5
#define STRING          6
#define UNKNOWN         7

int ttype;




void ErrorIni( char *msg ) {


        // Display an .ini error and quit.

        printf( "ERROR[%s] : line %d : %s\n", INI.Filename, fline, msg );

        exit( 1 );


}


// These are general parsing helpers.

void AddToToken( int c ) {


        // Add c to TOKEN, checking for token buffer overflow.

        if ( (token - TOKEN) < MAXTOKEN )

                *token++ = c;

        else
                ErrorIni( "internal token length exceeded" );


}
int SkipSpace( int c ) {


        // Skip over white space, including comments.
skip:

        if ( isspace( c ) ) {

                if ( c == '\n' )

                        fline++;


                c = fgetc( IniFp );

                goto skip;

                }


        else if ( c == ';' ) { // ; comment ...

                while ( c != '\n' && c != EOF )

                        c = fgetc( IniFp );

                goto skip;

                }


        return( c );


}
int GetToken( void ) {

        static int c = ' ';



        // Skip over any white space.

        c = SkipSpace( c );



        // Gather the next token.

        token = TOKEN;


        if ( isalpha( c ) || c == '_' || c == '/' ) {

                ttype = SYMBOL;

                while ( isalnum( c ) || c == '_' || c == '/' )

                        { AddToToken( c ); c = fgetc( IniFp ); }
                }


        else if ( c == '=' ) {

                ttype = EQUALS;

                  AddToToken( c ); c = fgetc( IniFp );

                }


        else if ( isdigit( c ) ) {

                // ttype could be HEX, INTEGER, or FLOAT ...

                AddToToken( c );

                  c = fgetc( IniFp );



                if ( TOKEN[0] == '0' && c == 'x' ) {

                        ttype = HEX;
                        token = TOKEN;

                        c = fgetc( IniFp );

                        if ( isxdigit( c ) ) {

                                while ( isxdigit( c ) )

                                        { AddToToken( c ); c = fgetc( IniFp ); }
                                }

                        else

                                ttype = UNKNOWN;
                        }



                else {

                        ttype = INTEGER;

                        while ( isdigit( c ) )

                                { AddToToken( c ); c = fgetc( IniFp ); }



                        if ( c == '.' ) {

                                ttype = FLOAT;

                                AddToToken( c );

                                  c = fgetc( IniFp );

                                while ( isdigit( c ) )

                                        { AddToToken( c ); c = fgetc( IniFp ); }
                                }
                        }
                }



        else if ( c == '"' ) {

                ttype = STRING;

                c = fgetc( IniFp );

                  while ( c != '"' && c != '\n' && c != EOF )

                        { AddToToken( c ); c = fgetc( IniFp ); }



                if ( c == '"' )

                        c = fgetc( IniFp );

                else
                        ErrorIni( "missing quote on string" );

                }



        else if ( c == EOF )

                ttype = EOF;


        else {

                ttype = UNKNOWN;

                  AddToToken( c );


                c = fgetc( IniFp );

                }




        // Null terminate the TOKEN and return its type.

        AddToToken( '\0' );

        return( ttype );




}
int Match( char *s ) {

        char *t;


        // Check for a match between s and TOKEN.

        t = TOKEN;

        while ( *s && *t && toupper( *s ) == toupper( *t ) )

                { s++; t++; }


        return ( *s == '\0' && *t == '\0' ? 1 : 0 );



}
int KeywordEquals( char *keyword ) {

#define KEYWORD_EQUALS_ERROR   "expected = sign after %s keyword"

        if ( Match( keyword ) ) {


                if ( GetToken() == EQUALS )

                        return( 1 );

                else {
                        sprintf( msg, KEYWORD_EQUALS_ERROR, keyword );

                          ErrorIni( msg );

                        }
                }



        return( 0 );


}
void GetOnOff( int *parameter, char *error ) {


        // Get an {ON,OFF} parameter value.

        GetToken();

        if ( Match( "ON" ) )

                *parameter = ON;

        else if ( Match( "OFF" ) )

                *parameter = OFF;

        else
                ErrorIni( error );


}
void GetInt( int *parameter, char *error ) {


        // Get a decimal int parameter value.

        GetToken();

        if ( ttype != INTEGER )

                ErrorIni( error );


        sscanf( TOKEN, "%d", parameter );


}
void GetUint( unsigned int *parameter, char *error ) {


        // Get a decimal int parameter value.

        GetToken();

        if ( ttype != INTEGER )

                ErrorIni( error );


        sscanf( TOKEN, "%u", parameter );


}
void GetUlong( unsigned long *parameter, char *error ) {

        // Get an unsigned long parameter value.

        GetToken();

        if ( ttype != INTEGER )

                ErrorIni( error );


        sscanf( TOKEN, "%lu", parameter );


}
void GetHex( int *parameter, char *error ) {

        // Get a hex parameter value.

        GetToken();

        if ( ttype != HEX )

                ErrorIni( error );


        sscanf( TOKEN, "%x", parameter );



}
void GetFloat( double *parameter, char *error ) {


        // Get a floating point parameter value.

        GetToken();

        if ( (ttype != FLOAT) && (ttype != INTEGER) )

                ErrorIni( error );


        sscanf( TOKEN, "%lf", parameter );


}


// These are INI specific parsing helpers.

int ConvertGain( int value ) {

        int lnvalue, gtable[] = { 1, 2, 4, 8, 16 };



        // Compute the log of the gain value.

        for ( lnvalue = 0 ; lnvalue <= 4 ; lnvalue++ ) {

                if ( gtable[ lnvalue] == value )

                        return( lnvalue );

                }


        return( -1 );


}
int ConvertTurbo( int value ) {

        int lnvalue, ttable[] = { 1, 2, 4, 8, 16 };



        // Compute the log of the gain value.

        for ( lnvalue = 0 ; lnvalue <= 4 ; lnvalue++ ) {

                if ( ttable[ lnvalue] == value )

                        return( lnvalue );

                }


        return( -1 );


}


// These are the INI report helpers.

char *OnOff( int value ) {

        return( value ? "On" : "Off" );

}
char *PortMode( int value ) {

        char *s;

        switch ( value ) {

                case PARXCH_PORT_MODE_BPP:     s = "BPP";     break;
                case PARXCH_PORT_MODE_EPP:     s = "EPP";     break;
                case PARXCH_PORT_MODE_ECP_BPP: s = "ECP/BPP"; break;
                case PARXCH_PORT_MODE_ECP_EPP: s = "ECP/EPP"; break;

                default:

                        s = "unknown PortMode"; break;
                }

        return( s );

}
char *DataCoding( int value ) {

        char *s;

        switch ( value ) {

                case PARXCH_DF_SIGNED: s = "Signed"; break;
                case PARXCH_DF_OFFSET: s = "Offset"; break;

                default:

                        s = "unknown DataCoding"; break;
                }

        return( s );

}
char *Gain( int value ) {

        char *s, *gtable[] = { "1", "2", "4", "8", "16" };


       if ( 0 <= value && value <= 4 )

                s = gtable[ value];

       else
                s = "unknown Gain";


       return( s );


}
char *Turbo( int value ) {

        char *s, *ttable[] = { "1", "2", "4", "8", "16" };


       if ( 0 <= value && value <= 4 )

                s = ttable[ value];

       else
                s = "unknown Turbo";


       return( s );


}
char *Calibration( int value ) {

        char *s;

        switch ( value ) {

                case PARXCH_CAL_SELF:           s = "Self";    break;
                case PARXCH_CAL_SYSTEM_OFFSET:  s = "SysOff";  break;
                case PARXCH_CAL_SYSTEM_FS:      s = "SysFull"; break;
                case PARXCH_CAL_NORMAL:         s = "None";    break;

                default:

                        s = "unknown Calibration"; break;
                }

        return( s );

}
char *TriggerType( int value ) {

        char *s;

        switch ( value ) {

                case TRIG_NONE:         s = "None";            break;
                case TRIG_THRESHOLD:    s = "Threshold";       break;
                case TRIG_DIGLEVEL:     s = "DigitalLevel";    break;
                case TRIG_DIGEDGE:      s = "DigitalEdge";     break;

                default:

                        s = "unknown TriggerType value"; break;
                }

        return( s );

}
char *TriggerActiveLevel( int value ) {

        char *s;

        switch ( value ) {

                case TRIGLEVEL_HIGH:    s = "High";            break;
                case TRIGLEVEL_LOW:     s = "Low";             break;

                default:

                        s = "unknown TriggerLevel value"; break;
                }

        return( s );

}
char *TriggerActiveEdge( int value ) {

        char *s;

        switch ( value ) {

                case TRIGEDGE_RISING:    s = "Rising";    break;
                case TRIGEDGE_FALLING:   s = "Falling";   break;

                default:

                        s = "unknown TriggerEdge value"; break;
                }

        return( s );

}
char *OutputFileFormat( int value ) {

        char *s;

        switch ( value ) {

                case NOFILE:       s = "None";     break;
                case OUTFILE:      s = "Out";      break;
                case DATFILE:      s = "Dat";      break;

                default:

                        s = "unknown OutputFileFormat value"; break;
                }

        return( s );

}
char *OutputFileNaming( int value ) {

        char *s;

        switch ( value ) {

                case NONE:           s = "None";              break;
                case SINGLE:         s = "Single";            break;
                case SEQUENTIALDEC:  s = "Sequential";        break;
                case TIME:           s = "Time";              break;
                case STDOUT:         s = "StdOut";            break;

                case SEQUENTIALHEX:  // fall through to unknown case
                default:

                        s = "unknown OutputFileNaming value"; break;
                }

        return( s );

}





/* INI FUNCTIONS:
 *
 * These are the major functions which fill and display the INI
 * structure.   The INI structure is filled by initialization with
 * defaults, by parsing the .ini file, and by computation.
 *
 * The INI structure is declared as a global variable at the top of the
 * program and additional .ini helper functions are above.
 *
 */

void IniDefaults( void ) {

        int i;


        // Setup default INI parameters.

        INI.Filename           = NULL;

        INI.AtodDriverName     = PARXCH_0;
        INI.AtodModel          = PARXCH_MODEL_UNKNOWN;
        INI.PortMode           = PARXCH_PORT_MODE_ECP_BPP;
        INI.DataCoding         = PARXCH_DF_SIGNED;
        INI.SamplingRate       = 50.0;
        INI.GainLog            = PARXCH_GAIN_MIN;
        INI.Calibration        = PARXCH_CAL_SELF;
        INI.ToggleLed          = ON;

        INI.AtodEnable              = ON;
        INI.KeypressToRun           = ON;
        INI.ExitOnError             = ON;
        INI.DisplayData             = ON;
        INI.DisplayInfo             = ON;
        INI.OutputFileFormat        = OUTFILE;
        INI.OutputFileNaming        = SINGLE;
        INI.OutputSeqStart          = 0;
        INI.OutputBlocks            = 2;

        INI.TriggerType             = NONE;
        INI.TriggerThreshold        = 0L;
        INI.TriggerDigitalLevelLine = 0;
        INI.TriggerDigitalEdgeLine  = 0;
        INI.TriggerActiveLevel      = TRIGLEVEL_HIGH;
        INI.TriggerActiveEdge       = TRIGEDGE_RISING;
        INI.TriggerActionSave       = ON;
        INI.TriggerFilesToSave      = 2;
        INI.TriggerActionSignal     = ON;
        INI.TriggerLineToSignal     = 0;
        INI.TriggerLatchLine        = OFF;

        INI.DigitalEnable          = OFF;

        INI.GpsEnable              = OFF;
        INI.GpsModel               = GPSMODEL_NONE;
        INI.GpsDriverName          = PARGPS_0;
        INI.GpsSerialPort          = SERIAL_PORT_1;
        INI.GpsDisplayPps          = ON;
        INI.GpsDisplaySerial       = ON;

        INI.CounterEnable          = OFF;

        INI.TurboLog               = PARXCH_TURBO_MAX;
        INI.Decimation             = PARXCH_DECIMATION_100Hz_T16;
        INI.ExtraDecimation        = PARXCH_EXTRADECIMATION_MIN;

        INI.ChBufPts               = 128; // arbitrary but reasonable number of
                                          // points per channel per buffer


        sprintf( INI.OutputDirectory, ".%c", SR_DIRSLASH ); // current dir
        
        for ( i = 0 ; i < PARXCH_MAX_ANALOG_CHANNELS ; i++ ) {
                sprintf( INI.ChannelTitle[i], "CH %d", i );
                INI.ChannelTrigger[i] = OFF;
                }

        if (PARXCH_MAX_DIGITAL_CHANNELS > 0) {
                i = PARXCH_MAX_ANALOG_CHANNELS;
                sprintf( INI.ChannelTitle[i], "DGTL" );
                INI.ChannelTrigger[i] = OFF;

                for ( i = 0 ; i < PARXCH_MAX_DIGPLOT_CHANNELS ; i++ ) {
                        sprintf( INI.DigitalTitle[i], "DIG%d", i );
                        }
                }

        if (PARXCH_MAX_GPS_CHANNELS > 0) {
                i = PARXCH_MAX_ANALOG_CHANNELS + PARXCH_MAX_DIGITAL_CHANNELS;
                sprintf( INI.ChannelTitle[i], "MARK" );
                INI.ChannelTrigger[i] = OFF;
                }

        if (PARXCH_MAX_COUNTER_CHANNELS > 0) {
                i = PARXCH_MAX_ANALOG_CHANNELS + PARXCH_MAX_DIGITAL_CHANNELS +
                                                          PARXCH_MAX_GPS_CHANNELS;
                sprintf( INI.ChannelTitle[i], "CNTR" );
                INI.ChannelTrigger[i] = OFF;
                }

}
void IniConsistencyCheck( void ) {


        // Verify there are no conflicting INI selections.

#define BAD_GPSOUT  "GPS data will not be saved.  Use .DAT file instead."
#define BAD_DISPLAY "Data to std out.  Turn display/keypress options OFF."
#define BAD_COUNTER "Counter channel not supported for this AtodModelName."
#define BAD_GPSRATE "GPS timing not compatible with rates < 40 Hz for 1CH/4CH."


        // Check for requesting GPS but not writing it out

        if ( (INI.OutputFileFormat == OUTFILE) && (INI.GpsEnable == ON) )

                LogMessage( BAD_GPSOUT );



        // Check for requesting output sent to standard out while also
        // requesting any option causing screen output.

        if ( INI.OutputFileNaming == STDOUT) {

                if ( (INI.DisplayData == ON)      ||
                     (INI.DisplayInfo == ON)      ||
                     (INI.GpsDisplayPps == ON)    ||
                     (INI.GpsDisplaySerial == ON) ||
                     (INI.KeypressToRun == ON)    )

                        ErrorWarning( BAD_DISPLAY );
                        }


        if ( INI.CounterEnable == ON  &&  INI.ParXchCounterChannels == 0) {
                ErrorWarning( BAD_COUNTER );
                INI.CounterEnable = OFF;
                }


//        if ( INI.ExtraDecimation != PARXCH_EXTRADECIMATION_MIN;
//        INI.GpsEnable == ON  &&  INI.ParXchCounterChannels == 0) {
//                ErrorWarning( BAD_GPSRATE );
//                INI.GpsEnable = OFF;
//                }
        

}
void IniCompute( void ) {

        int TriggerState;
        long GpsBytesPerSec, NsecPerFile;

        // Fill in the additional computed INI parameters.
        //
        // The user doesn't specify these INI parameters.  They are
        // computed and filled in, at runtime for convienient use.
        //
        //

        ParXchSpsGainToTde(
                           INI.SamplingRate,
                           INI.GainLog,
                           &INI.TurboLog,
                           &INI.Decimation,
                           &INI.ExtraDecimation
                           );

        ParXchTdeToSpsGain(
                           INI.TurboLog,
                           INI.Decimation,
                           INI.ExtraDecimation,
                           &INI.TrueSamplingRate,
                           &INI.GainLog
                           );


        // Finalize selection of specific PARxCH model
        
        if (INI.AtodModel == PARXCH_MODEL_UNKNOWN) // Still unspecified
                INI.AtodModel = ParXchGetDefaultModel( INI.AtodDriverName );


        // Set PARxCH model specific channel counts

        switch( INI.AtodModel ) {
                case PARXCH_MODEL_PAR1CH:
                        INI.ParXchAnalogChannels  = PAR1CH_ANALOG_CHANNELS;
                        INI.ParXchDigitalChannels = PAR1CH_DIGITAL_CHANNELS;
                        INI.ParXchGpsChannels     = PAR1CH_GPS_CHANNELS;
                        INI.ParXchCounterChannels = PAR1CH_COUNTER_CHANNELS;
                        INI.ParXchAnaPlotChannels = PAR1CH_ANAPLOT_CHANNELS;
                        INI.ParXchDigPlotChannels = PAR1CH_DIGPLOT_CHANNELS;
                        INI.ParXchGpsPlotChannels = PAR1CH_GPSPLOT_CHANNELS;
                        INI.ParXchCntPlotChannels = PAR1CH_CNTPLOT_CHANNELS;
                        strncpy( INI.ParXchModelName, PAR1CH_MODELNAME, INIMAXTITLE );
                        break;

                case PARXCH_MODEL_PAR4CH:
                        INI.ParXchAnalogChannels  = PAR4CH_ANALOG_CHANNELS;
                        INI.ParXchDigitalChannels = PAR4CH_DIGITAL_CHANNELS;
                        INI.ParXchGpsChannels     = PAR4CH_GPS_CHANNELS;
                        INI.ParXchCounterChannels = PAR4CH_COUNTER_CHANNELS;
                        INI.ParXchAnaPlotChannels = PAR4CH_ANAPLOT_CHANNELS;
                        INI.ParXchDigPlotChannels = PAR4CH_DIGPLOT_CHANNELS;
                        INI.ParXchGpsPlotChannels = PAR4CH_GPSPLOT_CHANNELS;
                        INI.ParXchCntPlotChannels = PAR4CH_CNTPLOT_CHANNELS;
                        strncpy( INI.ParXchModelName, PAR4CH_MODELNAME, INIMAXTITLE );
                        break;

                case PARXCH_MODEL_PAR8CH:
                        INI.ParXchAnalogChannels  = PAR8CH_ANALOG_CHANNELS;
                        INI.ParXchDigitalChannels = PAR8CH_DIGITAL_CHANNELS;
                        INI.ParXchGpsChannels     = PAR8CH_GPS_CHANNELS;
                        INI.ParXchCounterChannels = PAR8CH_COUNTER_CHANNELS;
                        INI.ParXchAnaPlotChannels = PAR8CH_ANAPLOT_CHANNELS;
                        INI.ParXchDigPlotChannels = PAR8CH_DIGPLOT_CHANNELS;
                        INI.ParXchGpsPlotChannels = PAR8CH_GPSPLOT_CHANNELS;
                        INI.ParXchCntPlotChannels = PAR8CH_CNTPLOT_CHANNELS;
                        strncpy( INI.ParXchModelName, PAR8CH_MODELNAME, INIMAXTITLE );
                        break;

                default:
                        ErrorFatal( "Invalid PARxCH model, please set AtodModelName correctly" );
                }
        
        INI.ParXchTotalChannels     = INI.ParXchAnalogChannels  +
                                      INI.ParXchDigitalChannels +
                                      INI.ParXchGpsChannels     +
                                      INI.ParXchCounterChannels;
        INI.ParXchTotalPlotChannels = INI.ParXchAnaPlotChannels +
                                      INI.ParXchDigPlotChannels +
                                      INI.ParXchGpsPlotChannels +
                                      INI.ParXchCntPlotChannels;


        INI.Channels = INI.ParXchAnalogChannels;

        // We don't allow any digital trigger channels
        // (amplitude triggering is on analog channels only)

        if ( INI.DigitalEnable == ON ) {
                DigitalIndex = INI.Channels;
                INI.Channels += INI.ParXchDigitalChannels;
                TriggerState = OFF;
                INI.ChannelTrigger[DigitalIndex] = TriggerState;
                }
        else
                DigitalIndex = -1;
                

        if ( INI.GpsEnable == ON ) {
                GpsMarkIndex = INI.Channels;
                INI.Channels += INI.ParXchGpsChannels;
                }
        else
                GpsMarkIndex = SRDAT_INDEXCHANNEL_NONE;



        // Keep CounterEnable in synch with GpsEnable until there
        // is a way to control it in the driver
        
        if (INI.ParXchCounterChannels > 0 && INI.GpsEnable)
                INI.CounterEnable = ON;
        else
                INI.CounterEnable = OFF;


        if ( INI.CounterEnable == ON ) {
                CounterIndex = INI.Channels;
                INI.Channels += INI.ParXchCounterChannels;
                }
        else
                CounterIndex = SRDAT_INDEXCHANNEL_NONE;

#ifdef GPS_AVAILABLE
        
        // Finalize GpsModel selection

        if (INI.GpsModel == GPSMODEL_UNKNOWN) // Still unspecified
                INI.GpsModel = ParGpsGetDefaultModel( INI.GpsDriverName );

        
        switch( INI.GpsModel ) {
                case GPSMODEL_GARMIN:
                        strncpy( INI.GpsModelName, GPSMODELNAME_GARMIN, INIMAXTITLE );
                        break;
                case GPSMODEL_TRIMBLE:
                        strncpy( INI.GpsModelName, GPSMODELNAME_TRIMBLE, INIMAXTITLE );
                        break;
                case GPSMODEL_ONCORE:
                        strncpy( INI.GpsModelName, GPSMODELNAME_ONCORE, INIMAXTITLE );
                        break;
                case GPSMODEL_PCTIME:
                        strncpy( INI.GpsModelName, GPSMODELNAME_PCTIME, INIMAXTITLE );
                        break;
                default:
                        if (INI.GpsEnable == ON)
                                ErrorFatal( "Invalid GPS model, please set GpsModelName correctly" );
                        else
                                strncpy( INI.GpsModelName, GPSMODELNAME_TRIMBLE, INIMAXTITLE );
                }
        
#endif // GPS_AVAILABLE
                        


        INI.Nvalues        = INI.ChBufPts * INI.Channels;
        INI.AcqBufSize     = INI.Nvalues * sizeof( long );
        INI.BufferTime     = INI.ChBufPts / INI.TrueSamplingRate;
        INI.TriggerTimeToSave  = (int)((float)(INI.TriggerFilesToSave *
                                 INI.OutputBlocks * INI.ChBufPts) /
                                 INI.TrueSamplingRate);

        // The sleepms variable controls how many milliseconds to wait
        // between each attempt to read new data from the PARxCH.  A
        // value of BufferTime * 100 means wait about 1/10th the amount
        // of time needed to acquire one buffer worth of data.
        //
        //   SleepMs = (BufferTime sec / 10) * 1000 ms/sec

        INI.SleepMs        = (int)(INI.BufferTime * 100.0L);


        // Compute basic output file size, then extend to include DAT
        // headers and tags.

        INI.OutputFileSize = INI.AcqBufSize * INI.OutputBlocks;
        
        if ( INI.OutputFileFormat == DATFILE )
                INI.OutputFileSize += (SRDAT_HDR_TOTALSIZE +
                                       SRDAT_TAGSIZE * (INI.OutputBlocks+1));



        // Gps related parameters.
        //

        INI.NgpsValues     = (int)INI.BufferTime + 2;
        INI.GpsBufSize     = INI.NgpsValues * sizeof( PPSTDATA );
        INI.SerBufSize     = INI.NgpsValues * sizeof( SERIALDATA );

        GpsBytesPerSec     = sizeof( PPSTDATA ) + sizeof( SERIALDATA );
        NsecPerFile        = (long)(((INI.ChBufPts * INI.OutputBlocks) /
                                     INI.TrueSamplingRate) + 0.5);


#ifdef GPS_AVAILABLE

        // Don't ask for more data than MAX_PARGPS_STORAGE since that is
        // the total amount that can be stored in the PARGPS device driver
        // software FIFO without losing data.

        if (INI.NgpsValues > MAX_PARGPS_STORAGE) {

                sprintf(
                   msg,
                   "Requested sample rate overflows driver PARGPS software FIFO\n"
                   );
                ErrorFatal( msg );
                }

        // Update file size estimate to include GPS parameters

        if (INI.GpsEnable)
                INI.OutputFileSize += (GpsBytesPerSec * NsecPerFile);
        
#endif


        // Verify there are no conflicting INI selections.

        IniConsistencyCheck( );


}
int IniParse( char *IniFilename ) {

        int n, Gain, Channel, Digital, MaxTotalChannels, TokenType;
        unsigned long MaxThreshold;
        static int LastChannel = -1;
        static int LastDigital = -1;
        static int IsChannel   = -1;
        static char drivername[DIM_STR], gpsname[DIM_STR], ininame[DIM_STR];
        static char tmpmsg[DIM_STR];


        // Open the .ini file.

        IniFp = fopen( IniFilename, "r" );

        if ( !IniFp ) {
                sprintf( tmpmsg, "unable to open .ini file %s", IniFilename ); 
                Usage( tmpmsg );
                }

        else {
                SrStringCopy( ininame, IniFilename, DIM_STR );
                INI.Filename = ininame;
                }


        MaxTotalChannels = PARXCH_MAX_ANALOG_CHANNELS  +
                           PARXCH_MAX_DIGITAL_CHANNELS +
                           PARXCH_MAX_GPS_CHANNELS     +
                           PARXCH_MAX_COUNTER_CHANNELS;



        // Parse the ini file.

parse:
        GetToken();


        // Atod keywords

        if ( KeywordEquals( "AtodDriverName" ) ||
             KeywordEquals( "PortName" )       ) {

                TokenType = GetToken();
                if ( TokenType != STRING  &&  TokenType != SYMBOL )

                        ErrorIni( "bad AtoD Driver Name string" );

                SrStringCopy( drivername, TOKEN, DIM_STR );
                INI.AtodDriverName = drivername;

                }



        else if ( KeywordEquals( "AtodPortMode" ) ||
                  KeywordEquals( "PortMode" )     ) {

                GetToken();


                if ( Match( "BPP" ) )

                        INI.PortMode = PARXCH_PORT_MODE_BPP;

                else if ( Match( "EPP" ) )

                        INI.PortMode = PARXCH_PORT_MODE_EPP;

                else if ( Match( "ECP/BPP" ) )

                        INI.PortMode = PARXCH_PORT_MODE_ECP_BPP;

                else if ( Match( "ECP/EPP" ) )

                        INI.PortMode = PARXCH_PORT_MODE_ECP_EPP;

                else
                        ErrorIni( "PortMode must = {BPP,EPP,ECP/BPP,ECP/EPP}" );

                }



        else if ( KeywordEquals( "AtodModelName" ) ) {

                GetToken();


                if ( Match( "PAR1CH" ) )

                        INI.AtodModel = PARXCH_MODEL_PAR1CH;

                else if ( Match( "PAR4CH" ) )

                        INI.AtodModel = PARXCH_MODEL_PAR4CH;

                else if ( Match( "PAR8CH" ) )

                        INI.AtodModel = PARXCH_MODEL_PAR8CH;

                else if ( Match( "DEFAULT" ) )

                        INI.AtodModel = PARXCH_MODEL_UNKNOWN;

                else
                        ErrorIni( "AtodModelName must = {PAR1CH,PAR4CH,PAR8CH,DEFAULT}" );

                }



        else if ( KeywordEquals( "DataCoding" ) ) {

                GetToken();


                if ( Match( "Signed" ) )

                        INI.DataCoding = PARXCH_DF_SIGNED;

                else if ( Match( "Offset" ) )

                        INI.DataCoding = PARXCH_DF_OFFSET;

                else
                        ErrorIni( "DataCoding must = {Signed,Offset}" );

                }



        else if ( KeywordEquals( "SamplingRate" ) ) {

                GetFloat( &INI.SamplingRate, "SamplingRate must = float" );

                if ( INI.SamplingRate < 0.00001 || 15625.00 < INI.SamplingRate )

                        ErrorIni("SamplingRate must be 0.00001 to 15625.00 Hz");

                }



        else if ( KeywordEquals( "Gain" ) ) {

                GetInt( &Gain, "illegal Gain integer value" );

                INI.GainLog = ConvertGain( Gain );

                if ( INI.GainLog == -1 )

                      ErrorIni( "Gain must = {1,2,4,8,16}" );
                else
                      INI.TurboLog = PARXCH_GAIN_MAX - INI.GainLog;

                }



        else if ( KeywordEquals( "Decimation" ) ) {

                GetInt( &INI.Decimation, "illegal Decimation integer value" );

                if ( INI.Decimation < PARXCH_DECIMATION_MIN ||
                     INI.Decimation > PARXCH_DECIMATION_MAX )

                      ErrorIni( "Decimation must be 19 to 8000" );

                }



        else if ( KeywordEquals( "ExtraDecimation" ) ) {

                GetInt( &INI.ExtraDecimation, "illegal ExtraDecimation integer value" );

                if ( INI.ExtraDecimation < PARXCH_EXTRADECIMATION_MIN )

                      ErrorIni( "ExtraDecimation must 1 or greater" );

                }



        else if ( KeywordEquals( "Calibration" ) ) {

                GetToken();


                if ( Match( "Self" ) )

                        INI.Calibration = PARXCH_CAL_SELF;

                else if ( Match( "SysOff" ) )

                        INI.Calibration = PARXCH_CAL_SYSTEM_OFFSET;

                else if ( Match( "SysFull" ) )

                        INI.Calibration = PARXCH_CAL_SYSTEM_FS;

                else if ( Match( "None" ) )

                        INI.Calibration = PARXCH_CAL_NORMAL;

                else
                     ErrorIni("Calibration must = {Self,SysOff,SysFull,None}");

                }



         else if ( KeywordEquals( "ToggleLed" ) )

                GetOnOff( &INI.ToggleLed, "ToggleLed must = {ON,OFF}" );



        else if ( KeywordEquals( "AtodEnable" )   ||
                  KeywordEquals( "SystemStatus" ) )

                GetOnOff( &INI.AtodEnable, "AtodEnable must = {ON,OFF}" );


        // Running keywords

        else if ( KeywordEquals( "KeypressToRun" ) )

                GetOnOff( &INI.KeypressToRun, "KeypressToRun must = {ON,OFF}" );


        else if ( KeywordEquals( "ExitOnError" ) )

                GetOnOff( &INI.ExitOnError, "ExitOnError must = {ON,OFF}" );



        // Display keywords

        else if ( KeywordEquals( "DisplayDataText" ) ||
                  KeywordEquals( "DisplayData" ) ||
                  KeywordEquals( "DataDisplay" ) )

                GetOnOff( &INI.DisplayData, "DisplayDataText must = {ON,OFF}" );


        else if ( KeywordEquals( "DisplayInfo" ) ||
                  KeywordEquals( "InfoDisplay" ) )

                GetOnOff( &INI.DisplayInfo, "DisplayInfo must = {ON,OFF}" );


        // Output keywords

        else if ( KeywordEquals( "OutputDirectory" ) ) {

                if ( GetToken() != STRING )
                        ErrorIni( "illegal OutputDirectory string" );
                else {
                        if ( !SrFileExist( TOKEN ) ) {
                                sprintf(
                                        msg,
                                        "OutputDirectory %s must already exist",
                                        TOKEN
                                        );
                                ErrorIni( msg );
                                }

                        // Make sure directory name ends in a slash
                        // and leaves enough room for an 8.3 filename

                        n = strlen( TOKEN ) - 1;
                        if ( TOKEN[n] != SR_DIRSLASH ) {
                                n++;
                                TOKEN[n] = SR_DIRSLASH;
                                TOKEN[n+1] = '\0';
                                }
                        if (n >= INIMAXSTR-12-1)
                                ErrorIni( "OutputDirectory name too long" );

                        SrStringCopy( INI.OutputDirectory, TOKEN, INIMAXSTR-12 );
                        }
                }


        else if ( KeywordEquals( "OutputFileFormat" ) ) {

                GetToken();


                if ( Match( "None" ) )

                        INI.OutputFileFormat = NOFILE;

                else if ( Match( "Out" ) )

                        INI.OutputFileFormat = OUTFILE;

                else if ( Match( "Dat" ) )

                        INI.OutputFileFormat = DATFILE;

                else

                    ErrorIni( "OutputFileFormat must = {None,Out,Dat}" );
                }


        else if ( KeywordEquals( "OutputFileNaming" ) ||
                  KeywordEquals( "OutputFileMode" )   ) {

                GetToken();


                if ( Match( "NONE" ) ) {

                        INI.OutputFileNaming = NONE;
                        printf( "Use of this option is discouraged.\n" );
                        printf( "Please use OutputFileFormat = NONE instead.\n" );
                        }

                else if ( Match( "Single" ) )

                        INI.OutputFileNaming = SINGLE;

                else if ( Match( "Sequential" ) )

                        INI.OutputFileNaming = SEQUENTIALDEC;

                else if ( Match( "Time" ) )

                        INI.OutputFileNaming = TIME;

                else if ( Match( "StdOut" ) )

                        INI.OutputFileNaming = STDOUT;

                else

                    ErrorIni(

                        "OutputFileNaming must = {Single,Sequential,Time,StdOut}"

                            );
                }


        else if ( KeywordEquals( "OutputSeqStart" ) ) {

                GetToken();
                if ( ttype == INTEGER )
                        sscanf( TOKEN, "%u", &INI.OutputSeqStart );
                else if ( ttype == HEX ) {
                        sscanf( TOKEN, "%x", &INI.OutputSeqStart );
                        printf( "WARNING[%s] : line %d : OutputSeqStart is now a decimal integer, not hex\n", INI.Filename, fline );
                        }
                else
                        ErrorIni( "illegal OutputSeqStart integer value" );
                }


        else if ( KeywordEquals( "OutputBlocks" ) ||
                  KeywordEquals( "Blocks" ) )

                GetUlong( &INI.OutputBlocks,
                          "illegal OutputBlocks integer value" );



        // Channel keywords

        else if ( Match( "Channel") ) {

                GetInt( &Channel, "illegal CHANNEL number");

                if ( GetToken() != EQUALS )

                        ErrorIni( "missing = sign on CHANNEL");

                if ( GetToken() != STRING )

                        ErrorIni( "illegal CHANNEL title");


                // Make sure the channel is in the range.

                if ( Channel < 0 || MaxTotalChannels <= Channel )

                        ErrorIni( "channel number out of range");


                // Assign title to channel and null terminate.

                strncpy( INI.ChannelTitle[Channel], TOKEN, INIMAXTITLE );
                INI.ChannelTitle[Channel][INIMAXTITLE-1] = '\0';
                LastChannel = Channel;
                IsChannel   = 1;
                }



        // Digital keywords

        else if ( KeywordEquals( "DigitalEnable" ) )

                GetOnOff( &INI.DigitalEnable, "DigitalEnable must = {ON,OFF}" );


        else if ( Match( "Digital") ) {

                GetInt( &Digital, "illegal Digital number");

                if ( GetToken() != EQUALS )

                        ErrorIni( "missing = sign on Digital");

                if ( GetToken() != STRING )

                        ErrorIni( "illegal Digital title");


                // Make sure the Digital is in the range.

                if ( Digital < 0 || PARXCH_MAX_DIGPLOT_CHANNELS <= Digital )

                        ErrorIni( "Digital number out of range");


                // Assign title to Digital and null terminate.

                strncpy( INI.DigitalTitle[Digital], TOKEN, INIMAXTITLE );
                INI.DigitalTitle[Digital][INIMAXTITLE-1] = '\0';
                LastDigital = Digital;
                IsChannel   = 0;
                }
        



        // Trigger keywords

        else if ( KeywordEquals( "Trigger") ) {

                if ( IsChannel == -1 )
                        ErrorIni( "missing Channel keyword before trigger" );

                if (IsChannel == 1) 
                        GetOnOff( &INI.ChannelTrigger[LastChannel],
                                  "Trigger must = {ON,OFF}" );
                else { // IsChannel == 0
                        sprintf( msg, "WARNING: Digital channels may not be used for amplitude\ntriggering.  Use digital level or edge triggering instead.\n" );
                        ErrorIni( msg );
                        }

                IsChannel   = -1;
                LastChannel = -1;
                LastDigital = -1;
                }



        else if ( KeywordEquals( "TriggerType" ) ) {

                GetToken();

                if ( Match( "None" ) )

                        INI.TriggerType = TRIG_NONE;

                else if ( Match( "Threshold" ) )

                        INI.TriggerType = TRIG_THRESHOLD;

                else if ( Match( "DigitalLevel" ) ||
                          Match( "DigitalSignal" ) )

                        INI.TriggerType = TRIG_DIGLEVEL;

                else if ( Match( "DigitalEdge" ) )

                        INI.TriggerType = TRIG_DIGEDGE;

                else
                    ErrorIni( "TriggerType must = {None,Threshold,DigitalLevel,DigitalEdge}" );

                }


        else if ( KeywordEquals( "TriggerThreshold" ) ) {

                MaxThreshold = 8388607L;
                GetUlong( &INI.TriggerThreshold,
                          "illegal TriggerThreshold integer value" );

                if ( (INI.TriggerThreshold <= 0) ||
                     (INI.TriggerThreshold > MaxThreshold)  ) {

                        sprintf( msg,
                                 "TriggerThreshold must be > 0 and less than max amplitude %d (0x%X)",
                                 MaxThreshold, MaxThreshold );
                        ErrorIni( msg );
                        }
                }


        else if ( KeywordEquals( "TriggerDigitalLevelLine" ) ||
                  KeywordEquals( "TriggerDigitalLine" )      ) {

                GetInt( &INI.TriggerDigitalLevelLine,
                        "illegal TriggerDigitalLevelLine integer value" );

                if ( (INI.TriggerDigitalLevelLine < 0) ||
                     (INI.TriggerDigitalLevelLine > 3)  ) {

                        ErrorIni( "TriggerDigitalLevelLine must be from 0 to 3" );
                        }
                }


        else if ( KeywordEquals( "TriggerDigitalEdgeLine" ) ) {

                GetInt( &INI.TriggerDigitalEdgeLine,
                        "illegal TriggerDigitalEdgeLine integer value" );

                if ( (INI.TriggerDigitalEdgeLine < 0) ||
                     (INI.TriggerDigitalEdgeLine > 3)  ) {

                        ErrorIni( "TriggerDigitalEdgeLine must be from 0 to 3" );
                        }
                }



        else if ( KeywordEquals( "TriggerActiveLevel" ) ) {

                GetToken();

                if ( Match( "High" ) )

                        INI.TriggerActiveLevel = TRIGLEVEL_HIGH;

                else if ( Match( "Low" ) )

                        INI.TriggerActiveLevel = TRIGLEVEL_LOW;

                else
                    ErrorIni( "TriggerActiveLevel must = {High,Low}" );

                }


        else if ( KeywordEquals( "TriggerActiveEdge" ) ) {

                GetToken();

                if ( Match( "Rising" ) )

                        INI.TriggerActiveEdge = TRIGEDGE_RISING;

                else if ( Match( "Falling" ) )

                        INI.TriggerActiveEdge = TRIGEDGE_FALLING;

                else
                    ErrorIni( "TriggerActiveEdge must = {Rising,Falling}" );

                }


        else if ( KeywordEquals( "TriggerActionSave" ) )

                GetOnOff( &INI.TriggerActionSave, "TriggerActionSave must = {On,Off}" );


        else if ( KeywordEquals( "TriggerFilesToSave" ) ) {

                GetInt( &INI.TriggerFilesToSave,
                          "illegal TriggerFilesToSave integer value" );

                if ( INI.TriggerFilesToSave <= 0 )

                        ErrorIni( "TriggerFilesToSave must be > 0" );
                }


        else if ( KeywordEquals( "TriggerActionSignal" ) )

                GetOnOff( &INI.TriggerActionSignal, "TriggerActionSignal must = {On,Off}" );


        else if ( KeywordEquals( "TriggerLineToSignal" ) ) {

                GetInt( &INI.TriggerLineToSignal,
                          "illegal TriggerLineToSignal integer value" );

                if ( (INI.TriggerLineToSignal < 0) ||
                     (INI.TriggerLineToSignal > 3)  ) {

                        ErrorIni( "TriggerLineToSignal must be from 0 to 3" );
                        }
                }


        else if ( KeywordEquals( "TriggerLatch" ) )

                GetOnOff( &INI.TriggerLatchLine, "TriggerLatch must = {On,Off}" );





        // Gps keywords

        else if ( KeywordEquals( "GpsEnable" ) ) {

                GetOnOff( &INI.GpsEnable, "GpsEnable must = {On,Off}" );

#if !defined( GPS_AVAILABLE )

                if ( INI.GpsEnable == ON )
                        LogMessage( "For full GPS support, recompile with GPS_AVAILABLE defined." );
#endif

                }



        else if ( KeywordEquals( "GpsDriverName" ) ) {

                TokenType = GetToken();
                if ( TokenType != STRING  &&  TokenType != SYMBOL )

                        ErrorIni( "bad Gps Driver Name string" );

                strncpy( gpsname, TOKEN, DIM_STR );
                gpsname[DIM_STR-1] = '\0';
                INI.GpsDriverName = gpsname;

                }


        else if ( KeywordEquals( "GpsModelName" ) ) {

                GetToken();


#ifdef GPS_AVAILABLE
                
                if ( Match( "GARMIN" ) )

                        INI.GpsModel = GPSMODEL_GARMIN;

                else if ( Match( "TRIMBLE" ) )

                        INI.GpsModel = GPSMODEL_TRIMBLE;

                else if ( Match( "ONCORE" ) )

                        INI.GpsModel = GPSMODEL_ONCORE;

                else if ( Match( "PCTIME" ) )

                        INI.GpsModel = GPSMODEL_PCTIME;

                else if ( Match( "PARGPS" ) ) {
                        ErrorIni( "PARGPS is no longer a valid GpsModel option, replace with TRIMBLE." );
                        INI.GpsModel = GPSMODEL_TRIMBLE;
                        }

                else if ( Match( "DEFAULT" ) )

                        INI.GpsModel = GPSMODEL_UNKNOWN;

                else
                        ErrorIni( "GpsModelName must = {GARMIN,TRIMBLE,ONCORE,PCTIME,DEFAULT}" );

#else // !defined( GPS_AVAILABLE )

                if ( INI.GpsEnable == ON )
                        LogMessage( "For full GPS support, recompile with GPS_AVAILABLE defined." );
#endif
                }



        else if ( KeywordEquals( "GpsSerialPort" ) ) {

                GetInt( &INI.GpsSerialPort,
                        "illegal GpsSerialPort integer value" );

                if ( INI.GpsSerialPort < 0 || INI.GpsSerialPort > 4)

                        ErrorIni( "GpsSerialPort must be 0 to 4, typically 1" );

                }


        else if ( KeywordEquals( "GpsDisplayPps" ) )

                GetOnOff( &INI.GpsDisplayPps, "GpsDisplayPps must = {On,Off}" );


        else if ( KeywordEquals( "GpsDisplaySerial" ) )

                GetOnOff( &INI.GpsDisplaySerial,
                          "GpsDisplaySerial must = {On,Off}" );




        // Counter keywords

        else if ( KeywordEquals( "CounterEnable" ) ) {

                GetOnOff( &INI.CounterEnable, "CounterEnable must = {On,Off}" );

                }




        // Unused keywords
        //
        // These options are not used in simp at this time, but they
        // are used by the scope acquisition program so just pass over
        // them here ...
        //
        //

        else if ( KeywordEquals( "Title" )                ||
                  KeywordEquals( "ForbidMultiple" )       ||
                  KeywordEquals( "Interactive" )          ||
                  KeywordEquals( "DisplayDataTrace" )     ||
                  KeywordEquals( "DisplayColor" )         ||
                  KeywordEquals( "DisplayNbits" )         ||
                  KeywordEquals( "DisplayAutoDC" )        ||
                  KeywordEquals( "DisplayClipping" )      ||
                  KeywordEquals( "DisplayFirstSample" )   ||
                  KeywordEquals( "DisplayFirstTrace" )    ||
                  KeywordEquals( "DisplayNtraces" ) ) {

                GetToken();  // ignore them ...

                }


        else if ( ttype == EOF ) {

                // Close the ini file

                fclose( IniFp );

                return( 1 );
                }



        else // token unrecognized ...

                ErrorIni( "unknown simp.ini syntax" );




        // Continue parsing.

        goto parse;



}




char *IniReportFmt[] = {

        "; PARxCH INI DATA from [%s]:                            \n",
        "                                                        \n",
        "              AtodEnable = %s                           \n",
        "          AtodDriverName = %s                           \n",
        "           AtodModelName = %s                           \n",
        "            AtodPortMode = %s                           \n",
        "                                                        \n",
        "              DataCoding = %s                           \n",
        "            SamplingRate = %f                           \n",
        "                    Gain = %s                           \n",
        "               ToggleLed = %s                           \n",
        "                                                        \n",
        "               GpsEnable = %s                           \n",
        "            GpsModelName = %s                           \n",
        "           GpsDriverName = %s                           \n",
        "           GpsSerialPort = %d                           \n",
        "           GpsDisplayPps = %s                           \n",
        "        GpsDisplaySerial = %s                           \n",
        "                                                        \n",
        "              Channel %2d = \"%s\"   Trigger = %s          \n",
        "                                                        \n",
        "           DigitalEnable = %s                           \n",
        "              Digital %2d = \"%s\"                         \n",
        "                                                        \n",
        "             TriggerType = %s                           \n",
        "        TriggerThreshold = %lu                          \n",
        " TriggerDigitalLevelLine = %d                           \n",
        "  TriggerDigitalEdgeLine = %d                           \n",
        "      TriggerActiveLevel = %s                           \n",
        "       TriggerActiveEdge = %s                           \n",
        "       TriggerActionSave = %s                           \n",
        "      TriggerFilesToSave = %d                           \n",
        "     TriggerActionSignal = %s                           \n",
        "     TriggerLineToSignal = %d                           \n",
        "            TriggerLatch = %s                           \n",
        "                                                        \n",
        "             DisplayInfo = %s                           \n",
        "         DisplayDataText = %s                           \n",
        "                                                        \n",
        "        OutputFileFormat = %s                           \n",
        "        OutputFileNaming = %s                           \n",
        "          OutputSeqStart = %d                           \n",
        "            OutputBlocks = %lu                          \n",
        "                                                        \n",
        "           KeypressToRun = %s                           \n",
        "             ExitOnError = %s                           \n",
        "                                                        \n",
        "                                                        \n",
        "; Additional items implied by these parameters:         \n",
        ";                                                       \n",
        ";               Channels = %d                           \n",
        ";               ChBufPts = %u                           \n",
        ";                Nvalues = %u                           \n",
        ";             NgpsValues = %u                           \n",
        ";             AcqBufSize = %u bytes                     \n",
        ";             GpsBufSize = %u bytes                     \n",
        ";             SerBufSize = %u bytes                     \n",
        ";         OutputFileSize = %lu bytes                    \n",
        ";      TriggerTimeToSave = %d seconds                   \n",
        ";                                                       \n",
        ";                  Turbo = %s                           \n",
        ";             Decimation = %d                           \n",
        ";        ExtraDecimation = %d                           \n",
        ";       TrueSamplingRate = %f Hz                        \n",
        ";             BufferTime = %f sec                       \n",
        ";                                                       \n",

        NULL,

        };



void IniReport( void ) {

        FILE *RptFp;
        char **fmt;
        int  i;


        // Generate the ASCII INI report file, simp.rpt.

        sprintf( FullFileName, "%sSIMP.RPT", INI.OutputDirectory );

        RptFp = fopen( FullFileName, "w" );

        if ( !RptFp )

                ErrorFatal( "unable to open simp.rpt report file" );




        DisplayInfo( "Generating simp.rpt ... \n" );



        fmt = IniReportFmt;

        fprintf( RptFp, Chop(*fmt++),                  INI.Filename          );
        fprintf( RptFp, Chop(*fmt++)                                         );

        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.AtodEnable         ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.AtodDriverName       );
        fprintf( RptFp, Chop(*fmt++),                  INI.ParXchModelName      );
        fprintf( RptFp, Chop(*fmt++),        PortMode( INI.PortMode           ) );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),      DataCoding( INI.DataCoding         ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.SamplingRate         );
        fprintf( RptFp, Chop(*fmt++),            Gain( INI.GainLog            ) );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.ToggleLed          ) );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.GpsEnable          ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.GpsModelName         );
        fprintf( RptFp, Chop(*fmt++),                  INI.GpsDriverName        );
        fprintf( RptFp, Chop(*fmt++),                  INI.GpsSerialPort        );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.GpsDisplayPps      ) );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.GpsDisplaySerial   ) );
        fprintf( RptFp, Chop(*fmt++)                                            );

        
        for ( i = 0 ; i < INI.Channels ; i++ )
                fprintf( RptFp, Chop( *fmt ), i, INI.ChannelTitle[i],
                                              OnOff( INI.ChannelTrigger[i] ) );
        fmt++;

        fprintf( RptFp, Chop(*fmt++)                                         );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.DigitalEnable   ) );
        if (INI.DigitalEnable) {
                for ( i = 0 ; i < INI.ParXchDigPlotChannels ; i++ )
                        fprintf( RptFp, Chop( *fmt ), i, INI.DigitalTitle[i] );
                }
        fmt++;

        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),     TriggerType( INI.TriggerType        ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.TriggerThreshold     );
        fprintf( RptFp, Chop(*fmt++),                  INI.TriggerDigitalLevelLine);
        fprintf( RptFp, Chop(*fmt++),                  INI.TriggerDigitalEdgeLine );
        fprintf( RptFp, Chop(*fmt++), TriggerActiveLevel(INI.TriggerActiveLevel   ) );
        fprintf( RptFp, Chop(*fmt++),  TriggerActiveEdge(INI.TriggerActiveEdge    ) );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.TriggerActionSave  ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.TriggerFilesToSave   );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.TriggerActionSignal) );
        fprintf( RptFp, Chop(*fmt++),                  INI.TriggerLineToSignal  );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.TriggerLatchLine   ) );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.DisplayInfo        ) );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.DisplayData        ) );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++), OutputFileFormat(INI.OutputFileFormat   ) );
        fprintf( RptFp, Chop(*fmt++), OutputFileNaming(INI.OutputFileNaming   ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.OutputSeqStart       );
        fprintf( RptFp, Chop(*fmt++),                  INI.OutputBlocks         );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.KeypressToRun      ) );
        fprintf( RptFp, Chop(*fmt++),           OnOff( INI.ExitOnError        ) );
        fprintf( RptFp, Chop(*fmt++)                                            );
        fprintf( RptFp, Chop(*fmt++)                                            );

        fprintf( RptFp, Chop(*fmt++)                                         );
        fprintf( RptFp, Chop(*fmt++)                                         );
        fprintf( RptFp, Chop(*fmt++),                  INI.Channels          );
        fprintf( RptFp, Chop(*fmt++),                  INI.ChBufPts          );
        fprintf( RptFp, Chop(*fmt++),                  INI.Nvalues           );
        fprintf( RptFp, Chop(*fmt++),                  INI.NgpsValues        );
        fprintf( RptFp, Chop(*fmt++),                  INI.AcqBufSize        );
        fprintf( RptFp, Chop(*fmt++),                  INI.GpsBufSize        );
        fprintf( RptFp, Chop(*fmt++),                  INI.SerBufSize        );
        fprintf( RptFp, Chop(*fmt++),                  INI.OutputFileSize    );
        fprintf( RptFp, Chop(*fmt++),                  INI.TriggerTimeToSave );
        fprintf( RptFp, Chop(*fmt++)                                         );
        fprintf( RptFp, Chop(*fmt++),           Turbo( INI.TurboLog        ) );
        fprintf( RptFp, Chop(*fmt++),                  INI.Decimation        );
        fprintf( RptFp, Chop(*fmt++),                  INI.ExtraDecimation   );
        fprintf( RptFp, Chop(*fmt++),                  INI.TrueSamplingRate  );
        fprintf( RptFp, Chop(*fmt++),                  INI.BufferTime        );
        fprintf( RptFp, Chop(*fmt++)                                         );

        fclose( RptFp );


}










/* HEADER DATA AND REPORTING:
 *
 * Two different header formats currently in use.  One, defined in
 * simp.h, is for the .HDR file used with .OUT data files.  The
 * other, defined in srdat.h, is for the 4096 byte header that starts
 * each .DAT file.
 *
 *
 */

SROUT_HDR OutHdr;  // <<< see simp.h for .OUT file .HDR layout ...
SRDAT_HDR DatHdr;  // <<< see srdat.h for .DAT file header layout ...

void HdrDefaults( void ) {

        int  i;


        // Fill the specified header type with appropriate default values.


        switch ( INI.OutputFileFormat ) {

          case OUTFILE:

            // Initialize the .OUT file .HDR structure

            SrStringCopy( OutHdr.Id, HDR_PRODUCT_STRING, 32 );

            OutHdr.Channels     = INI.Channels;
            OutHdr.RecordPts    = INI.ChBufPts * INI.OutputBlocks;
            OutHdr.WordSize     = 4;
            OutHdr.DataCoding   = !INI.DataCoding;  // XCH sense reversed from hdr
            OutHdr.SamplingRate = INI.TrueSamplingRate;
            OutHdr.Gain         = 1 << INI.GainLog;
            OutHdr.FilterCoeff  = 0.0;
            OutHdr.FilterScale  = 1.0;


            for ( i = 0 ; i < OutHdr.Channels ; i++ ) {

                    OutHdr.ChannelList[i].channel = i;
                    sprintf( OutHdr.ChannelList[i].title, INI.ChannelTitle[i] );
                    }

            break;


          case DATFILE:

            // Initialize .DAT file 4096 byte header.
            // Fill with generic defaults, then with PARxCH specific values.

            SrDatHdrInit( &DatHdr );

            SrStringCopy( DatHdr.DatProduct, HDR_PRODUCT_STRING, SRDAT_DIM_STRING );
            SrStringCopy( DatHdr.DatApplication, "SR SIMP", SRDAT_DIM_STRING );

            if (INI.GpsEnable)
                    DatHdr.DatGpsModel = INI.GpsModel;
            else
                    DatHdr.DatGpsModel = GPSMODEL_NONE;
            DatHdr.DatAtodModel        = INI.AtodModel;

            DatHdr.FileNameType      = INI.OutputFileNaming;
            if (INI.OutputFileNaming == SEQUENTIALDEC ||
                INI.OutputFileNaming == SEQUENTIALHEX )
                    DatHdr.FileNameInc = 1;
            else if (INI.OutputFileNaming == TIME)
                    DatHdr.FileNameInc = (long)(INI.OutputBlocks *
                                                 INI.BufferTime + 0.5);


            DatHdr.Channels       = INI.Channels;
            DatHdr.ChannelsAnalog = INI.ParXchAnalogChannels;
            
            if ( INI.DigitalEnable == ON  &&  INI.ParXchDigitalChannels > 0) {
                    DatHdr.ChannelsDigital  = INI.ParXchDigitalChannels;
                    DatHdr.DigitalPerChan   = INI.ParXchDigPlotChannels;
                    DatHdr.ChannelDigIndex  = DigitalIndex;
                    }
            else {
                    DatHdr.ChannelsDigital  = 0;
                    DatHdr.DigitalPerChan   = 0;
                    DatHdr.ChannelDigIndex  = SRDAT_INDEXCHANNEL_NONE;
                    }
            
            if ( INI.GpsEnable == ON  &&  INI.ParXchGpsChannels > 0) {
                    DatHdr.ChannelsGps      = INI.ParXchGpsChannels;
                    DatHdr.ChannelMarkIndex = GpsMarkIndex;
                    }
            else {
                    DatHdr.ChannelsGps      = 0;
                    DatHdr.ChannelMarkIndex = SRDAT_INDEXCHANNEL_NONE;
                    }
            
            if ( INI.CounterEnable == ON  &&  INI.ParXchCounterChannels > 0) {
                    DatHdr.ChannelsCounter  = INI.ParXchCounterChannels;
                    DatHdr.ChannelObcIndex  = CounterIndex;
                    }
            else {
                    DatHdr.ChannelsCounter  = 0;
                    DatHdr.ChannelObcIndex  = SRDAT_INDEXCHANNEL_NONE;
                    }


            
            DatHdr.ChannelPtsPerRecord = INI.ChBufPts;
            DatHdr.ChannelPtsPerFile   = INI.ChBufPts * INI.OutputBlocks;
            DatHdr.AtodWordSize        = 4;
            DatHdr.AtodDataCoding      = !INI.DataCoding;  // XCH sense reversed from hdr
            DatHdr.AtodSamplingRate    = INI.TrueSamplingRate;
            DatHdr.AtodGain            = 1 << INI.GainLog;


            for ( i = 0 ; i < DatHdr.Channels ; i++ ) {
                    DatHdr.ChannelList[i].Channel = i;
                    sprintf( DatHdr.ChannelList[i].Title, INI.ChannelTitle[i] );
                    }

            for ( i = 0 ; i < DatHdr.DigitalPerChan ; i++ ) {
                    DatHdr.DigitalList[i].Channel = i;
                    sprintf( DatHdr.DigitalList[i].Title, INI.DigitalTitle[i] );
                    }
            

            DatHdr.TriggerType = INI.TriggerType;
            if (INI.TriggerType == TRIG_THRESHOLD) {
                    DatHdr.TriggerParm1 = INI.TriggerThreshold;
                    DatHdr.TriggerParm2 = INI.ChannelTrigger[INI.Channels-1];
                    for ( i = 2 ; i < INI.Channels ; i++ ) {
                            DatHdr.TriggerParm2 *= 2; // shift left 1 bit
                            DatHdr.TriggerParm2 |= INI.ChannelTrigger[INI.Channels-i];
                            }

                    }
            else if (INI.TriggerType == TRIG_DIGLEVEL) {
                    DatHdr.TriggerParm1 = INI.TriggerDigitalLevelLine;
                    DatHdr.TriggerParm2 = INI.TriggerActiveLevel;
                    }
            else if (INI.TriggerType == TRIG_DIGEDGE) {
                    DatHdr.TriggerParm1 = INI.TriggerDigitalEdgeLine;
                    DatHdr.TriggerParm2 = INI.TriggerActiveEdge;
                    }

            // We would like to set the default Gps counter frequency
            // here, but the PARGPS has not been opened yet.  So we wait
            // and get the counter frequency just after the open call.

            break;



          default:

            // No header requested

            break;


          } // end switch


}

void HdrWriteFile( void ) {

        FILE *HdrFp;


        // Write the .OUT header structure out to a binary .HDR file.
        //
        // This function only pertains to the .OUT header structure which
        // is written to a single separate .HDR file once.
        //
        // See the HdrUpdateWrite function for .DAT file headers which are
        // written at the beginning of each .DAT file.
        //
        //

        if (INI.OutputFileFormat != OUTFILE)
                return;



        DisplayInfo( "Generating simp.hdr ... \n" );



        // Open file or get a handle to standard out as requested.

        if (INI.OutputFileNaming != STDOUT) {

                sprintf( FullFileName, "%sSIMP.HDR", INI.OutputDirectory );

                HdrFp = fopen( FullFileName, "wb" );

                if ( !HdrFp )
                        ErrorFatal( "unable to open simp.hdr ..." );
                }

        else
                HdrFp = stdout;



        // Write the .HDR information.

        if ( fwrite( &OutHdr, sizeof( OutHdr ), 1, HdrFp ) != 1 )

                ErrorFatal( "unable to write simp.hdr ..." );



        // Close file or flush standard out.

        if (INI.OutputFileNaming != STDOUT)

                fclose( HdrFp );

        else
                fflush( HdrFp );

}

void HdrUpdateWrite( char *FinalFileName,
                     FILE *fp,
                     time_t *AcquireTime,
                     int TimeSource
                    ) {

        struct tm *utc_time;
        long       modtime;



        // Update a .DAT header and write it out.
        //
        // This function pertains only to .DAT headers which are written
        // at the beginning of each .DAT file.
        //
        //

        if (INI.OutputFileFormat != DATFILE)
                return;



        // Update the DAT header structure with current information.

        strncpy( DatHdr.FileName, FinalFileName, SRDAT_DIM_STRING-1 );

        // Get file "create" time.  The st_ctime value doesn't seem
        // to be properly updated when re-creating a file, so use the
        // "modify" time instead since we've just opened it for creation.

        modtime = SrFileModTime( fp );
        if ( modtime != -1 ) {
                DatHdr.FileCreateTimeHi    = 0L;
                DatHdr.FileCreateTime      = modtime;
                }
// WCT: This section will need to be added to support times greater than year 2038
//        LARGE_INTEGER tmp64;
//        if ( fstat64( fh, &filestats ) != -1 ) {
//                        tmp64.QuadPart = (LONGLONG)filestats.st_mtime;
//                        DatHdr.FileCreateTimeHi    = tmp64.HighPart;
//                        DatHdr.FileCreateTime      = tmp64.LowPart;
//                }

        DatHdr.FileSeqNum                  = INI.OutputSeqStart;
        DatHdr.FileStartPtNum              = NpointsFromStart;
        DatHdr.TimePoint[0].Type           = SRDAT_TIME_TYPE_FIRSTPT;
        DatHdr.TimePoint[0].SampleFromRun  = NpointsFromStart;
        DatHdr.TimePoint[0].SampleFromFile = 0;
        DatHdr.TimePoint[0].GpsMarkIndex   = 0;


        utc_time = gmtime( AcquireTime );  // Split time into YMD and HMS

        DatHdr.TimePoint[0].Year           = utc_time->tm_year + 1900;
        DatHdr.TimePoint[0].Month          = utc_time->tm_mon + 1;
        DatHdr.TimePoint[0].Day            = utc_time->tm_mday;
        DatHdr.TimePoint[0].YMDSource      = TimeSource;
        DatHdr.TimePoint[0].Hour           = utc_time->tm_hour;
        DatHdr.TimePoint[0].Minute         = utc_time->tm_min;
        DatHdr.TimePoint[0].Second         = utc_time->tm_sec;
        DatHdr.TimePoint[0].Microsecond    = 0L;
        DatHdr.TimePoint[0].HMSSource      = TimeSource;



        // Write out the updated header

        if ( !SrDatHdrWrite( fp, &DatHdr ) )
                ErrorWarning( "Could not write updated .DAT header" );

}







/* START MESSAGE:
 *
 *
 *
 */

char *StartMsgFmt[] = {

        "                                                              \n",
        "Initializing PARxCH with: (see simp.rpt for more details)     \n",
        "                                                              \n",
        "       AtodDriverName = %s                                    \n",
        "        AtodModelName = %s                                    \n",
        "             PortMode = %s                                    \n",
        "           DataCoding = %s                                    \n",
        "     TrueSamplingRate = %f Hz                                 \n",
        "                 Gain = %s                                    \n",
        "          TriggerType = %s                                    \n",
        "        DigitalEnable = %s                                    \n",
        "            GpsEnable = %s                                    \n",
        "         GpsModelName = %s                                    \n",
        "        GpsDriverName = %s                                    \n",
        "        GpsSerialPort = %d                                    \n",
        "                                                              \n",

        NULL,

        };


void StartMsg( void ) {

        char **fmt;

        LogMessage( "Beginning simp." );

        if (INI.DisplayInfo == OFF)
                return;

        // Display a few of the initialization parameters.

        fmt = StartMsgFmt;

            printf( Chop( *fmt++ )                                       );
            printf( Chop( *fmt++ )                                       );
            printf( Chop( *fmt++ )                                       );
            printf( Chop( *fmt++ ),              INI.AtodDriverName      );
            printf( Chop( *fmt++ ),              INI.ParXchModelName     );
            printf( Chop( *fmt++ ),    PortMode( INI.PortMode         )  );
            printf( Chop( *fmt++ ),  DataCoding( INI.DataCoding       )  );
            printf( Chop( *fmt++ ),              INI.TrueSamplingRate    );
            printf( Chop( *fmt++ ),        Gain( INI.GainLog          )  );
            printf( Chop( *fmt++ ), TriggerType( INI.TriggerType      )  );
            printf( Chop( *fmt++ ),       OnOff( INI.DigitalEnable    )  );
            printf( Chop( *fmt++ ),       OnOff( INI.GpsEnable        )  );
            printf( Chop( *fmt++ ),              INI.GpsModelName        );
            printf( Chop( *fmt++ ),              INI.GpsDriverName       );
            printf( Chop( *fmt++ ),              INI.GpsSerialPort       );
            printf( Chop( *fmt++ )                                       );

}







/* KEYBOARD FUNCTIONS:
 *
 *
 *
 */

void KeyWaitForRun( void ) {

        int c;

        // Wait for a keypress.

        c = SrGetKeyWait( );


        // Quit if it was Ctrl+C.

        if ( c == 0x3 ) // exit Ctrl+C
                CleanExit( 0 );
        
}
int KeyProcess( int key ) {

        // Special key y is found

        if ( key == 'y' || key == 'Y' ) {

                // Turn off a latched trigger that is staYing on

                if ( INI.TriggerType != TRIG_NONE  && // triggering
                     INI.TriggerActionSignal       )  // digital output action
                        ParXchUserIoWr( ParXchHandle, 0x00 );


                // Now flush any remaining keys like the ENTER required
                // under Linux for the original key to be recognized

                while ( SrGetKeyCheck() != EOF )
                        ;


                // Return EOF so our special key won't cause a quit

                return( EOF );
                }


        // Add any other special key processing here ...
        // else if (key == )

        
        // No special key found, pass it on through untouched

        else
                return( key );
}
int KeyCheckForQuit( void ) {

        int c;

        // When running interactively, check for a quit keypress.

        if ( INI.KeypressToRun == ON ) {

                c = SrGetKeyCheck( );

                c = KeyProcess( c );

                if ( c != EOF )
                        return( 1 );
                }

        return( 0 );
}








/* DISK FUNCTIONS:
 *
 * These functions handle writing binary data output files.
 *
 *
 */

char *TmpFileComputeFinalName( time_t *ReadyTime ) {

        char *suffix;
        static char FinalName[DIM_STR];

        // ReadyTime should be the time of the first point in the file
        // to be named.


        // Set file suffix based on the requested data format.

        if (INI.OutputFileFormat == OUTFILE)
                suffix = "OUT";

        else if (INI.OutputFileFormat == DATFILE)
                suffix = "DAT";

        else
                suffix = "TXT";



        // Compute the final file name using the requested naming mode.

        if ( INI.OutputFileNaming == SINGLE )

                sprintf( FinalName, "%sSIMP.%s", INI.OutputDirectory, suffix );


        else if ( INI.OutputFileNaming == SEQUENTIALDEC )

                sprintf( FinalName, "%s%08ld.%s", INI.OutputDirectory,
                                             (long)INI.OutputSeqStart, suffix );


        else if ( INI.OutputFileNaming == TIME )

                sprintf( FinalName, "%s%08lX.%s", INI.OutputDirectory,
                                                     (long)*ReadyTime, suffix );


        else if ( INI.OutputFileNaming == STDOUT )

                sprintf( FinalName, "STDOUT" );


        else  // INI.OutputFileNaming == NONE

                sprintf( FinalName, "" );


        return( FinalName );

}
FILE *TmpFileOpen( unsigned long nblock ) {

        FILE *fp;
        char *oflag;

        // If using standard out, return already open handle.

        if ( INI.OutputFileNaming == STDOUT )
                return( stdout );



        // Otherwise open the tmp file.

        // If this is the first block, it must be created when opening.

        if (nblock == 0)
                oflag = "w+b"; // was OFLAG | (O_CREAT|O_TRUNC);
        else
                oflag = "a+b"; // was OFLAG | O_APPEND;



        sprintf( FullFileName, "%sSIMP.TMP", INI.OutputDirectory );

        fp = fopen( FullFileName, oflag );

        if (!fp)
                ErrorWarning( "unable to open simp.tmp" );

        return( fp );

}
void TmpFileClose( FILE *fp ) {

        // If sending data to standard out, leave it open.

        if ( INI.OutputFileNaming == STDOUT )
                return;


        // Otherwise, close tmp file.

        fclose( fp );

}
void TmpFileRename( char *FinalName ) {


        // If sending data to standard out, there is nothing to rename.

        if ( INI.OutputFileNaming == STDOUT )
                return;



        // Verify that new name is available.

        if ( SrFileExist( FinalName ) ) {

                sprintf( msg, "Rename failed.  File %s already exists.",
                         FinalName );
                ErrorWarning( msg );
                }



        // Rename the .tmp file.

        sprintf( FullFileName, "%sSIMP.TMP", INI.OutputDirectory );
        
        if ( rename( FullFileName, FinalName ) == 0 ) {

                sprintf( msg, "SIMP.TMP renamed to %s ... \n", FinalName );
                DisplayInfo( msg );
                }

        else
                ErrorWarning( "simp.tmp rename failed" );


}
void SrOutWrite( FILE *fp, long *buffer, unsigned int nbytes ) {

        // Write the .OUT binary data to the .tmp file.

        if ( fwrite( buffer, 1, nbytes, fp ) != nbytes )

                ErrorWarning( "simp.tmp write failed" );

        if ( fflush( fp ) != 0 )

                ErrorWarning( "unable to commit simp.tmp to disk" );

}
void SaveToDisk(
                char *FinalFileName,

                long *ParXchDataBuffer,
                unsigned int ndata,

                PPSTDATA *GpsPpsBuffer,
                unsigned int npps,

                SERIALDATA *GpsSerialBuffer,
                unsigned int nserial,

                unsigned long nbuffer,
                time_t *AcquireTime,
                time_t *ReadyTime,
                int TimeSource,

                unsigned long *nblock,
                unsigned long *ntriggerfile,
                unsigned long *nblocktrig
               ) {

        FILE *fp;


        // Don't save anything if no file was requested.
        // But track buffer and file count for triggering.

        if ( (INI.OutputFileFormat == NOFILE) ||
             (INI.OutputFileNaming == NONE)   ) {

                sprintf( msg, "Buffer[%08lX] processed\n", nbuffer );
                DisplayInfo( msg );

                *nblock += 1;
                if (*nblock >= INI.OutputBlocks) {
                        *nblock = 0;
                        if (*ntriggerfile > 0)
                                *ntriggerfile = *ntriggerfile - 1;
                        }

                *nblocktrig = *nblock;
                return;
                }



        // Open tmp file

        fp = TmpFileOpen( *nblock );



        // Before first buffer, update and write the header (if .DAT
        // files requested).  The final file name is now computed just
        // after the previous file was written.

        if (*nblock == 0)
                HdrUpdateWrite( FinalFileName,
                                fp,
                                AcquireTime,
                                TimeSource
                              );



        // Save data in requested format.

        if (INI.OutputFileFormat == OUTFILE) {

                SrOutWrite( fp, ParXchDataBuffer, (unsigned int)INI.AcqBufSize );
                }


        else if (INI.OutputFileFormat == DATFILE) {

                if ( !SrDatRecordWrite( fp,
                                  SRDAT_TAGID_ANALOG24,
                                  ParXchDataBuffer,
                                  INI.AcqBufSize ) )
                        ErrorWarning( "Could not write analog data to .DAT" );

                if (npps > 0)
                        if ( !SrDatRecordWrite( fp,
                                          SRDAT_TAGID_GPSPPST,
                                          GpsPpsBuffer,
                                          npps * sizeof(PPSTDATA) ) )
                                ErrorWarning( "Could not write PPS(T) data to .DAT" );

                if (nserial > 0)
                        if ( !SrDatRecordWrite( fp,
                                          SRDAT_TAGID_GPSSERIAL,
                                          GpsSerialBuffer,
                                          nserial * sizeof(SERIALDATA) ) )
                                ErrorWarning( "Could not write serial data to .DAT" );

                if (*nblock >= INI.OutputBlocks-1) // nblock not incremented yet
                        if ( !SrDatRecordWrite( fp, SRDAT_TAGID_EOF, NULL, 0 ) )
                                ErrorWarning( "Could not write EOF to .DAT" );

                }




        // Close tmp file

        TmpFileClose( fp );


        sprintf( msg, "Buffer[%08lX] block %02lu -> simp.tmp\n",
                 nbuffer, *nblock );
        DisplayInfo( msg );



        // Increment the number of blocks processed.

        *nblock += 1;

        if (*nblock >= INI.OutputBlocks) {
                TmpFileRename( FinalFileName );
                INI.OutputSeqStart++;
                *nblock = 0;
                if (*ntriggerfile > 0)
                        *ntriggerfile = *ntriggerfile - 1;

                // Prepare the final filename for the NEXT file, passing in
                // the ReadyTime of the last buffer in the file just written
                // as an approximation of the start time of the first point
                // in the next file.  This is a change from versions of the
                // code before 6/30/03 that computed the file name using
                // the time at the END of the first buffer in the next file.

                FinalFileName = TmpFileComputeFinalName( ReadyTime );
                }

        *nblocktrig = *nblock;
}

void TriggerThresholdSetRange( long *PC_BUFFER, long ch,
                               long *ThMin, long *ThMax ) {

        double TMPoffset;
        long   DCoffset, j, k;

        // Find channel offset

        k         = ch;
        TMPoffset = 0;

        for ( j = 0 ; j < INI.ChBufPts ; j++ ) {
                TMPoffset += PC_BUFFER[k];
                k         += INI.Channels;
                }
        DCoffset = (long)(TMPoffset / (double)INI.ChBufPts);


        // Compute threshold range from offset

        *ThMin = DCoffset - INI.TriggerThreshold;
        *ThMax = DCoffset + INI.TriggerThreshold;
}

int TriggerCheckAnalog( long *PC_BUFFER, int Channel ) {

        int j, pt;
        long ThresholdMin, ThresholdMax;

        TriggerThresholdSetRange( PC_BUFFER, Channel,
                                  &ThresholdMin, &ThresholdMax );

        pt = Channel;
        for ( j = 0 ; j < INI.ChBufPts ; j++ ) {

            if ( (PC_BUFFER[pt] < ThresholdMin) ||
                 (PC_BUFFER[pt] > ThresholdMax) ) {

                    sprintf( msg, "Amplitude exceeds threshold\n" );
                    DisplayInfo( msg );

                    return( 1 );
                    }

                pt += INI.Channels;

            } // end for j

        return( 0 );
}

int TriggerCondition( long *PC_BUFFER, unsigned long *ntriggerfile,
                      unsigned long nblocktrig, long *triggerstate ) {

        int i;
        int SignalState, DigitalIoNibble, trig;


        // Continuous recording

        if ( INI.TriggerType == TRIG_NONE )
                return( 1 );


        // Amplitude threshold

        else if (INI.TriggerType == TRIG_THRESHOLD) {

            // Update in progress state based on number of files saved

              if ( (*triggerstate == TRIGSTATE_INPROGRESS) &&
                   (*ntriggerfile <= 0)                 )
                   *triggerstate = TRIGSTATE_ARMED;

            // Compute the amplitude threshold trigger condition

            for ( i = 0 ; i < INI.Channels ; i++ ) {

                if (INI.ChannelTrigger[i] == ON) {

                    if (i == DigitalIndex)
                        trig = 0;
                    else
                        trig = TriggerCheckAnalog( PC_BUFFER, i );


                    if (trig) {

                        // If already saving, don't count current partial file

                        if (*ntriggerfile > 0)
                            *ntriggerfile = INI.TriggerFilesToSave + 1;
                        else
                            *ntriggerfile = INI.TriggerFilesToSave;

                        *triggerstate = TRIGSTATE_INPROGRESS;
                        
                        } // end if trig
                        
                    } // end if ON

                } // end for i
            
             } // end TRIG_THRESHOLD



        // Digital I/O line signal

        else if (INI.TriggerType == TRIG_DIGLEVEL) {

                if (INI.TriggerDigitalLevelLine < 0)
                        return( 0 );

                // Update in progress state based on number of files saved

                  if ( (*triggerstate == TRIGSTATE_INPROGRESS) &&
                       (*ntriggerfile <= 0)                 )
                       *triggerstate = TRIGSTATE_ARMED;


                ParXchUserIoRd( ParXchHandle, &DigitalIoNibble );
                SignalState = (DigitalIoNibble >> INI.TriggerDigitalLevelLine) & 0x1;
                if ( SignalState == INI.TriggerActiveLevel ) {

                        sprintf( msg, "Digital level detected\n" );
                        DisplayInfo( msg );
                        
                        // If already saving, don't count
                        // current partial file

                        if (*ntriggerfile > 0)
                                *ntriggerfile = INI.TriggerFilesToSave + 1;
                        else
                                *ntriggerfile = INI.TriggerFilesToSave;

                        *triggerstate = TRIGSTATE_INPROGRESS;

                        } // end signalstate == activestate

                } // end TRIG_DIGLEVEL


        // Digital signal on EDGE triggered I/O line (rising or falling)

        else if (INI.TriggerType == TRIG_DIGEDGE) {

                // Immediate return if this feature not available

                if (INI.TriggerDigitalEdgeLine < 0)
                        return( 0 );

                // Update in progress state based on number of files saved

                if ( (*triggerstate == TRIGSTATE_INPROGRESS) &&
                     (*ntriggerfile <= 0)                 ) {
                     *triggerstate = TRIGSTATE_WAITING;
                        }

                
                // Test digital I/O line for matching the active trigger
                // state (ie rising edge is active when high, falling edge
                // is active when low)

                ParXchUserIoRd( ParXchHandle, &DigitalIoNibble );
                SignalState = (DigitalIoNibble >> INI.TriggerDigitalEdgeLine) & 0x1;

                // Now process the match results based on the current
                // trigger state ...

                
                // We are armed and looking for trigger.  If we find one,
                // setup the number of files to save.

                if (*triggerstate == TRIGSTATE_ARMED) {
                        if ( SignalState == INI.TriggerActiveEdge ) { // triggered
                                sprintf( msg, "Digital edge detected\n" );
                                DisplayInfo( msg );
                                *triggerstate = TRIGSTATE_INPROGRESS;
                                *ntriggerfile = INI.TriggerFilesToSave;
                                return( 1 ); 
                                }
                        else
                                return( 0 );
                        }


                // We are currently processing a trigger.  If we are
                // almost done, we check to see if we've been re-armed.
                // This lets us skip the waiting state so we're ready to
                // trigger again immediately after the current trigger
                // finishes.  In any case, we indicate the current
                // trigger continues.
                
                else if (*triggerstate == TRIGSTATE_INPROGRESS) {
                        if ((*ntriggerfile == 1)                  && // if last file +
                            (nblocktrig == INI.OutputBlocks-1)    && // last buffer +
                            (SignalState != INI.TriggerActiveEdge) ) { // re-armed then
                                *triggerstate = TRIGSTATE_ARMED;     // skip waiting
                                }
                        return( 1 );
                        
                        }

                
                // We are waiting to be re-armed by seeing the digital
                // line go to the inactive or non-triggered state.  With
                // rising edge triggering, a high value causes a trigger
                // and a low value re-arms.

                else { // (*triggerstate == TRIGSTATE_WAITING)
                        if (SignalState != INI.TriggerActiveEdge) { // re-armed
                                *triggerstate = TRIGSTATE_ARMED;
                                }
                        
                        return( 0 );     // Never return a trigger when waiting
                        }

                } // end if TRIG_DIGEDGE


        // Unknown

        else {
                ErrorWarning( "Unknown trigger type" );
                return( 0 );
                }




        // A positive trigger file count may come from either
        //   1) An amplitude exceeding the threshold in the current buffer
        //   2) More files still remaining to be written after a
        //      previous high amplitude was found

        if (*ntriggerfile > 0)
                return( 1 );
        else
                return( 0 );
}




/* SCREEN DATA DISPLAY:
 *
 * Print out acquired data values to the screen.
 *
 *
 *
 *
 */

void DisplayValues( long *values, unsigned int n ) {

        unsigned int i;
        int          j, MaxChan;


        // Decide how many channels to display

        MaxChan = INI.ParXchAnalogChannels;             // all the analog ones
        if (INI.DigitalEnable == ON)
                MaxChan += INI.ParXchDigitalChannels;   // add the digital one


        // Print out the values, using one hex digit for the digital channel
        
        for ( i = 0 ; i < n ; i += INI.Channels ) { // inc by all channels
            for ( j = 0 ; j < MaxChan; j++ ) {      // only show chosen ones
                if (j == DigitalIndex)
                    printf( " %1X", values[i+j] );
                else
                    printf( " %08lX", values[i+j] );
                }
            if (INI.GpsEnable == ON && values[i+GpsMarkIndex])
                    printf( " <PPS\n" );
            else
                    printf( "\n" );
            }
}

void DisplayPpsValues( PPSTDATA *values, unsigned int n ) {

#ifdef GPS_AVAILABLE

        unsigned int i;

        for ( i = 0 ; i < n ; i ++ ) {

                printf( "PPS: Event=%u   Count@PPS=%08lX%08lX  Count@Dready=%08lX%08lX\n",

                          values[i].PpsEventNum,
                          values[i].CountAtPps.u.HighPart,
                          values[i].CountAtPps.u.LowPart,
                          values[i].CountAtDready.u.HighPart,
                          values[i].CountAtDready.u.LowPart
                      );
                printf( "           %u  PcTime@PPS=%08lX%08lX PcTime@Dready=%08lX%08lX\n",

                          values[i].PpsEventNum,
                          values[i].PctimeAtPps.u.HighPart,
                          values[i].PctimeAtPps.u.LowPart,
                          values[i].PctimeAtDready.u.HighPart,
                          values[i].PctimeAtDready.u.LowPart
                      );
                }
#else
        if (n > 0)
                printf( "%u PPS data can not be displayed\n", n );
#endif

}

void DisplaySerialValues( SERIALDATA *values, unsigned int n ) {

#ifdef GPS_AVAILABLE

        unsigned int i, j, jstart, jmax;

        for ( i = 0 ; i < n ; i ++ ) {

                printf( "Serial: PpsEvent=%u, NMEA string count = %d\n",
                          values[i].PpsEventNum,
                          values[i].NmeaCount
                      );

                jstart = 0;
                jmax   = values[i].NmeaCount;
                for ( j = 0 ; j < jmax ; j++ ) {
                        printf( "%s", &values[i].NmeaMsg[jstart] );
                        jstart += MAX_NMEA_SIZE;
                        }
                }
        if (n > 0)  printf( "\n" );

#else
        if (n > 0)
                printf( "%u SERIAL data can not be displayed\n\n", n );
#endif

}










/* BUFFER ALLOCATION AND CHECK ERROR HELPER FUNCTIONS:
 *
 * These helpers allow the primary program flow to read more cleanly.
 *
 */

void AllocateBuffers(

                long **ParXchDataBuffer,
                long **ParXchDataBufferEnd,
                PPSTDATA **GpsPpsBuffer,
                PPSTDATA **GpsPpsBufferEnd,
                SERIALDATA **GpsSerialBuffer,
                SERIALDATA **GpsSerialBufferEnd

               ) {



        /* Allocate PARxCH and PARGPS buffer memory:
         *
         * The ParXchDataBuffer array in PC memory is used to store analog
         * data on its way out from the PARxCH to the disk and screen.
         *
         * The GpsPpsBuffer and GpsSerialBuffer arrays are for GPS timing
         * data.
         *
         */


        // Allocate PARxCH data buffer.

        *ParXchDataBuffer    = (long *)malloc( INI.AcqBufSize );
        *ParXchDataBufferEnd = *ParXchDataBuffer + INI.Nvalues;

        if ( !*ParXchDataBuffer ) {
                sprintf( msg, "unable to malloc %s data buffer", INI.ParXchModelName );
                ErrorFatal( msg );
                }


        // Allocate GPS PPS and serial buffers, if requested.

        if ( INI.GpsEnable == ON ) {

                *GpsPpsBuffer       = (PPSTDATA *)malloc( INI.GpsBufSize );
                *GpsPpsBufferEnd    = *GpsPpsBuffer + INI.NgpsValues;

                *GpsSerialBuffer    = (SERIALDATA *)malloc( INI.SerBufSize );
                *GpsSerialBufferEnd = *GpsSerialBuffer + INI.NgpsValues;

                if ( !*GpsPpsBuffer || !*GpsSerialBuffer )
                        ErrorFatal( "unable to malloc GPS buffers" );
                }

        else {
                *GpsPpsBuffer       = NULL;
                *GpsPpsBufferEnd    = NULL;
                *GpsSerialBuffer    = NULL;
                *GpsSerialBufferEnd = NULL;
                }
}

int CheckErrorReadData( int ErrorCode, int n ) {

        // Trap abnormal overflow where no points are coming in
        // so we avoid getting stuck in an infinite polling loop

        if (ErrorCode == PARXCH_ERROR_OVERFLOW && n <= 0) {
                sprintf( msg, "ReadData error %s", PARXCH_ERROR_MSG[ErrorCode]);
                ErrorFatal( msg );
                return( 0 );
                }


        // Check for an error, but ignore normal overflow which
        // will be handled elsewhere.

        if ( (ErrorCode != PARXCH_ERROR_NONE    ) &&
             (ErrorCode != PARXCH_ERROR_OVERFLOW) ) {

                sprintf( msg, "ReadData error %s", PARXCH_ERROR_MSG[ErrorCode]);
                ErrorWarning( msg );
                return( 0 );
                }

        return( 1 );
}

#ifdef GPS_AVAILABLE

void CheckErrorReadPps( int ErrorCode ) {

        if ( ErrorCode != PARGPS_ERROR_NONE ) {

                sprintf( msg, "ReadGpsPps error %s", PARGPS_ERROR_MSG[ErrorCode] );
                ErrorFatal( msg );
                }

}

void CheckErrorReadSerial( int ErrorCode ) {

        if ( ErrorCode != PARGPS_ERROR_NONE ) {

                sprintf(msg,"ReadGpsSerial error %s",PARGPS_ERROR_MSG[ErrorCode]);
                ErrorFatal( msg );
                }
}

#endif // GPS_AVAILABLE







/* PARxCH AND PARGPS OPERATION FUNCTIONS:
 *
 * These functions handle various basic operations for the PARxCH and
 * PARGPS such as opening and closing the drivers, plus initializing,
 * starting, and stopping the hardware.
 *
 *
 *
 */

void OpenParXchAndParGps( void ) {

        int  ErrorCode, IrqNum;
        long ADS1210_CrValue, cftemphi, cftemplo, cftemp2hi, cftemp2lo;

#define INIT_SKIPPED          "%s open skipped, AtodEnable = OFF\n\n"
#define INIT_SUCCESS          "Success\n"
#define INIT_FAILURE          "Failure\n"
#define INIT_XCH_FAILURE      "%s open failed (Err=%s)"
#define INIT_GPS_FAILURE      "PARGPS open failed (Err=%s)"
#define INIT_GPS_FAILURE_SER  "PARGPS open failed (Err=%s)\n\t\tIs serial port %d already in use?"
#define INIT_ATTACH_FAILURE   "%s attach to PARGPS failed (Err=%s)"


        // Initialize to bad values

        ParXchHandle = BAD_DEVHANDLE;
        GpsHandle    = BAD_DEVHANDLE;


        // Skip opening the PARxCH and PARGPS if the user has indicated
        // the A/D is not available.

        if ( INI.AtodEnable == OFF ) {
                sprintf( msg, INIT_SKIPPED, INI.ParXchModelName );
                DisplayInfo( msg );
                DisplayInfo( "\n" );
                return;
                }



        // Open the PARxCH A/D driver and initialize the board

        DisplayInfoFlush( "Opening the PARxCH driver ...       " );

        ParXchHandle = ParXchFullOpen(

                        INI.AtodDriverName,
                        INI.AtodModel,
                        INI.PortMode,

                        INI.DataCoding,
                        INI.GainLog,
                        INI.TurboLog,
                        INI.Decimation,
                        INI.ExtraDecimation,
                        PARXCH_UNUSED,

                        &ADS1210_CrValue,
                        &ErrorCode
                        );

        if ( ParXchHandle == BAD_DEVHANDLE ) {
                DisplayInfo( INIT_FAILURE );
                sprintf( msg, INIT_XCH_FAILURE,
                         INI.ParXchModelName, PARXCH_ERROR_MSG[ErrorCode] );
                ErrorFatal( msg );
                }
        DisplayInfo( INIT_SUCCESS );






        // Not using GPS

        if ( INI.GpsEnable == OFF )
                return;


        // Fill GPS counter frequencies with dummy value of 0

        cftemphi  = cftemplo  = 0L;     // quiet compiler complaints
        cftemp2hi = cftemp2lo = 0L;

        DatHdr.CounterFreqHi  = cftemphi;
        DatHdr.CounterFreq    = cftemplo;
        DatHdr.Counter2FreqHi = cftemp2hi;
        DatHdr.Counter2Freq   = cftemp2lo;

        IrqNum = 0;

#ifdef GPS_AVAILABLE

        // Open the PARGPS driver and initialize the board

        DisplayInfoFlush( "Opening the PARGPS driver ...       " );

        GpsHandle = ParGpsFullOpen(
                           INI.GpsDriverName,
                           INI.GpsModel,
                           INI.GpsSerialPort,
                           NULL,
                           &ErrorCode
                           );

        if ( GpsHandle == BAD_DEVHANDLE ) {
                DisplayInfo( INIT_FAILURE );
                if (ErrorCode == PARGPS_ERROR_SERIAL_PORT_NOT_OPEN)
                        sprintf( msg, INIT_GPS_FAILURE_SER, PARGPS_ERROR_MSG[ErrorCode],
                                 INI.GpsSerialPort );
                else
                        sprintf( msg, INIT_GPS_FAILURE, PARGPS_ERROR_MSG[ErrorCode] );
                ErrorFatal( msg );
                }
        DisplayInfo( INIT_SUCCESS );



        // Attach the PARGPS driver to the PARxCH driver.

        DisplayInfoFlush( "Attaching the PARxCH and PARGPS ... " );
        if ( !ParXchAttachGps( ParXchHandle, GpsHandle, &ErrorCode ) ) {
                DisplayInfo( INIT_FAILURE );
                sprintf( msg, INIT_ATTACH_FAILURE,
                         INI.ParXchModelName, PARXCH_ERROR_MSG[ErrorCode] );
                ErrorFatal( msg );
                }
        DisplayInfo( INIT_SUCCESS );


        // Check that an interrupt number is actually assigned to the driver.
        // If not, give a warning since the user MAY really want to do this.
        
        DisplayInfoFlush( "Checking PARxCH interrupt num ...   " );
        IrqNum = ParXchInterruptGetNumber( ParXchHandle );
        if ( IrqNum == 0 ) {
                DisplayInfo( "Warning - No IRQ assigned\n" );
                ErrorWarning( "Reinstall PARxCH driver to assign IRQ" );
                }
        else {
                sprintf( msg, "Success IRQ = %d\n", IrqNum );
                DisplayInfo( msg );
                }


   
        // Set GPS counter frequency in the Dat Hdr.

        if ( (INI.OutputFileFormat == DATFILE) &&
             (INI.GpsEnable == ON)             ) {
                if ( ParGpsGetFullCounterFrequency( GpsHandle,
                                                    &cftemphi, &cftemplo,
                                                    &cftemp2hi, &cftemp2lo) ) {

                        DatHdr.CounterFreqHi  = cftemphi;
                        DatHdr.CounterFreq    = cftemplo;
                        DatHdr.Counter2FreqHi = cftemp2hi;
                        DatHdr.Counter2Freq   = cftemp2lo;

                        // Warn if GPS counters are 0.  This should never
                        // happen, but it might under linux if the counter
                        // is not supported.

                        if ( cftemplo == 0L && cftemphi == 0L )
                                ErrorWarning( "Gps Counter Frequency is 0" );
                        if ( cftemp2lo == 0L && cftemp2hi == 0L )
                                ErrorWarning( "Gps Counter 2 Frequency is 0" );
                        }
                else
                        ErrorWarning( "Can't get Gps CounterFrequency value" );
                }

#endif // GPS_AVAILABLE

}


void StartParXchAndParGps( void ) {

        // Perform an orderly startup of the hardware.

        // Don't start anything if the A/D is not available

        if ( INI.AtodEnable == OFF )
                return;


        LogMessage( "Starting acquisition." );
        
#ifdef GPS_AVAILABLE

        // Start the PARGPS if requested.

        if ( INI.GpsEnable == ON )
                ParGpsStart( GpsHandle );
#endif


        // Start the PARxCH A/D acquiring data.

        ParXchStart( ParXchHandle );
}


void ShutdownParXchAndParGps( void ) {

        // Perform an orderly shutdown.

        LogMessage( "Stopping acquisition." );


        // Stop the PARxCH from acquiring data.

        ParXchStop( ParXchHandle );


#ifdef GPS_AVAILABLE

        // If the PARGPS was requested, stop it, release it from the
        // PARxCH, and close the PARGPS driver.

        if ( INI.GpsEnable == ON ) {
                ParGpsStop( GpsHandle );
                ParXchReleaseGps( ParXchHandle );
                ParGpsClose( GpsHandle );
                }

#endif

        // Close the PARxCH A/D driver.

        ParXchClose( ParXchHandle );

}

void OverflowRecovery( void ) {

        char FinalName[DIM_STR];
        FILE *fp;


        // After an overflow has been detected, the PARxCH must be stopped
        // and restarted.  Some users might want to read out the valid data
        // data currently saved in the fifo first.  If so, put that code
        // here ...


        // Perform orderly shutdown and restart after overflow.


        // Shutdown

        ShutdownParXchAndParGps( );


        // Data is collected in a temporary file until the full
        // requested amount has been acquired.  When an error occurs,
        // add an end of file record if needed, compute a new name
        // ending in .FIX, and rename the partial file of data collected
        // up to the time of the error.  Then we increment the
        // sequential start number so scope will leave a gap in the
        // normal .DAT file numbers at time of the error.

        if ( INI.OutputFileFormat == DATFILE ) {
                fp = TmpFileOpen( 1 );
                SrDatRecordWrite( fp, SRDAT_TAGID_EOF, NULL, 0 );
                TmpFileClose( fp );
                }

        sprintf( FinalName, "%s%08ld.FIX", INI.OutputDirectory, INI.OutputSeqStart );
        TmpFileRename( FinalName );
        INI.OutputSeqStart++;


        // Check for a 'q' to quit

        if ( KeyCheckForQuit( ) )
                CleanExit( 0 );  // shuts down PARxCH + PARGPS


        // Restart

         OpenParXchAndParGps( );
        StartParXchAndParGps( );
}

void CleanExit( int value ) {

        // Perform orderly shutdown and exit.

        ShutdownParXchAndParGps( );

        LogMessage( "Ending simp." );

        exit( value );

}







/* MAIN:
 *
 *
 *
 *
 */

int main( int argc, char *argv[] ) {

        int LedValue, ErrorCode, ReadRequest, IniArg, Ok;
        long nbuffer;
        time_t StartTime, AcquireTime, ReadyTime;
        double ElapsedBufferTime;
        unsigned int n, ndata, npps, nserial;
        unsigned long nblock, ntriggerfile, nblocktrig, triggerstate;
        long *ParXchDataBuffer, *ParXchDataBufferEnd, *ParXchDataBufferPtr;
        PPSTDATA *GpsPpsBuffer, *GpsPpsBufferEnd, *GpsPpsBufferPtr;
        SERIALDATA *GpsSerialBuffer, *GpsSerialBufferEnd, *GpsSerialBufferPtr;
        char *FinalFileName;



        // Macros computing the number of remaining unfilled buffer points
        // and resetting the current pointers to the buffer beginning

#define NREMAIN_PARXCHDATA (ParXchDataBufferEnd-ParXchDataBufferPtr)
#define NREMAIN_GPSPPS     (GpsPpsBufferEnd-GpsPpsBufferPtr)
#define NREMAIN_GPSSERIAL  (GpsSerialBufferEnd-GpsSerialBufferPtr)

#define RESET_PARXCHDATA (ParXchDataBufferPtr = ParXchDataBuffer)
#define RESET_GPSPPS     (GpsPpsBufferPtr = GpsPpsBuffer)
#define RESET_GPSSERIAL  (GpsSerialBufferPtr = GpsSerialBuffer)




        // Allow getchar to return even if no character is waiting.

        if ( SrKeyboardNonBlock() == -1 )

                ErrorFatal( "Can't set non-blocking read ..." );



        // Get command line arguments

        IniArg = ProcessCommandLine( argc, argv );

        



        // Initialize INI parameters ...
        //
        //    Setup defaults
        //    Parse the user's .ini file
        //    Fill in additional computed values
        //    Generate the (ini) .rpt file
        //
        //

        IniDefaults();
        IniParse( argv[IniArg] );
        Rev();  // Announce program version as soon as INI.InfoDisplay is known.
        IniCompute();
        IniReport();



        





        // Initialize HDR parameters ...
        //
        //    Setup defaults
        //    Write .HDR file (if .OUT files were requested)
        //
        //

        HdrDefaults();
        HdrWriteFile();




        // Display the starting message

        StartMsg();





        // Open and initialize the PARxCH and PARGPS.

        OpenParXchAndParGps();





        // Allocate PC acquisition and GPS buffer memory.

        AllocateBuffers(
                        &ParXchDataBuffer,
                        &ParXchDataBufferEnd,
                        &GpsPpsBuffer,
                        &GpsPpsBufferEnd,
                        &GpsSerialBuffer,
                        &GpsSerialBufferEnd
                       );






        /* Start executing:
         *
         * The PARxCH executes continuously, saving data in its FIFO.
         * This FIFO is drained when it is partly full, so acquisition
         * continues uninterrupted while the PC saves to disk.
         *
         *
         */

        if ( INI.KeypressToRun == ON ) {
                printf( "\nPress ENTER to start, then ENTER again to quit ... \n" );
                KeyWaitForRun( );
                }
        else
                DisplayInfo( "\n" );



        // Start acquisition.

        StartParXchAndParGps( );


acq_started:


        // Get approximate start time from the PC and set first file name

        time( &StartTime );
        AcquireTime       = StartTime;
        ReadyTime         = StartTime;
        ElapsedBufferTime = 0.0;
        NpointsFromStart  = 0;
        FinalFileName     = TmpFileComputeFinalName( &ReadyTime );


        // Initialize data buffer count, trigger file count and buffer pointers.

        nblock       = 0;
        nblocktrig   = 0;
        nbuffer      = 0;
        ntriggerfile = 0;
        triggerstate = TRIGSTATE_WAITING;

        RESET_PARXCHDATA;
        RESET_GPSPPS;
        RESET_GPSSERIAL;


        // Indicate what to read

        ReadRequest = PARXCH_REQUEST_ANALOG;
        if (INI.DigitalEnable)
                ReadRequest |= PARXCH_REQUEST_DIGITAL;
        if (INI.GpsEnable)
                ReadRequest |= PARXCH_REQUEST_GPS;
        if (INI.CounterEnable)
                ReadRequest |= PARXCH_REQUEST_COUNTER;




        // Initialize the PARxCH Yellow LED to off.

        ParXchUserLed( ParXchHandle, LedValue = 0 );





        // Poll the PARxCH for available data.
        //
        // This loop repeats until a complete buffer worth of analog
        // data has been read from the PARxCH.  Then the buffer is written
        // to disk and the loop started again.
        //
        //

        sprintf(msg,"\nWaiting %4.2f sec for first buffer ...\n",INI.BufferTime);
        DisplayInfo( msg );


polling_loop:

        // Read data that is ready

        if ( INI.AtodEnable == ON ) {

                n = ParXchFullReadData(
                                       ParXchHandle,
                                       ParXchDataBufferPtr,
                                       NREMAIN_PARXCHDATA,
                                       ReadRequest,
                                       &ErrorCode
                                      );
//WCT - Consider adding code like this to indicate not keeping up with acquisition
//                if ( n == INI.Nvalues ) {
//                        sprintf( msg, "Read ALL data at once, so NOT keeping up with acquisition!");
//                        ErrorWarning( msg );
//                        }
                ParXchDataBufferPtr += n;
                Ok = CheckErrorReadData( ErrorCode, n );
                if ( !Ok ) {
                        OverflowRecovery();
                        goto acq_started;
                        }


                
#ifdef GPS_AVAILABLE

                // Read GPS PPS and serial data, if requested

                if (INI.GpsEnable == ON ) {

                        n = ParGpsReadPpsData(
                                        GpsHandle,
                                        GpsPpsBufferPtr,
                                        NREMAIN_GPSPPS,
                                        &ErrorCode
                                        );

                        GpsPpsBufferPtr += n;
                        CheckErrorReadPps( ErrorCode );


                        n = ParGpsReadSerialData(
                                        GpsHandle,
                                        GpsSerialBufferPtr,
                                        NREMAIN_GPSSERIAL,
                                        &ErrorCode
                                        );

                        GpsSerialBufferPtr += n;
                        CheckErrorReadSerial( ErrorCode );
                        }
#endif



                // If more data is needed to fill the PARxCH data buffer,
                // wait a while and loop around to read again.

                if ( NREMAIN_PARXCHDATA > 0 ) {

                        if ( KeyCheckForQuit( ) )
                                CleanExit( 0 );  // shuts down PARxCH + PARGPS

                        SrSleep( INI.SleepMs );
                        goto polling_loop;
                        }


                // Else current PARxCH data buffer is complete so
                // save number read and reset to buffer beginning.
                // (Code continues on below to process the data.)

                else { // NREMAIN_PARXCHDATA == 0

                        ndata   = ParXchDataBufferPtr - ParXchDataBuffer;
                        npps    =     GpsPpsBufferPtr -     GpsPpsBuffer;
                        nserial =  GpsSerialBufferPtr -  GpsSerialBuffer;

                        RESET_PARXCHDATA;
                        RESET_GPSPPS;
                        RESET_GPSSERIAL;
                        }

                } // end if AtodEnable






        // At this point, we have a complete buffer of analog data.  So
        // it is time to do what ever processing we want.  For example,
        // saving the data to a disk file.


        // Twiddle the yellow LED.

        if ( INI.ToggleLed == ON )

                ParXchUserLed( ParXchHandle, LedValue ^= 1 );



        // Save the time when this complete buffer was ready
        // (which is approximately the start of the NEXT buffer).

        time( &ReadyTime );



        // If trigger condition met

        if ( TriggerCondition( ParXchDataBuffer, &ntriggerfile,
                               nblocktrig, &triggerstate ) ) {

                // Save the buffers to disk

                if ( INI.TriggerActionSave )
                        SaveToDisk( FinalFileName,
                                    ParXchDataBuffer, ndata,
                                    GpsPpsBuffer,     npps,
                                    GpsSerialBuffer,  nserial,
                                    nbuffer, &AcquireTime, &ReadyTime,
                                    SRDAT_SOURCE_ACQ, &nblock, 
                                    &ntriggerfile, &nblocktrig );
                
                // Otherwise just count down files that would have been
                // written so trigger will finish as usual

                else {
                        nblocktrig++;
                        if (nblocktrig >= INI.OutputBlocks) {
                                nblocktrig = 0;
                                if (ntriggerfile > 0)
                                        ntriggerfile = ntriggerfile - 1;
                                }
                        sprintf( msg, "Buffer[%08lX] discarded\n", nbuffer );
                        DisplayInfo( msg );
                        }


                 // Turn on requested digital output bit

                if ( INI.TriggerType != TRIG_NONE  &&  INI.TriggerActionSignal )
                        ParXchUserIoWr( ParXchHandle,
                                        (0x01 << INI.TriggerLineToSignal) );
                }


        // Otherwise, trigger is done

        else {
                 // Turn off all digital output bits

                if ( INI.TriggerType != TRIG_NONE  && // triggering
                     INI.TriggerActionSignal       && // digital output action
                    !INI.TriggerLatchLine          )  // not latching
                        ParXchUserIoWr( ParXchHandle, 0x00 );
                sprintf( msg, "Buffer[%08lX] discarded\n", nbuffer );
                DisplayInfo( msg );
                nblocktrig = 0;
                }





        // Display the acquired data.

        if ( INI.DisplayData == ON )
                DisplayValues( ParXchDataBuffer, ndata );

        if ( INI.GpsDisplayPps == ON )
                DisplayPpsValues( GpsPpsBuffer, npps );

        if ( INI.GpsDisplaySerial == ON )
                DisplaySerialValues( GpsSerialBuffer, nserial );



        // Increment the number of buffers and points processed.
        // Roughly update the next buffer time.  For a more accurate
        // value, use the GPS data to compute time.

        nbuffer++;
        NpointsFromStart  += INI.ChBufPts;
        ElapsedBufferTime += INI.BufferTime;
        AcquireTime        = StartTime + (unsigned int)ElapsedBufferTime;




        // Has the PARxCH DRAM FIFO overflowed ?

        if ( (INI.AtodEnable == ON) && ParXchOverflow( ParXchHandle ) ) {

                ErrorWarning( "FIFO overflow during PC processing ..." );

                // If error was not fatal, reset and continue

                OverflowRecovery(  );
                goto acq_started;
                }



        // Check for a 'q' to quit, and then continue polling.

        if ( KeyCheckForQuit( ) )
                CleanExit( 0 );  // shuts down PARxCH + PARGPS

    goto polling_loop;


    return( 0 );
}
