/* FILE: parxchkd.c             Copyright (c), Symmetric Research, 2003-2010

ParXch device driver functions for all OS's.

For Windows 2K/XP and Linux, the device driver must first be installed
by running the provided indriver utility.

For Windows 95/98 and DOS, the device driver functions are made
available by linking them in with the parxch.c library functions at
compile time.

Users then interact with the driver functions by calling the parxch.c
library functions.


DRIVER ORGANIZATION OVERVIEW

The following information is provided as an overview for those wishing
to understand how this driver code is organized.  A thorough
understanding of device driver programming is recommended for anyone
wanting to modify this code.  The following resources may help:

   Linux Device Drivers by Jonathan Corbet, Alessandro Rubini & Greg Kroah-Hartman
   Writing NT Device Drivers by Tony Mason & Peter Giscarola
   WDM Device Drivers by Walter Oney


CONDITIONAL COMPILE TIME DEFINES

Although there is a great deal of similarity from one OS to the next
in how device drivers work, they are not identical.  This means that
some OS specific code is required.  However, we have tried to share
and reuse as much common code as possible since this eases maintenance
and thus increases reliablity.  OS specific sections of code are
selected at compile time by use of #if defined statements that check
the SROS_xxx group of defines where xxx indicates one of the supported
OS's.  Currently these are: SROS_WINXP, SROS_WIN2K, SROS_LINUX,
SROS_WIN9X, SROS_MSDOS

There is a little bit of assembly code in the demux functions.  This
makes porting to other architectures like ARM chips difficult.  To use
C code only, define USE_C_CODE_DEMUX at compile time.


DEVINFO DATA STRUCTURE

The most important data structure in this driver is the device
information structure (DEVINFO).  It contains all of the driver's data
and is passed to almost all of the functions.  To make it easier to
support multiple OS's, these functions have been divided into three
main groups described below.


FUNCTION GROUPS

The first group contains functions that interact with the PARxCH hardware
and contain the fundamental behavior of the driver common to all the
OS's.  Because they are so important, they have been designated as group
1 have been given names that start with ParXch_.  For example ParXch_DrvOpen
and ParXch_DramFifoReset.  These common functions are sandwiched between
two layers of OS specific functions.

The second group contains OS specific helper functions which are used to
insulate the common PARxCH functions from differences between the OS's.
They have been given names that start with Os.  For example, OsPortIn or
OsSetIoctlBuffers.  There is one complete set of these helper functions
for each supported OS.  Although all versions are included in this
source file, the conditional compile time SROS_xxx defines are used to
select which set is actually used.

The third group contains top level functions that are specific to an
OS.  They are typically called from the OS itself and have been given
names prefixed with the OS.  For example, Wdm_PnpStartDevice or
Linux_DrvOpen.


CALLING THE DRIVER

The functions in this driver are typically called in one of two ways;
directly from the OS or from the OS as a result of call from user code.

When the indriver installation utility is run, the OS calls some of
the OS specific top level functions to initialize the driver upon
installation.

Once the device driver has been installed, user code can use the
provided parxch.c library to initiate calls into this driver.  Typically
these calls will be used to do things like open, start, and read data
from the driver.  Many of these call from user code to the PARxCH
library and on to the OS and one of the top level OS specific
functions before reaching the shared ParXch_DrvIoctl function which deals
with any data transfer before calling an appropriate ParXch_ function to
actually service the request.  Although this sounds complicated, it is
all relatively straightforward.


WORKING WITH THE PARGPS

To work with the PARGPS, the PARxCH A/D driver must first be compiled
with PARGPS support.  Then, after both drivers have been installed, the
PARxCH driver must be "attached" to the PARGPS driver using a call
from the parxch.c library.  This results in both drivers having a
pointer to a common kernel data area which they share and use to store
pointers to selected functions which the other driver might need to
call.  Finally, interrupts must be enabled with ParXch_InterruptEnable.
This is done automatically in the ParXch_Start function if the PARGPS
driver has been attached.


HARDWARE BEHAVIOR

The PARxCH has several logical resources down on the board.  These
include the A/D's, the digital I/O, the FIFO, and the board control
register.

The board control register is used to control the yellow user led,
reset the FIFO, and the "mode" in which the PARxCH operates.

In the HALT_INIT mode, the PC controls the sclk and sdi pins on the
A/D's and can use them to clock setup parameters like the sampling rate
into the A/D's.  In this mode, the PC can both write to and read from
the FIFO.  This is useful for testing the FIFO.  This is the only mode
in which the PC can write to the FIFO.  In all other modes, the PC can
only read the FIFO which is filled by the Xilinx CPLD.  The digital I/O
is not accessable in this mode.

In the HALT_UDIO mode, the PC can read and write from the digital I/O.

In RUN_EF mode, the Xilinx CPLD starts running and moves data from the
A/D's to the FIFO.  It also passes digital I/O signals through.  If
parallel port interrupts have been enabled, a signal on digital IN 3
gets passed on as a PC parallel port interrupt.

When board control register DRDYIRQREQ is set, the Xilinx CPLD will
generate a PC parallel port interrupt on the next data ready (DREADY)
signal from the A/D's.


INTERRUPT PROCESSING

Currently, the PARxCH driver does nothing for itself in its interrupt
service routine.  Instead, if the PARGPS driver has been attached, it
calls the PARGPS interrupt service routine instead.  See the PARGPS
driver file comments for more info.

SUMMARY

The organization of the PARxCH device driver has been discussed.
Hopefully, this will help anyone who needs to understand the workings
of this code and that further questions can be answered by studying
the code itself.  If you still have questions, please send email to
info@symres.com and we will be happy to discuss them with you.


Change History:
   2010/10/05  WCT  Support Fedora13 and add USE_C_CODE_DEMUX to ease ARM ports
   2009/07/25  WCT  Correct Rbyte -> ReadbackCrValues code so it works in DOS too
   2008/05/25  WCT  Support Fedora9 by removing int irq from Linux_ParPortIrq
   2007/08/15  WCT  Add Linux support for Fedora7 by changing irq func args
   2007/07/25  WCT  PWRITE+PDIR consistent for EPP, reorder ADS1210 Init bytes
   2007/02/12  WCT  Add Linux support for FC6 by removing config.h and devfs
   2007/01/08  WCT  Change to SrParXch0 style driver names
   2006/09/15  WCT  Make Linux version layer itself over the parport driver
   2005/11/25  WCT  Add Earthworm and GPSMODEL improvements
   2005/08/20  WCT  Make 8CH GpsCounter work for DOS and with decimation
   2005/07/25  WCT  Fix Wdm_DrvWmi stub
   2005/03/20  WCT  Support 47 ohm fix by masking data with fifo size
   2004/02/01  WCT  Combine 1CH, 4CH and 8CH into a single xCH driver
*/





/* DEBUGGING DEFINE:

This define controls the level of printf type debugging that will come
from the kernel mode driver.  -1 = no printf debugging statements.
SRDBG = (1,2,3,4,5) -> various levels of reporting.  Production
compiles are done with -1.

Examples of functions turned on at each debug level are:

  0 DrverEntry, Wdm_DrvAddDevice, Wdm_WorkItemNotifyCallback, AttachGps
  1 DrvOpen, DrvClose, DrvRead
  2 Start, Stop, Wdm_ParPortAcquire
  3 ParPortInit, Wait, Init
  4 BdCtrlXxx, ADS1210_Xxx, DramFifoXxx, PowerUp/Down
  5 ParPortXxxRd/Wr, OsSetIoctlxxx

*/

#define SRDBG -1





/* PARXCH TEST VOLTAGE DEFINES:

These defines control whether the voltage good signal is tested or
not.  Setting the _BAD define to 1 tests the voltage at the end of
the PowerDown function when the voltage should be off.  Setting the
_GOOD define to 1 tests the voltage at the end of the PowerUp
function when the voltage should be good.  Setting either define to 0
ignores the respective voltage test.  Production compiles are done
with both defines set to 0.

If you enable either define, be sure to also set the _VOLTAGE_IS_GOOD
define described below appropriately for the Rev of your PARXCH board.

Some PC's don't provide a pull-up resistor on the voltage good line.
This resistor is needed to pull the active low voltage good signal up
to a 1 when power on the PARXCH board is off.  Once the PARXCH board
has power, it can reliably pull the signal low to indicate good
voltage.  Because this type of PC will always see 0, they should set
the _BAD define to 0.

If your PC is among the majority that do provide a pull-up resistor
on the voltage good line, you can get the additional checking by
setting both defines to 1.

Most PARXCH Rev B boards did not support the voltage good signal, so
they should use 0 for both defines.

*/

#define PARXCH_TEST_VOLTAGE_BAD   0
#define PARXCH_TEST_VOLTAGE_GOOD  0



/* PARXCH VOLTAGE IS GOOD DEFINE:

Rev C and Rev D boards differ in the value they return to mean that the
voltage is good.  For Rev C, a low value means the voltage is good.
While for Rev D, a high value means the voltage is good.

Only ONE of the following two defines should be used:
    For Rev C   #define PARXCH_VOLTAGE_IS_GOOD    0
    For Rev D   #define PARXCH_VOLTAGE_IS_GOOD    1

*/

#define PARXCH_VOLTAGE_IS_GOOD    1  // For Rev D



/* PARXCH REV B TEST OVERFLOW DEFINE:

This define controls whether the overflow signal is tested or not.
Set it to 1 to test the overflow signal and 0 to ignore it.

Production compiles are done with this define set to 1.  Most PARXCH
Rev B boards did not support the overflow signal, so they should use 0.

*/

#define PARXCH_TEST_OVERFLOW   1





/* LOCAL DEFINE:

This define controls whether the driver functions are visible to
other source files.  Most users should let SRLOCAL be defined as
static to avoid namespace clutter and possible function name
contention.  The only exceptions would be for specialized debugging
and diagnostic programs.

*/

#if !defined( SRLOCAL )
#define SRLOCAL static
#endif



/* OS dependent includes, defines, and typedefs.
 *
 *
 *
 */

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
#include <wdm.h>
#include <stdio.h>
#include <parallel.h>  // for ParClass and ParPort driver functions
#include <initguid.h>  // for defining GUID struct and IsEqualGUID
#include <wdmguid.h>   // for PnP notification GUIDs
#include <ntddpar.h>   // for ParClass and ParPort interface GUIDs
#include "parxchlg.h"  // for PARXCH error log codes
#define ATOMIC  long
typedef int CDEV;
typedef int DEVT;
typedef int PARPORT_DRIVER;
typedef int PARPORT_CLASS;
typedef int PARPORT_DEVICE;
typedef int PARPORT_PORT;
typedef struct _KINTERRUPT KINTERRUPT;


#elif defined( SROS_WIN9X )
#include <windows.h>
#include <stdio.h>
#define ATOMIC  long

#define STATUS_SUCCESS                          (0x00000000L) // ntsubauth
#define STATUS_PORT_CONNECTION_REFUSED          (0xC0000041L)
#define STATUS_TOKEN_ALREADY_IN_USE             (0xC000012BL)
#define STATUS_NO_TOKEN                         (0xC000007CL)
#define STATUS_INVALID_PARAMETER                (0xC000000DL)
#define STATUS_INVALID_USER_BUFFER              (0xC00000E8L)
#define STATUS_INVALID_DEVICE_REQUEST           (0xC0000010L)
#define STATUS_DEVICE_CONFIGURATION_ERROR       (0xC0000182L)
#define STATUS_DEVICE_DATA_ERROR                (0xC000009CL)
#define STATUS_PIPE_EMPTY                       (0xC00000D9L)

typedef int DEVICE_POWER_STATE;
typedef int SYSTEM_POWER_STATE;
typedef int DEVICE_CAPABILITIES;
typedef int IO_REMOVE_LOCK;
typedef long DRIVER_OBJECT;
typedef long DEVICE_OBJECT;
typedef long FILE_OBJECT;
typedef char *UNICODE_STRING;
typedef long KMUTEX;
typedef long *KEVENT;
typedef unsigned long KIRQL;
typedef int CDEV;
typedef int DEVT;
typedef int PARPORT_DRIVER;
typedef int PARPORT_CLASS;
typedef int PARPORT_DEVICE;
typedef int PARPORT_PORT;
typedef struct _KINTERRUPT KINTERRUPT;


#elif defined( SROS_MSDOS )
#include <stdio.h>
#include <string.h>    // for strcmp function
#include <malloc.h>
#define WCHAR             char // unsigned short
#define ATOMIC            long
#define PowerDeviceD0      1
#define PowerSystemWorking 1

#define STATUS_SUCCESS                          (0x00000000L) // ntsubauth
#define STATUS_PORT_CONNECTION_REFUSED          (0xC0000041L)
#define STATUS_TOKEN_ALREADY_IN_USE             (0xC000012BL)
#define STATUS_NO_TOKEN                         (0xC000007CL)
#define STATUS_INVALID_PARAMETER                (0xC000000DL)
#define STATUS_INVALID_USER_BUFFER              (0xC00000E8L)
#define STATUS_INVALID_DEVICE_REQUEST           (0xC0000010L)
#define STATUS_DEVICE_CONFIGURATION_ERROR       (0xC0000182L)
#define STATUS_DEVICE_DATA_ERROR                (0xC000009CL)
#define STATUS_PIPE_EMPTY                       (0xC00000D9L)

typedef int DEVICE_POWER_STATE;
typedef int SYSTEM_POWER_STATE;
typedef int DEVICE_CAPABILITIES;
typedef int IO_REMOVE_LOCK;
typedef long DRIVER_OBJECT;
typedef long DEVICE_OBJECT;
typedef long FILE_OBJECT;
typedef char *UNICODE_STRING;
typedef unsigned long NTSTATUS;
typedef long KMUTEX;
typedef long *KEVENT;
typedef unsigned int *PKINTERRUPT;
typedef unsigned long KIRQL;
typedef long KSPIN_LOCK;
typedef int CDEV;
typedef int DEVT;
typedef int PARPORT_DRIVER;
typedef int PARPORT_CLASS;
typedef int PARPORT_DEVICE;
typedef int PARPORT_PORT;
typedef unsigned int KINTERRUPT;


#elif defined( SROS_LINUX )
#include <linux/module.h>
#include <linux/kernel.h>          // printk
#include <linux/init.h>            // module_init
#include <linux/moduleparam.h>     // module_param
#include <linux/ioport.h>          // check_region, etc
#include <linux/interrupt.h>       // request_irq
#include <linux/fs.h>              // everything ...
#include <linux/cdev.h>            // cdev
#include <linux/parport.h>         // functions and constants for low level parport driver
#include <linux/device.h>          // class_create, class_device_create, etc
#include <asm/uaccess.h>           // copy_to_user, VERIFY_WRITE, etc
#include <asm/io.h>                // inb, outb, etc
//#include "lnxver.h"                // Account for kernel changes. Mostly nop now.
                                   // Used to include
                                   // linux/poll.h, linux/sched.h, linux/init.h,
                                   // and maybe linux/vmalloc.h

#define WCHAR             char
#define ATOMIC            atomic_t
#define PowerDeviceD0      1
#define PowerSystemWorking 1

#define STATUS_SUCCESS                          (0x00000000L)
#define STATUS_PORT_CONNECTION_REFUSED          ((long)(-EACCES))
#define STATUS_TOKEN_ALREADY_IN_USE             ((long)(-EBUSY))
#define STATUS_NO_TOKEN                         ((long)(-ENXIO))
#define STATUS_INVALID_PARAMETER                ((long)(-EINVAL))
#define STATUS_INVALID_USER_BUFFER              ((long)(-EFAULT))
#define STATUS_INVALID_DEVICE_REQUEST           ((long)(-EINVAL))
#define STATUS_DEVICE_CONFIGURATION_ERROR       ((long)(-ENODEV))
#define STATUS_DEVICE_DATA_ERROR                ((long)(-EIO))
#define STATUS_PIPE_EMPTY                       ((long)(-EAGAIN))
#define STATUS_SECTION_NOT_EXTENDED             ((long)(-ENOBUFS))

typedef int DEVICE_POWER_STATE;
typedef int SYSTEM_POWER_STATE;
typedef int DEVICE_CAPABILITIES;
typedef int IO_REMOVE_LOCK;
typedef long DRIVER_OBJECT;
typedef long DEVICE_OBJECT;
typedef long FILE_OBJECT;
typedef char *UNICODE_STRING;
typedef unsigned long NTSTATUS;
typedef long KMUTEX;
typedef struct wait_queue *KEVENT;
typedef unsigned int *PKINTERRUPT;
typedef unsigned long KIRQL;
typedef spinlock_t KSPIN_LOCK;
typedef struct cdev CDEV;
typedef dev_t DEVT;
typedef struct parport_driver PARPORT_DRIVER;
typedef struct class PARPORT_CLASS;
typedef struct pardevice PARPORT_DEVICE;
typedef struct parport PARPORT_PORT;
typedef unsigned int KINTERRUPT;

#endif  // SROS_xxxxx



//WCT: Consider adding device interface functionality
//#include "srguids.h"  // user mode code will need to use these GUIDs too
// srguids would contain for {CAF53C68-A94C-11D2-BB4A-00C04FA330A6}
//DEFINE_GUID(GUID_WDMPORT,0xcaf53c68,0xa94c,0x11d2,0xbb,0x4a,0x00,0xc0,0x4f,0xa3,0x30,0xa6);





/* Common includes and defines:

These includes and defines are the same across all the OS's.

*/

#include "parxch.h"    // includes PARXCH defines shared with the user & DLL
#include "parxchkd.h"  // includes IOCTL codes and device name defines
#include "dbgprint.h"  // for debugging

#ifdef GPS_AVAILABLE
#include "gpsa2dkd.h"  // includes GPS + A/D Common Kernal Area defines
#else
typedef void GPSCKA;
#endif

#define DEFAULT_XCH_NAME     PARXCH_0
#define DEFAULT_XCH_ADDRESS  0x378
#define DEFAULT_XCH_COUNT    5
#define DEFAULT_XCH_IRQ      7
#define DEFAULT_XCH_MAJOR    123
#define DEFAULT_XCH_CLASS    "sr_instrument"
#define DEFAULT_XCH_SUFFIX   ""

#define MAX_STR_LENGTH        256

#define FH_MASK               0x03FF // 10 bits is 1024


#define RUNSTATE int
#define PWRSTATE int

#define RUNSTATE_STOPPED  0
#define RUNSTATE_RUNNING  1

#define PWRSTATE_POWEROFF 0
#define PWRSTATE_POWERON  1

#define IOCTL_STANDARD    0
#define IOCTL_INTERNAL    1






/* DEVICE INFORMATION STRUCTURE:

This structure contains all the global data which the driver code
needs.  It is passed to most functions.

PKFN_IOWR and PKFN_IORD are typedefs for the EPP/BPP read/write
function pointers in struct DeviceInformation.

*/

struct DeviceInformation;  // forward declaration

typedef void (*PKFN_IOWR)( struct DeviceInformation *Dev, unsigned char  Value);
typedef void (*PKFN_IORD)( struct DeviceInformation *Dev, unsigned char *Value);

SRLOCAL struct DeviceInformation {

        // Name of driver

        char DriverNameBuffer[MAX_STR_LENGTH];

        // Atod specific data

        int AtodXchModel;
        int AtodFifoBits;
        int AtodAnalogChannels;
        int AtodDigitalChannels;
        int AtodGpsChannels;
        int AtodCounterChannels;
        unsigned char AtodSizeMask;
        unsigned char AtodFifoSanityMask;
        unsigned char AtodFifoMarkMask;


        // Parallel port base address, mode, and irq.

        int ParPortBase;
        int ParPortCount;
        int ParPortMode;
        int ParPortIrq;
        int ParPortMapped;  // memory mapped or IO space
        int ParPortFound;
        int ParPortEcpOffset;

        // Parallel port register addresses.

        int ParPortData;  // ParPortBase + 0
        int ParPortStat;  // ParPortBase + 1
        int ParPortCtrl;  // ParPortBase + 2
        int EppPortAddr;  // ParPortBase + 3
        int EppPortData;  // ParPortBase + 4
        int EcpPortFifo;  // ParPortBase + 0x400, Data/Test Fifo or Config A
        int EcpPortCfgb;  // ParPortBase + 0x401, Configuration B
        int EcpPortExcr;  // ParPortBase + 0x402, Extended Control Register


        // Parallel port read/write function pointers that vary with
        // port EPP/BPP mode.

        PKFN_IORD ParPortAsRd;
        PKFN_IOWR ParPortAsWr;
        PKFN_IORD ParPortDsRd;
        PKFN_IOWR ParPortDsWr;


        // Parallel port current control register value.

        unsigned char ParPortCtrlValue;


        // ECP Parallel port original extended control register value.

        unsigned char EcpPortSaveEcrValue;


        // ParXch on board control register value.

        unsigned char BdCtrlValue;


        // ParXch / ADS1210 command register value (32 bits).

        long ADS1210SaveCrValue;


        // ParXch / ADS1210 data format (signed vs offset).

        int DataFormat;


        // Expected DRAM FIFO sanity bit value and FIFO history.

        unsigned char ExpectedFifoSanity;
        unsigned char FifoHistory[FH_MASK+1];
        int  FHindex;
        int  FHsanityInterval;


        // Decimation data for lower sampling rates ( < 40Hz ).

        int  ExtraDecNavg;
        int  ExtraDecNcount;
        long ExtraDecSum[PARXCH_MAX_ANALOG_CHANNELS];
        long ExtraDecSumGpsCount;
        long ExtraDecObcPerSample;


        // Controls whether the ParXch powers down on exit

        int PowerDownOnExitFlag;


        // Voltage is good (high Rev D, low Rev C)

        unsigned char VoltageIsGood;


        // Interrupt Object

        KINTERRUPT *IrqObj;


        // Spin lock for protecting in/out from DPC

        KSPIN_LOCK PortLock;
        KSPIN_LOCK ADS1210InitLock;


        // Common kernel data area for sharing with GPS driver

        GPSCKA *SrCommonArea;


        // Variables for working with the Mark (GPS/info) channel

        unsigned long AcqDataCount;  // pts acquired since A/D started
        long          MarkCount;
        long          MarkCurrent;
        int           IncludeMark;


        // Variables for working with the digital I/O channels

        int           IncludeDigital;


        // Variables for working with the gps counter channel

        int           IncludeCounter;



        // WDM support

        int            StartDone;
        int            ParPortAcquired;
        KMUTEX         ParPortAcquireMutex;
        DEVICE_OBJECT *ParPortDevObj;
        FILE_OBJECT   *ParPortFileObj;

        DRIVER_OBJECT      *DriverObj;
        DEVICE_OBJECT      *ThisFdoDevObj;
        DEVICE_OBJECT      *LowerDevObj;
        DEVICE_OBJECT      *PdoDevObj;
        void               *NotificationHandle;      // PlugPlay Notification Handle
        UNICODE_STRING      LinkName;
        RUNSTATE            CurrDevState;
        RUNSTATE            PrevDevState;
        DEVICE_POWER_STATE  CurrDevPower;
        SYSTEM_POWER_STATE  CurrSysPower;
        DEVICE_CAPABILITIES DevCaps;
        IRP                *SystemIrp;
        IO_REMOVE_LOCK      RemoveLock;
        long                WorkItemCount;

        // New for Linux 2.6
        CDEV     Cdev;         // Char device structure
        IRP      DriverIrp;    // Copied from passed user space IRP
        DEVT     DevNum;       // Device number

        // Linux ParPort PNP support

        PARPORT_DRIVER *ParPortDriver;
        PARPORT_CLASS  *ParPortClass;
        PARPORT_DEVICE *ParPortDevice;
        PARPORT_PORT   *ParPortPort;
        char            SysFsClassName[MAX_STR_LENGTH];

// WCT: Check these
//      POWERSTATE CurrPowerState;     // Oney notes say not needed
//      UNICODE_STRING InterfaceName;  // LinkName may be sufficient



        // Usage lock so only 1 process is using the driver at a time.

        int NumberOpensLock;


        // Error message buffer.

        WCHAR MsgBuffer[MAX_STR_LENGTH];


} ParXchDeviceInfo;

typedef struct DeviceInformation DEVINFO;







// Global Variable

UNICODE_STRING DriverRegPath;



////////////////////////////////////////////////////////////////////////
/* FUNCTION PROTOTYPES:
 *
 *
 *
 *
 */

// Group 2: OS specific helper functions:
//
//
//
//
//

SRLOCAL void OsPortIn( int port, unsigned char *value );
SRLOCAL void OsPortOut( int port, unsigned char value );
SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes );
SRLOCAL unsigned long OsCopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes );
SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info );
SRLOCAL long OsSetReadParms( IRP *Irp, unsigned long *OutBufferSize, long **IOBuffer );
SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command );
SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout );
SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                                  unsigned long Command,
                                  int *OutBuffer,
                                  unsigned long ReturnedBytes );
SRLOCAL void OsWriteFifoHistory( DEVINFO *Dev,
                                 unsigned char MuxedBits,
                                 unsigned char Expected );

// OS specific macro helper functions
// OsAtomicSet,OsAtomicInc,OsAtomicDec,OsAtomicAdd,OsAtomicSub,OsAtomicRead
// OsSpinLockAcquire, OsSpinLockRelease
// OsModuleUseCountInc,OsModuleUseCountDec


// Group 1: PARxCH hardware and behavior functions:
//
//
//
//
//

SRLOCAL void ParXch_ParPortInit( DEVINFO *Dev );
SRLOCAL void ParXch_EcpPortInit( DEVINFO *Dev );
SRLOCAL void ParXch_EcpPortRestore( DEVINFO *Dev );
SRLOCAL void ParXch_EppPortClearTimeout( DEVINFO *Dev );

SRLOCAL void ParXch_ParPortDataRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_ParPortDataWr( DEVINFO *Dev, unsigned char Value );
SRLOCAL void ParXch_ParPortStatRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_ParPortStatWr( DEVINFO *Dev, unsigned char Value );
SRLOCAL void ParXch_ParPortCtrlWr( DEVINFO *Dev, unsigned char Value );
SRLOCAL void ParXch_ParPortCtrlRw( DEVINFO *Dev, unsigned char Value);
SRLOCAL void ParXch_ParPortCtrlIrqen( DEVINFO *Dev, unsigned char Value);
SRLOCAL void ParXch_ParPortCtrlDirRead( DEVINFO *Dev);
SRLOCAL void ParXch_ParPortCtrlDirWrite( DEVINFO *Dev);

SRLOCAL void ParXch_Wait( DEVINFO *Dev, unsigned int Milliseconds );
SRLOCAL int  ParXch_VoltageGood( DEVINFO *Dev );
SRLOCAL int  ParXch_PowerDown( DEVINFO *Dev );
SRLOCAL int  ParXch_PowerUp( DEVINFO *Dev );

SRLOCAL void ParXch_EppAsRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_EppAsWr( DEVINFO *Dev, unsigned char  Value );
SRLOCAL void ParXch_EppDsRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_EppDsWr( DEVINFO *Dev, unsigned char  Value );

SRLOCAL void ParXch_BppAsRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_BppAsWr( DEVINFO *Dev, unsigned char  Value );
SRLOCAL void ParXch_BppDsRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_BppDsWr( DEVINFO *Dev, unsigned char  Value );

SRLOCAL void ParXch_BdCtrlWr( DEVINFO *Dev, unsigned char Value );
SRLOCAL void ParXch_BdCtrlUnlock( DEVINFO *Dev );
SRLOCAL void ParXch_BdCtrlInit( DEVINFO *Dev );
SRLOCAL void ParXch_BdCtrlMode( DEVINFO *Dev, unsigned char Value );
SRLOCAL void ParXch_BdCtrlDrdyIrqReq( DEVINFO *Dev);
SRLOCAL void ParXch_BdCtrlFifoReset( DEVINFO *Dev, unsigned char Value );
SRLOCAL void ParXch_BdCtrlUserLed( DEVINFO *Dev, int State );

SRLOCAL void ParXch_ADS1210_WriteByte( DEVINFO *Dev, unsigned char  Value );
SRLOCAL void ParXch_ADS1210_ReadByte( DEVINFO *Dev, unsigned char *Value );

SRLOCAL void ParXch_ADS1210_ReadReg8( DEVINFO *Dev, int reg, unsigned char *Value );
SRLOCAL void ParXch_ADS1210_WriteReg8( DEVINFO *Dev, int reg, unsigned char Value );

SRLOCAL void ParXch_DramFifoReset( DEVINFO *Dev );
SRLOCAL void ParXch_DramFifoRd( DEVINFO *Dev, unsigned char *Value );
SRLOCAL void ParXch_DramFifoWr( DEVINFO *Dev, unsigned char Value );
SRLOCAL int  ParXch_DramFifoReady( DEVINFO *Dev );
SRLOCAL int  ParXch_DramFifoFull( DEVINFO *Dev );
SRLOCAL int  ParXch_DramFifoReadDemux( DEVINFO *Dev,
                                       long *AnalogValues,
                                       long *DigitalValues,
                                       int *Mark,
                                       long *GpsCounter);

SRLOCAL void ParXch_UserIoRd( DEVINFO *Dev, int *Value );
SRLOCAL void ParXch_UserIoWr( DEVINFO *Dev, int Value );

SRLOCAL void ParXch_InterruptEnable( DEVINFO *Dev );
SRLOCAL void ParXch_InterruptDisable( DEVINFO *Dev );

SRLOCAL void ParXch_Init( DEVINFO *Dev, int *IOBuffer );
SRLOCAL void ParXch_Start( DEVINFO *Dev );
SRLOCAL void ParXch_Stop( DEVINFO *Dev );


SRLOCAL void ParXch_FillDevInfo( DEVINFO *Dev, char *DriverName,
                                 int ParPortBase, int ParPortEcpOffset,
                                 int ParPortIrq );
SRLOCAL int  ParXch_CheckHealth( DEVINFO *Dev );
SRLOCAL int  ParXch_AttachGps( DEVINFO *Dev, void *CommonArea );
SRLOCAL void ParXch_ReleaseGps( DEVINFO *Dev );
SRLOCAL void ParXch_ResetMarkCount( DEVINFO *Dev );

SRLOCAL int ParXch_DrvInterrupt( DEVINFO *Dev );


// Driver dispatch functions called "directly" from the OS

SRLOCAL long ParXch_DrvOpen( DEVINFO *Dev, IRP *Irp );
SRLOCAL long ParXch_DrvClose( DEVINFO *Dev, IRP *Irp );
SRLOCAL long ParXch_DrvRead( DEVINFO *Dev, IRP *Irp );
SRLOCAL long ParXch_DrvIoctl( DEVINFO *Dev, IRP *Irp );

SRLOCAL void ParXch_DrvUnload( DEVINFO *Dev );



// Group 3: Top level OS dependent initialization and "dispatch" functions:
//
//
//
//
//

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )

struct PnpNotifyContextStruct {
        PIO_WORKITEM  WorkItem;
        WCHAR         NotifySymLink[MAX_STR_LENGTH];
};
typedef struct PnpNotifyContextStruct PNP_NOTIFY_CONTEXT;

DRIVER_INITIALIZE     DriverEntry;
DRIVER_ADD_DEVICE     Wdm_DrvAddDevice;
DRIVER_DISPATCH       Wdm_DrvPnp;
DRIVER_DISPATCH       Wdm_DrvPower;
DRIVER_DISPATCH       Wdm_DrvWmi;
DRIVER_DISPATCH       Wdm_DrvOpen;
DRIVER_DISPATCH       Wdm_DrvClose;
DRIVER_DISPATCH       Wdm_DrvRead;
DRIVER_DISPATCH       Wdm_DrvIoctl;
DRIVER_UNLOAD         Wdm_DrvUnload;
IO_WORKITEM_ROUTINE   Wdm_WorkItemNotifyCallback;
KSERVICE_ROUTINE      Wdm_DrvInterrupt;
IO_COMPLETION_ROUTINE Wdm_OnRequestComplete;
IO_COMPLETION_ROUTINE Wdm_OnCompletePowerSystem;
IO_COMPLETION_ROUTINE Wdm_OnCompletePowerDeviceUp;


// Functions called from the OS

NTSTATUS DriverEntry( DRIVER_OBJECT *DriverObject,
                      UNICODE_STRING *RegistryPath );

void     Wdm_DrvUnload(    DRIVER_OBJECT *DriverObject );
NTSTATUS Wdm_DrvAddDevice( DRIVER_OBJECT *DriverObject,  DEVICE_OBJECT *PDO );

SRLOCAL NTSTATUS Wdm_GetParameters( UNICODE_STRING *RegistryPath,
                                    UNICODE_STRING *ParXchName,
                                    unsigned long *pPortBase,
                                    unsigned long *pPortIrq,
                                    unsigned long *pPortFound );
SRLOCAL NTSTATUS Wdm_ErrorLog( void *IoObject,
                               IRP *Irp,
                               NTSTATUS CurrStatus,
                               unsigned long ErrCode,
                               const WCHAR *Msg );

NTSTATUS Wdm_DrvOpen(  DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_DrvClose( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_DrvRead(  DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_DrvIoctl( DEVICE_OBJECT *FDO, IRP *Irp );

BOOLEAN Wdm_DrvInterrupt( PKINTERRUPT pInterruptObject, VOID *Context);


// Plug and Play Functions

NTSTATUS Wdm_DrvPnp(               DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpStartDevice(       DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpStopDevice(        DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpRemoveDevice(      DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryCapabilities( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpCancelStop(        DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpCancelRemove(      DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryStop(         DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryRemove(       DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryDevRelations( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryInterface(    DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpFilterResourceReq( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpSurpriseRemoval(   DEVICE_OBJECT *FDO, IRP *Irp );

NTSTATUS Wdm_PnpNotifyInterfaceChange(
               PDEVICE_INTERFACE_CHANGE_NOTIFICATION NotificationStruct,
               PDEVICE_OBJECT                        FDO
               );
VOID Wdm_WorkItemNotifyCallback( PDEVICE_OBJECT FDO, PVOID Context );
VOID Wdm_WorkItemNotifyCleanup( DEVINFO *Dev, PNP_NOTIFY_CONTEXT *PnpNotifyContext,
                                BOOLEAN FreeMutex );

SRLOCAL NTSTATUS Wdm_ParPortAcquire( DEVINFO *Dev, unsigned long DeviceBase,
                                     unsigned long DeviceIrq, UNICODE_STRING *PptName,
                                     VOID *Context);
SRLOCAL NTSTATUS Wdm_ParPortRelease( DEVINFO *Dev,
                                     DEVICE_OBJECT *PPDevObj,
                                     FILE_OBJECT *PPFileObj,
                                     KINTERRUPT *PPIrqObj );
SRLOCAL NTSTATUS Wdm_ParPortConnectInterrupt( DEVINFO *Dev );
SRLOCAL NTSTATUS Wdm_ParPortGetPdo( DEVINFO *Dev,
                                    DEVICE_OBJECT *TargetDeviceObject,
                                    FILE_OBJECT *TargetFileObject,
                                    DEVICE_OBJECT **ParportPDO );

SRLOCAL NTSTATUS Wdm_ResourceSet(   DEVICE_OBJECT *FDO, IRP *Irp );
SRLOCAL NTSTATUS Wdm_ResourceClear( DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_StartDevice(   DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_StopDevice(    DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_RemoveDevice(  DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_SetRegPortFound( DEVINFO *Dev );


// Power Functions

NTSTATUS Wdm_DrvPower(                DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PowerSetSystem(          DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_OnCompletePowerSystem(   DEVICE_OBJECT *FDO, IRP *Irp, void *Context );
void     Wdm_OnCompletePowerFinish(   DEVICE_OBJECT *FDO, UCHAR MinorFunction,
                                      POWER_STATE PowerState, void *Context,
                                      IO_STATUS_BLOCK *IoStatus );
NTSTATUS Wdm_PowerSetDevice(          DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_OnCompletePowerDeviceUp( DEVICE_OBJECT *FDO, IRP *Irp, void *Context );
NTSTATUS Wdm_PowerForward(            DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PowerFail( IRP *Irp, NTSTATUS Status );


// Windows Management Interface Functions

NTSTATUS Wdm_DrvWmi(   DEVICE_OBJECT *FDO, IRP *Irp );


// Helper functions

SRLOCAL NTSTATUS Wdm_OnRequestComplete( DEVICE_OBJECT *FDO,
                                        IRP *Irp,
                                        void *Context );

SRLOCAL NTSTATUS Wdm_IrpBuildSendAndWait(
                            ULONG           Internal,
                            ULONG           IoControlCode,
                            PDEVICE_OBJECT  TargetDeviceObject,
                            PVOID           InputBuffer,
                            ULONG           InputBufferLength,
                            PVOID           OutputBuffer,
                            ULONG           OutputBufferLength,
                            PLARGE_INTEGER  RequestedTimeout
                            );

SRLOCAL NTSTATUS Wdm_IrpForwardAndWait( DEVICE_OBJECT *FDO, IRP *Irp );

SRLOCAL NTSTATUS Wdm_IrpForward( DEVICE_OBJECT *FDO, IRP *Irp );

SRLOCAL void Wdm_ResourceFree( DEVICE_OBJECT *DeviceObject,
                               UNICODE_STRING *LinkDeviceName );






#elif defined( SROS_WIN9X ) || defined( SROS_MSDOS )

// MSDOS Helper Function
SRLOCAL int Msdos_NameToAddress( char *DriverName );
SRLOCAL int Msdos_NameToIrq( char *DriverName );

/* The rest of the functions are declared in parxchkd.h so they can
   be called directly by the parxch.c code instead of by the OS

         Msdos_DrvOpen
         Msdos_DrvRead
         Msdos_DrvIoctl
         Msdos_DrvClose
         Msdos_DrvGetLastError
*/





#elif defined( SROS_LINUX )

// Linux Helper Functions

SRLOCAL void Linux_GetParameters( void );
SRLOCAL void Linux_ResourceFree( DEVINFO *Dev, int DeviceCdev,
                                 int PortIrq,  int PortAddress,
                                 int DevMajor, char *DriverName );
SRLOCAL int Linux_ResourceReserve( DEVINFO *Dev );

// Linux functions for interacting with the low level ParPort driver

int  Linux_ParPortPreempt( void *handle );
void Linux_ParPortIrq( void *dev_id );
void Linux_ParPortAttachPort( struct parport *port );
void Linux_ParPortDetachPort( struct parport *port );
int  Linux_ParPortRegisterClaim( DEVINFO *Dev );
void Linux_ParPortUnregisterRelease( DEVINFO *Dev );

// Linux functions called from the OS (interrupt & dispatch)

irqreturn_t Linux_DrvInterrupt( int irq, void *dev_id );

int     Linux_DrvOpen(  struct inode *inode, struct file *filp );
int     Linux_DrvClose( struct inode *inode, struct file *filp );
int     Linux_DrvIoctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg );
ssize_t Linux_DrvRead(  struct file *filp, char *buf, size_t count, loff_t *offset );

// These two are declared below using the new module_init and module_exit macros
//int      DriverEntry( void );  // was    init_module
//void Linux_DrvUnload( void );  // was cleanup_module


#endif  // SROS_xxxxx









////////////////////////////////////////////////////////////////////////
// START OF CODE ...
//
//
//




////////////////////////////////////////////////////////////////////////
/* Group 2: OS SPECIFIC HELPER FUNCTIONS AND MACROS:
 *
 * These functions are repeated several times, so you MUST give a
 * valid SROS_xxx compile time define to select which set of these
 * functions is appropriate for your OS.
 *
 */

/*
In order to protect the rest of the code from having lots of
conditional sections, several helper functions and macros have been
defined.  The helper names are repeated once for each OS, but the
SROS_xxxxx define is used to select the one correct set to include at
compile time.

These helper functions include the in and out port I/O functions,
Atomic (ie non-interruptible) functions, and etc.  For example, under
Win2K and Linux, in and out are best called through macros provided
by the system, while under DOS inline assembler suffices.
*/


// *************************************************************************

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )

// Lowest level PC I/O functions

SRLOCAL void OsPortIn( int port, unsigned char *value ) {

        *value = READ_PORT_UCHAR( (UCHAR *)port );

}
SRLOCAL void OsPortOut( int port, unsigned char value ) {

        WRITE_PORT_UCHAR( (UCHAR *)port, value );
}

// Functions to transfer data between user and kernel space

SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes ) {
        RtlCopyMemory( UserBuff, KernelBuff, Nbytes );
        return( Nbytes );
}
SRLOCAL unsigned long CopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes ) {
        RtlCopyMemory( KernelBuff, UserBuff, Nbytes );
        return( Nbytes );
}

// Misc functions

SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info ) {
        Irp->IoStatus.Status      = (NTSTATUS)Status;
        Irp->IoStatus.Information = Info;
        IoCompleteRequest( Irp, IO_NO_INCREMENT );
}

SRLOCAL long OsSetReadParms( IRP *Irp, unsigned long *OutBufferSize, long **IOBuffer ) {

        IO_STACK_LOCATION *IrpStack;
        unsigned long      DevFlags;

        // Set up variables

        IrpStack        = IoGetCurrentIrpStackLocation( Irp );
        DevFlags        = IrpStack->DeviceObject->Flags;
        *OutBufferSize  = IrpStack->Parameters.Read.Length;


        // Get appropriate pointer for buffer

        if (DevFlags & DO_DIRECT_IO) {
                if (Irp->MdlAddress)
#if defined( SROS_WIN2K ) || defined( SROS_WINXP )
                        *IOBuffer = (long*)MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
#else // defined( SROS_WIN98 ) || defined( SROS_WINNT )
                        *IOBuffer = (long*)MmGetSystemAddressForMdl(Irp->MdlAddress);
#endif
                else
                        *IOBuffer = NULL;
                }

        else if (DevFlags & DO_BUFFERED_IO)
                *IOBuffer = (long *)Irp->AssociatedIrp.SystemBuffer;

        else // Must be NEITHER_IO

                *IOBuffer = (long *)Irp->UserBuffer;


        // Verify buffer pointer

        if ( (*OutBufferSize == 0) || (!(*IOBuffer)) )

                return( STATUS_INVALID_PARAMETER );
        else
                return( STATUS_SUCCESS );
}

SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command ) {

        IO_STACK_LOCATION *IrpStack;

        IrpStack  = IoGetCurrentIrpStackLocation( Irp );
        *Command  = IrpStack->Parameters.DeviceIoControl.IoControlCode;
}

SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout ) {

        IO_STACK_LOCATION *IrpStack;

        IrpStack  = IoGetCurrentIrpStackLocation( Irp );


        // Assign buffer sizes

        *Nin  = IrpStack->Parameters.DeviceIoControl.InputBufferLength;
        *Nout = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;


        // Assign buffers

        *InBuffer  = (int *)Irp->AssociatedIrp.SystemBuffer;
        *OutBuffer = (int *)Irp->AssociatedIrp.SystemBuffer;

        if ((IOCTL_METH(Command) == METHOD_IN_DIRECT) ||
            (IOCTL_METH(Command) == METHOD_OUT_DIRECT)) {

                if (Irp->MdlAddress)
#if defined( SROS_WINXP ) || defined( SROS_WIN2K )
                        *OutBuffer = (int*)MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
#else // defined( SROS_WIN98 ) || defined( SROS_WINNT )
                        *OutBuffer = (int*)MmGetSystemAddressForMdl(Irp->MdlAddress);
#endif
                else
                        *OutBuffer = NULL;
                }


        // Check that buffers exist if they are given a length

        if ( (*Nin > 0) && !(*InBuffer) ) {
                dbg5_printk("InBuffer at 0x%X, expected size %d\n", *InBuffer, *Nin );
                return( STATUS_INVALID_USER_BUFFER );
                }
        if ( (*Nout > 0) && !(*OutBuffer) ) {
                dbg5_printk("OutBuffer at 0x%X, expected size %d\n", *OutBuffer, *Nout );
                return( STATUS_INVALID_USER_BUFFER );
                }


        // No filling needed for InBuffer since OS functions do that

        return( STATUS_SUCCESS );
}

SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                                  unsigned long Command,
                                  int *OutBuffer,
                                  unsigned long ReturnedBytes ) {

        // No filling needed for OutBuffer since OS functions do that

        return;
}

SRLOCAL void OsWriteFifoHistory( DEVINFO *Dev,
                                 unsigned char MuxedBits,
                                 unsigned char Expected ) {
        HANDLE   ErrFp;
        NTSTATUS Status;
        IO_STATUS_BLOCK   IoStatusBlock;
        OBJECT_ATTRIBUTES ObjAttr;
        UNICODE_STRING    ObjName;
        ULONG             CreateOptions;
        int               nbuff, i, j, k;
        unsigned int      HistoryValue;
        char              tmpstr[MAX_STR_LENGTH];

        RtlInitUnicodeString( &ObjName, L"\\??\\C:\\sanity.err" );
        InitializeObjectAttributes(
                                   &ObjAttr,
                                   &ObjName,
                                   OBJ_OPENIF, // Attributes
                                   NULL,       // RootDirectory
                                   NULL        // SecurityDescriptor
                                  );
        CreateOptions = FILE_SYNCHRONOUS_IO_NONALERT | FILE_NON_DIRECTORY_FILE;

        Status = ZwCreateFile(
                              &ErrFp,
                              GENERIC_WRITE,           // Access mask
                              &ObjAttr,
                              &IoStatusBlock,
                              NULL,                    // AllocationSize
                              FILE_ATTRIBUTE_NORMAL,
                              0,                       // ShareAccess
                              FILE_OPEN_IF,            // CreateDisposition
                              CreateOptions,
                              NULL,
                              0
                             );
        if (!NT_SUCCESS(Status)) {
                dbg0_printk("Can't open sanity.err 0x%X\n", Status);
                return;
                }




        nbuff = sprintf(tmpstr,"Fifo Sanity Error: expected 0x%02X, found 0x%02X\nFifo Data most recent first:\n",
                 Expected, MuxedBits );

        Status  = ZwWriteFile(
                              ErrFp,
                              NULL,
                              NULL,
                              NULL,
                              &IoStatusBlock,
                              tmpstr,
                              nbuff*sizeof(char),
                              NULL, // FILE_USE_FILE_POINTER_POSITION,
                              NULL
                             );
        if (!NT_SUCCESS(Status)) {
                dbg0_printk("Can't write msg buffer to sanity.err 0x%X\n", Status);
                ZwClose( ErrFp );
                return;
                }

        j = Dev->FHsanityInterval;
        for ( i = 0 ; i < FH_MASK+1 ; i++ ) {
                HistoryValue = Dev->FifoHistory[Dev->FHindex] & 0xFF;
                k = Dev->FHsanityInterval - j;
                if (j == Dev->FHsanityInterval) {
                        nbuff = sprintf(tmpstr,"%4d: (%2d) 0x%02X   sanity\n\n",
                                        i, k, HistoryValue);
                        j = 1;
                        }
                else {
                        nbuff = sprintf(tmpstr,"%4d: (%2d) 0x%02X\n",
                                        i, k, HistoryValue);
                        j++;
                        }
                Dev->FHindex -= 1;
                Dev->FHindex &= FH_MASK;

                Status  = ZwWriteFile(
                                      ErrFp,
                                      NULL,
                                      NULL,
                                      NULL,
                                      &IoStatusBlock,
                                      tmpstr,
                                      nbuff*sizeof(char),
                                      NULL, // FILE_USE_FILE_POINTER_POSITION,
                                      NULL
                                     );
                if (!NT_SUCCESS(Status)) {
                        dbg0_printk("Can't write point %d to sanity.err 0x%X\n",
                                                                   i, Status);
                        ZwClose( ErrFp );
                        return;
                        }

                }



        ZwClose( ErrFp );

        return;
}

#define OsModuleUseCountInc
#define OsModuleUseCountDec

// Atomic operation functions

#define OsAtomicSet( pAtom, Value )  InterlockedExchange( (pAtom), (Value) )
#define OsAtomicInc( pAtom )         InterlockedIncrement( (pAtom) )
#define OsAtomicDec( pAtom )         InterlockedDecrement( (pAtom) )
#define OsAtomicAdd( pAtom, Amount ) InterlockedExchangeAdd( (pAtom), (Amount) )
#define OsAtomicSub( pAtom, Amount ) InterlockedExchangeAdd( (pAtom), -(Amount) )
#define OsAtomicRead( pAtom )        *(pAtom)

// Spinlock functions

#define OsSpinLockAcquire( pLock, OldIrqlOrFlags ) \
        KeAcquireSpinLock( (pLock), &(OldIrqlOrFlags) );
#define OsSpinLockRelease( pLock, OldIrqlOrFlags ) \
        KeReleaseSpinLock( (pLock), (OldIrqlOrFlags) );


// *************************************************************************

#elif defined( SROS_WIN9X ) || defined( SROS_MSDOS )

// Lowest level PC I/O functions

SRLOCAL void OsPortIn( int port, unsigned char *value) {

        unsigned char temp;

        _asm {

                mov     dx,WORD PTR port
                in      al,dx
                mov     BYTE PTR temp,al

                }

        *value = temp;

}
SRLOCAL void OsPortOut( int port, unsigned char value) {

        _asm {

                mov     dx,WORD PTR port
                mov     al,BYTE PTR value
                out     dx,al

                }

}

// Functions to transfer data between user and kernel space

SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes ) {

        unsigned long i;
        unsigned char *userptr;
        unsigned char *kernelptr;

        userptr   = (unsigned char *)UserBuff;
        kernelptr = (unsigned char *)KernelBuff;

        for ( i = 0 ; i < Nbytes ; i++ )
                *userptr++ = *kernelptr++;

        return( Nbytes );
}

SRLOCAL unsigned long OsCopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes ) {

        unsigned long i;
        unsigned char *userptr;
        unsigned char *kernelptr;

        userptr   = (unsigned char *)UserBuff;
        kernelptr = (unsigned char *)KernelBuff;
        for ( i = 0 ; i < Nbytes ; i++ )
                *userptr++ = *kernelptr++;

        return( Nbytes );
}

// Misc functions

SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info ) {

        if (Irp)
                Irp->ReturnedBytes = Info;

}

SRLOCAL long OsSetReadParms( IRP *Irp, unsigned long *OutBufferSize, long **IOBuffer ) {

        // Set parameters

        *OutBufferSize = Irp->OutSize;
        *IOBuffer      = (long *)Irp->OutBuffer;


        // Verify buffer pointer

        if ( (*OutBufferSize == 0) || !(*IOBuffer) )

                return( STATUS_INVALID_PARAMETER );
        else
                return( STATUS_SUCCESS );
}

SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command ) {
        *Command  = Irp->Command;
}

SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout ) {

        // Assign buffer sizes

        *Nin  = (int)Irp->InSize;
        *Nout = (int)Irp->OutSize;


        // Assign buffers

        *InBuffer  = (int *)Irp->InBuffer;
        *OutBuffer = (int *)Irp->OutBuffer;


        // Check that a buffer exists since all commands pass (or get) data

            if ( !(*OutBuffer) && !(*InBuffer) )
                    return( STATUS_INVALID_USER_BUFFER );



        // No filling needed for InBuffer since "kernel" space is user space

        return( STATUS_SUCCESS );
}

SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                          unsigned long Command,
                          int *OutBuffer,
                          unsigned long ReturnedBytes ) {

        // No filling needed for OutBuffer since "kernel" space is user space

        return;
}

SRLOCAL void OsWriteFifoHistory( DEVINFO *Dev,
                                 unsigned char MuxedBits,
                                 unsigned char Expected ) {
        FILE *ErrFp;
        int   i, j, k;
        unsigned int HistoryValue;

        ErrFp = fopen( "sanity.err", "a+b" );
        if (!ErrFp)
                return;

        fprintf( ErrFp, "Fifo Sanity Error: expected 0x%02X, found 0x%02X\n",
                 Expected, MuxedBits );
        fprintf( ErrFp, "Fifo Data most recent first:\n" );

        j = Dev->FHsanityInterval;
        for ( i = 0 ; i < FH_MASK+1 ; i++ ) {
                HistoryValue = Dev->FifoHistory[Dev->FHindex] & 0xFF;
                k = Dev->FHsanityInterval - j;
                if (j == Dev->FHsanityInterval) {
                        fprintf(ErrFp,"\n%4d: (%2d) 0x%02X   sanity\n",
                                  i, k, HistoryValue);
                        j = 1;
                        }
                else {
                        fprintf(ErrFp,"%4d: (%2d) 0x%02X\n",
                                  i, k, HistoryValue);
                        j++;
                        }
                Dev->FHindex -= 1;
                Dev->FHindex &= FH_MASK;
                }



        fclose( ErrFp );


        return;
}

#define OsModuleUseCountInc
#define OsModuleUseCountDec


// Atomic operation functions

#define OsAtomicSet( pAtom, Value )  *(pAtom) = (Value)
#define OsAtomicInc( pAtom )         *(pAtom) += 1
#define OsAtomicDec( pAtom )         *(pAtom) -= 1
#define OsAtomicAdd( pAtom, Amount ) *(pAtom) += (Amount)
#define OsAtomicSub( pAtom, Amount ) *(pAtom) -= (Amount)
#define OsAtomicRead( pAtom )        *(pAtom)

// Spinlock functions

#define OsSpinLockAcquire( pLock, OldIrqlOrFlags )    (OldIrqlOrFlags) = 1;
#define OsSpinLockRelease( pLock, OldIrqlOrFlags )    ;



// *************************************************************************

#elif defined( SROS_LINUX )

// Lowest level PC I/O functions

SRLOCAL void OsPortIn( int port, unsigned char *value) {

        rmb();                // Ensure that reads occur in the order written
        *value = inb( port );
}

SRLOCAL void OsPortOut( int port, unsigned char value) {

        wmb();                // Ensure that writes occur in the order written
        outb( value, port );
}

// Functions to transfer data between user and kernel space

SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes ) {
        unsigned long result;

        result = __copy_to_user( UserBuff, KernelBuff, Nbytes );

        return( result );
}
SRLOCAL unsigned long OsCopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes ) {
        unsigned long result;

        result = __copy_from_user( KernelBuff, UserBuff, Nbytes );

        // Note: result is number of bytes NOT transferred

        return( result );
}

// Misc functions

SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info ) {

        int retval;
        IRP *UserIrp;

        dbg5_printk("In OsIrpFinish\n");

        if (Irp) {
                Irp->ReturnedBytes = Info;

                // When called from DrvIoctl, we have a user space IRP
                // that needs to get this return Info.  But there is
                // no corresponding user space IRP when called by DrvRead.

                UserIrp = (IRP *)(Irp->UserIrp);
                if ( UserIrp )
                        retval = copy_to_user( UserIrp, Irp, sizeof(IRP) );
                }
}

SRLOCAL long OsSetReadParms( IRP *Irp, unsigned long *OutBufferSize, long **IOBuffer ) {

        dbg5_printk( "In OsSetReadParms\n" );

        // Set parameters

        *OutBufferSize = Irp->OutSize;
        *IOBuffer      = (long *)Irp->OutBuffer;


        // Verify buffer pointer

        if ( (*OutBufferSize == 0)   ||
             (!(*IOBuffer))          ||
             (!access_ok( VERIFY_WRITE, (void*)(*IOBuffer), *OutBufferSize)) )

                return( STATUS_INVALID_PARAMETER );

        else
                return( STATUS_SUCCESS );
}

SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command ) {

        dbg5_printk("In OsSetIoctlCommand\n");

        *Command  = Irp->Command;
}

SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout ) {

        // BufferArea is defined at maximum size needed for any command

        static unsigned char BufferArea[sizeof(INITXCHPARMS)];
        int ReadCopy, WriteCopy, Direct, MaxBuffArea, Err;

        dbg5_printk( "In OsSetIoctlBuffers\n" );


        /* _IOC type is user-oriented, verify_area is kernel-oriented
         * so read/write are reversed.  VERIFY_WRITE handles R/W.
         * eg _IOC_READ means read from the device, so driver must write
         * to user space.
         */

        // Check if driver must write to user space

        if ( _IOC_DIR(Command) & _IOC_READ )
                WriteCopy = 1;
        else
                WriteCopy = 0;


        // Check if driver must read from user space

        if ( _IOC_DIR(Command) & _IOC_WRITE )
                ReadCopy  = 1;
        else
                ReadCopy  = 0;




        // Unpack IoctlIrp structure

        if ( ( Irp->DataMethod == METHOD_IN_DIRECT ) ||
             ( Irp->DataMethod == METHOD_OUT_DIRECT ) )
                Direct  = 1;
        else
                Direct  = 0;


        *Nin  = (int)Irp->InSize;
        *Nout = (int)Irp->OutSize;


        if (Direct) {
                *InBuffer  = (int*)Irp->InBuffer;
                *OutBuffer = (int*)Irp->OutBuffer;
                }
        else {
                MaxBuffArea = sizeof(BufferArea);
                if ( (*Nin > MaxBuffArea) || (*Nout > MaxBuffArea) ) {
                        dbg5_printk( "In OsSetIoctlBuffers fail SECTION_NOT_EXTENDED\n" );
                        return( STATUS_SECTION_NOT_EXTENDED );
                        }
                *InBuffer = *OutBuffer = (int *)BufferArea;
                }


        // Check input buffer is ok to read

        if ( (*Nin > 0) && ReadCopy ) {
                Err = access_ok( VERIFY_READ,
                                 (void*)Irp->InBuffer,
                                  *Nin );
                if (!Err) {
                        dbg5_printk( "In OsSetIoctlBuffers fail INVALID ReadCopy\n" );
                        return( STATUS_INVALID_USER_BUFFER );
                        }
                }


        // Check output buffer is ok to read/write

        if ( (*Nout > 0) && WriteCopy ) {
                Err = access_ok( VERIFY_WRITE,
                                 (void*)Irp->OutBuffer,
                                  *Nout );
                if (!Err) {
                        dbg5_printk( "In OsSetIoctlBuffers fail INVALID WriteCopy\n" );
                        return( STATUS_INVALID_USER_BUFFER );
                        }
                }


        // Fill Ioctl buffer from user space if not using direct IO

        // WCT - Review for new IRP location

        if (ReadCopy && !Direct) {
                dbg5_printk( "OsSetIoctlBuffers readcopy for not direct IO\n" );
                OsCopyFromUser( *InBuffer, Irp->InBuffer, *Nin );
                }


        return( STATUS_SUCCESS );
}

SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                                  unsigned long Command,
                                  int *OutBuffer,
                                  unsigned long ReturnedBytes ) {

        int  retval;
        IRP *UserIrp;

        dbg5_printk( "In OsReturnIoctlBuffer\n" );


        // Save number of bytes being returned

        Irp->ReturnedBytes = ReturnedBytes;

        // Ensure user space IRP from DrvIoctl gets filled
        // but don't try from DrvRead since there is no user IRP

        UserIrp = (IRP *)(Irp->UserIrp);
        if (UserIrp)
                retval = copy_to_user( UserIrp, Irp, sizeof(IRP) );


        // Copy has been done else where for this command

        if ((Irp->DataMethod == METHOD_IN_DIRECT) ||
            (Irp->DataMethod == METHOD_OUT_DIRECT) ) {
                return;
                }


        // Check if driver must write to user space (by READing from kernel)

        if ( _IOC_DIR(Command) & _IOC_READ )
                OsCopyToUser( Irp->OutBuffer, OutBuffer, Irp->OutSize );
}


SRLOCAL void OsWriteFifoHistory( DEVINFO *Dev,
                                 unsigned char MuxedBits,
                                 unsigned char Expected ) {

// This function is not ready under Linux at this time

#ifdef  INCLUDE_WRITE_FIFO_HISTORY

        FILE *ErrFp;
        int   i, j, k;
        unsigned int HistoryValue;

        ErrFp = fopen( "sanity.err", "a+b" );
        if (!ErrFp)
                return;

        fprintf( ErrFp, "Fifo Sanity Error: expected 0x%02X, found 0x%02X\n",
                 Expected, MuxedBits);
        fprintf( ErrFp, "Fifo Data most recent first:\n" );

        j = Dev->FHsanityInterval;
        for ( i = 0 ; i < FH_MASK+1 ; i++ ) {
                HistoryValue = Dev->FifoHistory[Dev->FHindex] & 0xFF;
                k = Dev->FHsanityInterval - j;
                if (j == Dev->FHsanityInterval) {
                        fprintf(ErrFp,"\n%4d: (%2d) 0x%02X   sanity\n",
                                  i, k, HistoryValue);
                        j = 1;
                        }
                else {
                        fprintf(ErrFp,"%4d: (%2d) 0x%02X\n",
                                  i, k, HistoryValue);
                        j++;
                        }
                Dev->FHindex -= 1;
                Dev->FHindex &= FH_MASK;
                }



        fclose( ErrFp );

#else
        dbg1_printk("OsWriteFifoHistory currently disabled\n");

#endif // INCLUDE_WRITE_FIFO_HISTORY

        return;
}

#define OsModuleUseCountInc  // was MOD_INC_USE_COUNT before 2.6
#define OsModuleUseCountDec  // was MOD_DEC_USE_COUNT before 2.6

// Atomic operation functions

#define OsAtomicSet( pAtom, Value )  atomic_set( (pAtom), (Value) )
#define OsAtomicInc( pAtom )         atomic_inc( (pAtom) )
#define OsAtomicDec( pAtom )         atomic_dec( (pAtom) )
#define OsAtomicAdd( pAtom, Amount ) atomic_add( (Amount), (pAtom) )
#define OsAtomicSub( pAtom, Amount ) atomic_sub( (Amount), (pAtom) )
#define OsAtomicRead( pAtom )        atomic_read( (pAtom) )

// Spinlock functions

#define OsSpinLockAcquire( pLock, OldIrqlOrFlags ) \
        OldIrqlOrFlags = 1; spin_lock_bh( (pLock) );
#define OsSpinLockRelease( pLock, OldIrqlOrFlags ) \
        spin_unlock_bh( (pLock) );

#define OsSpinLockAcquireInt( pLock, OldIrqlOrFlags ) \
        spin_lock_irqsave( (pLock), (OldIrqlOrFlags) );
#define OsSpinLockReleaseInt( pLock, OldIrqlOrFlags ) \
        spin_unlock_irqrestore( (pLock), (OldIrqlOrFlags) );


#endif // SROS_xxxxx  End of OS dependent helper functions section








////////////////////////////////////////////////////////////////////////
/* Group 1: PARXCH HARDWARE AND BEHAVIOR FUNCTIONS:
 *
 *
 *
 *
 *
 *
 */

/* PC PARALLEL PORT REGISTERS AND FUNCTIONS:

The parallel port has been standard PC hardware since the earliest
days of the PC.  Almost all PCs have at least one, while some have
two or even more.  A single parallel port is comprised of several
byte wide registers mapped into IO space that can be manipulated with
in and out instructions for data IO.

At power on the PC scans for parallel ports in the following address
order.  The first parallel port found is commonly referred to as
LPT1, the second as LPT2, and so on.  The base address 0x378 is by
far the most common base address for systems with a single port.

 PC power on parallel port base addresses scan order

    0x3BC
    0x378 <<< LPT1 on most machines
    0x278 <<< LPT2 on most machines
    0x268
    0x26C
    0x27C


In the days of good old DOS LPT assigments are stored at 0:408.  Even
though Windows machines are virtualized, those locations can still be
examined by running debug in a DOS window to see what the LPT
addresses are.  The 0x3BC address was the parallel port on the old
Monochrome graphics card.

Early ports were designed primarily for data output only.  Later
modifications added data input and several other features.  IEEE has
now standardized the various operating modes into the IEEE 1284 spec.
The PARXCH board will operate with either the 1284 BPP (PS/2
bidirectional parallel port byte wide mode), or 1284 EPP (enhanced
parallel port mode).  You can also reach these modes as sub-modes of
the 1284 ECP (extended capabilities port mode).  EPP transfers are
faster, relieving the x86 cpu of some of the signaling chores.
Almost all PCs support BPP mode.  And, almost all recent vintage PCs
also support EPP.  Use EPP if it is available.


The basic PC parallel port has three byte wide registers.  These
registers and their layouts are the same across all IEEE 1284
operating modes.  They are:

 Basic parallel port registers

    Data register = PortBaseAddress + 0
    Stat register = PortBaseAddress + 1
    Ctrl register = PortBaseAddress + 2

The parallel port may have additional registers beyond these three
basic ones in the IEEE 1284 enhanced modes.  See the functions listed
further on down for the additional drivers that are required for BPP
or EPP by the PARXCH.  The following paragraphs describe the layout
and signal mapping for each of the basic registers.



>>> DATA REGISTER / BYTE WIDE / PORTBASEADDRESS + 0 / READ & WRITE:

This register was primarily intended on the original PC parallel port
to be write only.  The output was latched and could be read back with
the same data being read back that had just been latched.  The BPP
mode (aka PS2) enhanced this so that depending on the setting of the
direction bit in the control register, independent data could be read
back rather than the latched output.


>>> STATUS REGISTER / BYTE WIDE / PORTBASEADDRESS + 1 / READ ONLY:

This is a read only register designed to return status information
from a device attached to the parallel port.  Register bit to DB25 pin
to PARXCH signal mapping is:

Stat reg bit 0    NA             NA                      NA
Stat reg bit 1    NA             NA                      NA
Stat reg bit 2    NA             NA                      NA
Stat reg bit 3    DB25 pin 15    Parallel Error          PARXCH PERROR
Stat reg bit 4    DB25 pin 13    Parallel Select         PARXCH PS1
Stat reg bit 5    DB25 pin 12    Parallel Paper Empty    PARXCH PS0
Stat reg bit 6    DB25 pin 10    Parallel Acknowledge    PARXCH PIRQ
Stat reg bit 7    DB25 pin 11    Parallel Busy           PARXCH PWAIT

DB25 pin 10, Parallel Acknowledge, is also connected to IRQ7 on the
ISA bus via a tri state buffer that can be enabled in the parallel
port control register.

These bits may be variously inverted or not inverted from the DB25
pin senses.  The relationships are:

PARXCH PERROR high    DB25 pin 15 high      Stat reg bit 3 high
PARXCH PS1 high       DB25 pin 13 high      Stat reg bit 4 high
PARXCH PS0 high       DB25 pin 12 high      Stat reg bit 5 high
PARXCH PIRQ high      DB25 pin 10 high      Stat reg bit 6 high
PARXCH PWAIT high     DB25 pin 11 high      Stat reg bit 7 low



>>> CONTROL REGISTER, BYTE WIDE / PORTBASEADDRESS + 2 / READ & WRITE:

This register is primarily a write only register designed to send
control signals to a device attached to the parallel port.  The
signal values are latched and can be read back, where the read back
values are the same as the ones last programmed.  Register bit to
DB25 pin to PARXCH signal mapping is:

Ctrl reg bit 0    DB25 pin 1     Parallel Strobe         PARXCH PWRITE
Ctrl reg bit 1    DB25 pin 14    Parallel Auto LF        PARXCH PDS
Ctrl reg bit 2    DB25 pin 16    Parallel Initialize     PARXCH PINIT
Ctrl reg bit 3    DB25 pin 17    Parallel Select In      PARXCH PAS
Ctrl reg bit 4    NA             Parallel Irq enable     NA
Ctrl reg bit 5    NA             Parallel Direction      NA
Ctrl reg bit 6    NA             NA                      NA
Ctrl reg bit 7    NA             NA                      NA

These bits may be variously inverted or not inverted from the DB25
pin senses.  The relationships are:

Ctrl reg bit 0 high    DB25 pin 1 low     PARXCH PWRITE low
Ctrl reg bit 1 high    DB25 pin 14 low    PARXCH PDS low
Ctrl reg bit 2 high    DB25 pin 16 high   PARXCH PINIT high
Ctrl reg bit 3 high    DB25 pin 17 low    PARXCH PAS low
Ctrl reg bit 4 high    NA                 Parallel port irq 7 is ENABLED
Ctrl reg bit 5 high    NA                 Parallel data dir is INPUT TO PC
Ctrl reg bit 6 high    NA
Ctrl reg bit 7 high    NA

Ctrl reg bits are not inverted on read back from their programmed
values.

*/


// Status register bit assignments: (with PARXCH signal names)

#define PERROR                  (1 << 3)    // DB25 pin 15
#define PS1                     (1 << 4)    // DB25 pin 13
#define PS0                     (1 << 5)    // DB25 pin 12
#define PIRQ                    (1 << 6)    // DB25 pin 10
#define PWAIT                   (1 << 7)    // DB25 pin 11

#define PARXCH_FIFO_EF_MASK          PS0      //// NOT USED
#define PARXCH_FIFO_FF_MASK          PS1      //// NOT USED
#define PARXCH_VOLTAGE_GOOD_MASK     PERROR



// Control register bit assignments and values: (with PARXCH signal names)

#define PWR                     (1 << 0)    // DB25 pin 1    << PWRITE
#define PDS                     (1 << 1)    // DB25 pin 14
#define PINIT                   (1 << 2)    // DB25 pin 16
#define PAS                     (1 << 3)    // DB25 pin 17
#define PIRQEN                  (1 << 4)    // not on DB25
#define PDIR                    (1 << 5)    // not on DB25

#define PWR_READ                (0 << 0)
#define PWR_WRITE               (1 << 0)

#define PDS_NEGATE              (0 << 1)
#define PDS_ASSERT              (1 << 1)

#define PINIT_POWERDOWN         (0 << 2)
#define PINIT_POWERUP           (1 << 2)

#define PAS_NEGATE              (0 << 3)
#define PAS_ASSERT              (1 << 3)

#define PIRQ_DISABLE            (0 << 4)
#define PIRQ_ENABLE             (1 << 4)

#define PDIR_WRITE              (0 << 5)
#define PDIR_READ               (1 << 5)




// ECP extended control register bit assignments and values:

#define ECP_FIFO_EMPTY          (1 << 0)  // Read Only
#define ECP_FIFO_FULL           (1 << 1)  // Read Only
#define ECP_SVCINTR             (1 << 2)
#define ECP_DMA                 (1 << 3)
#define ECP_ERRINTR             (1 << 4)
#define ECP_SUBMODE             (7 << 5)  // 3 bits

#define ECP_SVCINTR_ENABLE      (0 << 2)
#define ECP_SVCINTR_DISABLE     (1 << 2)

#define ECP_DMA_DISABLE         (0 << 3)
#define ECP_DMA_ENABLE          (1 << 3)

#define ECP_ERRINTR_ENABLE      (0 << 4)
#define ECP_ERRINTR_DISABLE     (1 << 4)

#define ECP_SUBMODE_SPP         (0 << 5) // 000 SPP original
#define ECP_SUBMODE_PS2         (1 << 5) // 001 PS/2 byte, bidirectional
#define ECP_SUBMODE_FAST        (2 << 5) // 010 Fast Centronics (SPP)
#define ECP_SUBMODE_ECP         (3 << 5) // 011 ECP
#define ECP_SUBMODE_EPP         (4 << 5) // 100 EPP
#define ECP_SUBMODE_RESERVED    (5 << 5) // 101 Reserved
#define ECP_SUBMODE_TEST        (6 << 5) // 110 Test, no handshaking
#define ECP_SUBMODE_CONFIG      (7 << 5) // 111 Configuration




SRLOCAL void ParXch_ParPortInit( DEVINFO *Dev ) {

        dbg3_printk( "ParXch_ParPortInit\n" );


        // Set up port addresses.

        Dev->ParPortData  = (Dev->ParPortBase + 0);
        Dev->ParPortStat  = (Dev->ParPortBase + 1);
        Dev->ParPortCtrl  = (Dev->ParPortBase + 2);
        Dev->EppPortAddr  = (Dev->ParPortBase + 3);
        Dev->EppPortData  = (Dev->ParPortBase + 4);
        Dev->EcpPortFifo  = (Dev->ParPortBase + Dev->ParPortEcpOffset + 0);
        Dev->EcpPortCfgb  = (Dev->ParPortBase + Dev->ParPortEcpOffset + 1);
        Dev->EcpPortExcr  = (Dev->ParPortBase + Dev->ParPortEcpOffset + 2);



        // Point to EPP/BPP read/write functions according to the mode.

        if ( (Dev->ParPortMode == PARXCH_PORT_MODE_EPP)     ||
             (Dev->ParPortMode == PARXCH_PORT_MODE_ECP_EPP) ) {

                Dev->ParPortAsRd = ParXch_EppAsRd;
                Dev->ParPortAsWr = ParXch_EppAsWr;
                Dev->ParPortDsRd = ParXch_EppDsRd;
                Dev->ParPortDsWr = ParXch_EppDsWr;
                }

       else { // (Dev->ParPortMode == PARXCH_PORT_MODE_BPP)    ||
              // (Dev->ParPortMode == PARXCH_PORT_MODE_ECP_BPP)

                Dev->ParPortAsRd = ParXch_BppAsRd;
                Dev->ParPortAsWr = ParXch_BppAsWr;
                Dev->ParPortDsRd = ParXch_BppDsRd;
                Dev->ParPortDsWr = ParXch_BppDsWr;
                }

}
SRLOCAL void ParXch_EcpPortInit( DEVINFO *Dev ) {

        unsigned char EcrValue;

        dbg3_printk( "ParXch_EcpPortInit\n" );


        // Quick exit when not using ECP

        if (Dev->ParPortMode == PARXCH_PORT_MODE_BPP ||
            Dev->ParPortMode == PARXCH_PORT_MODE_EPP )
                return;


        // Read in and save original ECR value.

        OsPortIn( Dev->EcpPortExcr, &Dev->EcpPortSaveEcrValue );


        // Set extended control register value, except for submode.

        EcrValue = ECP_SVCINTR_DISABLE
                 | ECP_DMA_DISABLE
                 | ECP_ERRINTR_DISABLE;


        // Add mode and write it out.

        if (Dev->ParPortMode == PARXCH_PORT_MODE_ECP_BPP) {

                EcrValue |= ECP_SUBMODE_PS2;
                OsPortOut( Dev->EcpPortExcr, EcrValue );
                }


        else if (Dev->ParPortMode == PARXCH_PORT_MODE_ECP_EPP) {

                EcrValue |= ECP_SUBMODE_EPP;
                OsPortOut( Dev->EcpPortExcr, EcrValue );
                }


        else
                ; // Do nothing when not in ECP mode
}

SRLOCAL void ParXch_EcpPortRestore( DEVINFO *Dev ) {

        dbg3_printk( "ParXch_EcpPortRestore\n" );

        // Quick exit when not using ECP

        if (Dev->ParPortMode == PARXCH_PORT_MODE_BPP ||
            Dev->ParPortMode == PARXCH_PORT_MODE_EPP )
                return;


        // Restore the saved ECP parallel port ECR value.

        OsPortOut( Dev->EcpPortExcr, Dev->EcpPortSaveEcrValue );
}

SRLOCAL void ParXch_EppPortClearTimeout( DEVINFO *Dev ) {

        unsigned char temp;

        dbg2_printk( "EppPortClearTimeout\n" );

        // Make sure any EPP timeout bits are cleared.

        ParXch_ParPortStatWr( Dev, 0x01 );      // SMC '665
        ParXch_ParPortStatWr( Dev, 0x00 );      // odd-ball generics

        ParXch_ParPortStatRd( Dev, &temp );     // National Semi
        ParXch_ParPortStatRd( Dev, &temp );

}

SRLOCAL void ParXch_ParPortDataRd( DEVINFO *Dev, unsigned char *Value ) {

        dbg5_printk( "ParXch_ParPortDataRd\n" );


        // Read from the parallel port data register.

        OsPortIn( Dev->ParPortData, Value );


}
SRLOCAL void ParXch_ParPortDataWr( DEVINFO *Dev, unsigned char Value ) {

        dbg5_printk( "ParXch_ParPortDataWr\n" );


        // Write to the parallel port data register.

        OsPortOut( Dev->ParPortData, Value );
}

SRLOCAL void ParXch_ParPortStatRd( DEVINFO *Dev, unsigned char *Value ) {

        dbg5_printk( "ParXch_ParPortStatRd\n" );


        // Read from the parallel port status register.

        OsPortIn( Dev->ParPortStat, Value );
}

SRLOCAL void ParXch_ParPortStatWr( DEVINFO *Dev, unsigned char Value ) {

        dbg5_printk( "ParXch_ParPortStatWr\n" );


        // Write to the parallel port status register.

        OsPortOut( Dev->ParPortStat, Value );

}
SRLOCAL void ParXch_ParPortCtrlWr( DEVINFO *Dev, unsigned char Value ) {

        dbg5_printk( "ParXch_ParPortCtrlWr\n" );


        // Write to the parallel port control register.

        Dev->ParPortCtrlValue = Value;

        OsPortOut( Dev->ParPortCtrl, Value );
}

SRLOCAL void ParXch_ParPortCtrlRw( DEVINFO *Dev, unsigned char Value ) {

        dbg5_printk( "ParXch_ParPortCtrlRw\n" );


        // Set the parallel port control register read/write/as/ds bits.

        Dev->ParPortCtrlValue &= ~(PWR | PDS | PAS | PDIR);
        Dev->ParPortCtrlValue |= (Value & (PWR | PDS | PAS | PDIR));

        OsPortOut( Dev->ParPortCtrl, Dev->ParPortCtrlValue );

}
SRLOCAL void ParXch_ParPortCtrlIrqen( DEVINFO *Dev, unsigned char Value ) {

        dbg5_printk( "ParXch_ParPortCtrlIrqen\n" );


        // Set the parallel port control register PIRQEN bit.

        Dev->ParPortCtrlValue &= ~PIRQEN;
        Dev->ParPortCtrlValue |= (Value & PIRQEN);

        OsPortOut( Dev->ParPortCtrl, Dev->ParPortCtrlValue );

}
SRLOCAL void ParXch_ParPortCtrlDirRead( DEVINFO *Dev) {

        dbg5_printk( "ParXch_ParPortCtrlDirRead\n" );

        // Set the parallel port control register PDIR and PWR bits for reading

        Dev->ParPortCtrlValue &= ~(PWR | PDIR);
        Dev->ParPortCtrlValue |= (PWR_READ | PDIR_READ);

        OsPortOut( Dev->ParPortCtrl, Dev->ParPortCtrlValue );
}
SRLOCAL void ParXch_ParPortCtrlDirWrite( DEVINFO *Dev) {

        dbg5_printk( "ParXch_ParPortCtrlDirWrite\n" );

        // Set the parallel port control register PDIR and PWR bits for writing

        Dev->ParPortCtrlValue &= ~(PWR | PDIR);
        Dev->ParPortCtrlValue |= (PWR_WRITE | PDIR_WRITE);

        OsPortOut( Dev->ParPortCtrl, Dev->ParPortCtrlValue );
}





/* WAIT FOR A SPECIFIED AMOUNT OF TIME:

Operations like PowerDown and PowerUp require waiting for a period of
milliseconds.  Their wait periods go through the ParXch_Wait function.

The wait is done by simply executing x86 in instructions.  An in
instruction from the ISA bus takes approximately 1us.  The wait
periods for power up and down are not super critical.  Another
mechanism like calling sleep could be used if preferred.

*/

SRLOCAL void ParXch_Wait( DEVINFO *Dev, unsigned milliseconds ) {

        long microseconds;
        unsigned char value;

        dbg3_printk( "ParXch_Wait\n" );


        // Wait for the specified number of milliseconds:

        while ( milliseconds-- ) {

                microseconds = 1000; // 1000 us = 1 ms

                while ( microseconds-- )

                        ParXch_ParPortStatRd( Dev, &value); // ~1 us
                }

        return;
}




/* POWER GOOD:

Whether the PARXCH on board power supply voltage is good or too low
is indicated by the PERROR bit on the parallel port status register.
This signal is active low and, if good, the on board voltage is with
within 1% of 4.75 volts.

This bit is pulled up by a resistor on the PC side of the port.

*/

SRLOCAL int ParXch_VoltageGood( DEVINFO *Dev ) {

        unsigned char Status;

        dbg4_printk( "ParXch_VoltageGood\n" );



        // Get the voltage good bit from the status register.

        ParXch_ParPortStatRd( Dev, &Status );

        if ( (Status & PERROR) == Dev->VoltageIsGood )
                return( 1 );
        else
                return( 0 );
}




/* POWER DOWN AND UP:

The PC can power down and up the PARXCH board via the parallel port
control register bit PINIT.

Power down and up are not only useful for turning off the PARXCH when
conversion isn't needed, they are also the only way to ensure an
overall reset of the PARXCH board.  A PowerDown and PowerUp cycle is
automatically done in ParXch_Init.

Also note that turning off the PC will automatically force a PARXCH
power down, even if its wall transformer is still plugged in.

Finally, the PARXCH cannot power up and down instantly.  It takes
approx 300 to 900 ms to settle to power up or down respectively.
Settling time is programed into PowerDown and PowerUp by calling the
ParXch_Wait function.

*/

#define POWER_DOWN_WAIT         900 // milliseconds
#define POWER_UP_WAIT           300 // milliseconds

SRLOCAL void ParXch_TurnOff( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_TurnOff\n" );

        // Turn off the parallel port

        ParXch_ParPortCtrlWr( Dev,   // drive all DB25 rw pins low ...

                              PWR_WRITE
                            | PDS_ASSERT
                            | PINIT_POWERDOWN
                            | PAS_ASSERT
                            | PIRQ_DISABLE
                            | PDIR_WRITE

                            );
}

SRLOCAL void ParXch_TurnOn( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_TurnOn\n" );

        // Turn on the parallel port
        //
        // ** Other ctrl bits set for subsequent EPP/BPP operations. **

        ParXch_ParPortCtrlWr( Dev,

                              PWR_READ
                            | PDS_NEGATE
                            | PINIT_POWERUP
                            | PAS_NEGATE
                            | PIRQ_DISABLE
                            | PDIR_READ

                             );
}

SRLOCAL int ParXch_PowerDown( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_PowerDown\n" );


        // Program the control register to power down the PARXCH:
        //
        // Setting the PINIT bit on the parallel port control register
        // to zero powers down the PARXCH.  Other bits on the parallel
        // port registers are also programmed to minimize PARXCH
        // power consumption while powered down.


        // Set the parallel port data bits to zero.

        ParXch_ParPortDataWr( Dev, 0x00 );


        // Power down.

        ParXch_TurnOff( Dev );


        // Wait for the power down settling time.

        ParXch_Wait( Dev, POWER_DOWN_WAIT );



        // Check that the PARXCH voltage good bit shows power off.
        //
        // This code is conditionally compiled in depending on the
        // PARXCH_TEST_VOLTAGE_BAD define.
        //


#if PARXCH_TEST_VOLTAGE_BAD == 1


        if ( !ParXch_VoltageGood( Dev ) )

                // power is indeed off ... all is ok ...

                return( 1 );

        else
                // power is still up, something is wrong!

                return( 0);


#elif PARXCH_TEST_VOLTAGE_BAD == 0

        // don't bother testing for voltage good ... fake success

        return( 1 );

#else
#pragma message( "PARXCH_TEST_VOLTAGE_BAD must be defined as 1 or 0" )
#endif



}
SRLOCAL int ParXch_PowerUp( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_PowerUp\n" );


        // Program the control register to power up the PARXCH:
        //
        // Setting the PINIT bit on the parallel port control register
        // to one powers up the PARXCH.  Other bits are programmed so
        // the port is in a state ready for EPP/BPP mode.


        // Set the parallel port data bits to zero.

        ParXch_ParPortDataWr( Dev, 0x00 );



        // Power up.

        ParXch_TurnOn( Dev );



        // Make sure any EPP timeout bits are cleared.

        ParXch_EppPortClearTimeout( Dev );



        // Wait for the power up settling time.

        ParXch_Wait( Dev, POWER_UP_WAIT );



        // Check that the PARXCH voltage good bit shows power on.
        //
        // This code is conditionally compiled in depending on the
        // PARXCH_TEST_VOLTAGE_GOOD define.
        //


#if PARXCH_TEST_VOLTAGE_GOOD == 1

        if ( ParXch_VoltageGood( Dev ) )

                // then power is up and all is good ...

                return( 1 );

        else
                // power is still down, something is wrong!

                return( 0 );

#elif PARXCH_TEST_VOLTAGE_GOOD == 0

        // don't bother testing the voltage good signal ... fake success

        return( 1 );

#else
#pragma message( "PARXCH_TEST_VOLTAGE_GOOD must be defined as 1 or 0" )
#endif


}




/* ENHANCED PARALLEL PORT (EPP) READ/WRITE FUNCTIONS:

The PARXCH board is primarily designed to work with the parallel port
in IEEE 1284 EPP (enhanced parallel port) mode.  The following
functions are appropriate for accessing the port in that mode.

The PARXCH can also work with the lower performance BPP
(bidirectional parallel port) mode with appropriate read/write
functions.  The selection of EPP/BPP read/write functions should be
made by specifying the ParPortMode parameter to ParXchOpen.

Besides the automatic Write, As, and Ds strobing done in EPP mode,
those signals can still be controlled by programming the parallel
port control register.  For example, after PC power on As is often
asserted in the control register and needs to be turned off.  This is
automatically done in the PowerUp routine above.

Note:

A few machines require the parallel port control register PS/2 dir
bit, PDIR, to be toggled for reads and writes when in EPP mode.  This
is not correct according to the IEEE 1284 spec, but nevertheless is
the way some EPP ports behave.  The functions below assume the PDIR
is normally set to read, toggling it for writes.  Since the PC is
normally reading from the PARXCH, this results in little performance
hit and provides compatibility with the most machines.


Note:

When used with the SR PARGPS module, these port functions may be called
from the deferred procedure call (DPC) or bottom-half associated with
the PPS interrupt.  Therefore, any function with more than a single
in/out to the port must be protected by a spin lock to ensure that
the current operation is allowed to complete before the DPC runs.

*/

SRLOCAL void ParXch_EppAsRd( DEVINFO *Dev, unsigned char *Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Prepare direction and read from EPP AS

        ParXch_ParPortCtrlDirRead( Dev );

        OsPortIn( Dev->EppPortAddr, Value );


        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}

SRLOCAL void ParXch_EppAsWr( DEVINFO *Dev, unsigned char Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Prepare direction and write from EPP AS

        ParXch_ParPortCtrlDirWrite( Dev );

        OsPortOut( Dev->EppPortAddr, Value );


        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}

SRLOCAL void ParXch_EppDsRd( DEVINFO *Dev, unsigned char *Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Prepare direction and read from EPP DS

        ParXch_ParPortCtrlDirRead( Dev );

        OsPortIn( Dev->EppPortData, Value );


        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}

SRLOCAL void ParXch_EppDsWr( DEVINFO *Dev, unsigned char Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Prepare direction and write from EPP DS

        ParXch_ParPortCtrlDirWrite( Dev );

        OsPortOut( Dev->EppPortData, Value );


        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}




/* BIDIRECTIONAL PARALLEL PORT (BPP) READ/WRITE FUNCTIONS:

These BPP drivers work by synthesizing EPP timing in software, using
the basic parallel port data and control registers.  For the most
part this means several IO instructions are required where only a
single IO instruction suffices in EPP mode.

Note:

On some PCs (such as Toshiba), it is important that the parallel port
control register PS/2 dir bit, PDIR, be set *BEFORE* writing to the
data port.  On such machines, if the PDIR bit is set to read, the
write to the parallel port data register will be thrown away.

Note:

When used with the SR PARGPS module, these port functions may be called
from the deferred procedure call (DPC) or bottom-half associated with
the PPS interrupt.  Therefore, any function with more than a single
in/out to the port must be protected by a spin lock to ensure that
the current operation is allowed to complete before the DPC runs.

*/


// shorter names for some parallel port control register bit values ...

#define PWR_R           PWR_READ
#define PWR_W           PWR_WRITE

#define PDS_A           PDS_ASSERT
#define PDS_N           PDS_NEGATE

#define PAS_A           PAS_ASSERT
#define PAS_N           PAS_NEGATE

#define PDIR_R          PDIR_READ
#define PDIR_W          PDIR_WRITE


SRLOCAL void ParXch_BppAsRd( DEVINFO *Dev, unsigned char *Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Synthesize EPP AS read ...

        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_A | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_A | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_A | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_A | PDIR_R );

            ParXch_ParPortDataRd( Dev, Value );

        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );


        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}
SRLOCAL void ParXch_BppAsWr( DEVINFO *Dev, unsigned char Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Synthesize EPP AS write ...

        ParXch_ParPortCtrlRw( Dev, PWR_W | PDS_N | PAS_N | PDIR_W );

            ParXch_ParPortDataWr( Dev, Value );

        ParXch_ParPortCtrlRw( Dev, PWR_W | PDS_N | PAS_A | PDIR_W );
        ParXch_ParPortCtrlRw( Dev, PWR_W | PDS_N | PAS_A | PDIR_W );

        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );

        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}
SRLOCAL void ParXch_BppDsRd( DEVINFO *Dev, unsigned char *Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Synthesize EPP DS read ...

        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_A | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_A | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_A | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_A | PAS_N | PDIR_R );

            ParXch_ParPortDataRd( Dev, Value );

        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );

        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}
SRLOCAL void ParXch_BppDsWr( DEVINFO *Dev, unsigned char Value ) {

        KIRQL OldIrqlOrFlags;

        OsSpinLockAcquire( &Dev->PortLock, OldIrqlOrFlags );


        // Synthesize EPP DS write ...

        ParXch_ParPortCtrlRw( Dev, PWR_W | PDS_N | PAS_N | PDIR_W );

            ParXch_ParPortDataWr( Dev, Value );

        ParXch_ParPortCtrlRw( Dev, PWR_W | PDS_A | PAS_N | PDIR_W );
        ParXch_ParPortCtrlRw( Dev, PWR_W | PDS_A | PAS_N | PDIR_W );

        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );
        ParXch_ParPortCtrlRw( Dev, PWR_R | PDS_N | PAS_N | PDIR_R );

        OsSpinLockRelease( &Dev->PortLock, OldIrqlOrFlags );
}








/* PARXCH BOARD LEVEL CONTROL REGISTER, BdCtrl:

Primary control of the PARXCH from the PC is done via the byte wide
board control register, BdCtrl.  This write only register is mapped
into EPP AsWr.  The physical register is inside the XCR5128 CPLD on
the PARXCH board.

Since the BdCtrl register is write only, the value of its current bit
settings are saved in the variable BdCtrlValue.  See the following
defines for the various bit assigments.

*/


// PARXCH BdCtrl register bit assignments and values:

#define BDCTRL_MODE0                    (1 << 0)
#define BDCTRL_MODE1                    (1 << 1)
#define BDCTRL_DRDYIRQREQ               (1 << 2)
#define BDCTRL_SCLK                     (1 << 3)
#define BDCTRL_SDI                      (1 << 4)
#define BDCTRL_DSYNC                    (1 << 5)
#define BDCTRL_FIFORESET                (1 << 6)
#define BDCTRL_USERLED                  (1 << 7)

#define BDCTRL_MODE_HALT_INIT           (0x0 << 0)
#define BDCTRL_MODE_HALT_UDIO           (0x1 << 0)
#define BDCTRL_MODE_RUN_EF              (0x2 << 0)   // ef = empty flag
#define BDCTRL_MODE_RUN_BF              (0x3 << 0)   // bf = buffer ready

#define BDCTRL_DRDYIRQREQ_LOW           (0x0 << 2)
#define BDCTRL_DRDYIRQREQ_HIGH          (0x1 << 2)

#define BDCTRL_SCLK_LOW                 (0x0 << 3)
#define BDCTRL_SCLK_HIGH                (0x1 << 3)

#define BDCTRL_SDI_LOW                  (0x0 << 4)
#define BDCTRL_SDI_HIGH                 (0x1 << 4)

#define BDCTRL_DSYNC_LOW                (0x0 << 5)   // modulators halted
#define BDCTRL_DSYNC_HIGH               (0x1 << 5)   // modulators running

#define BDCTRL_FIFORESET_LOW            (0x0 << 6)
#define BDCTRL_FIFORESET_HIGH           (0x1 << 6)

#define BDCTRL_USERLED_OFF              (0x0 << 7)
#define BDCTRL_USERLED_ON               (0x1 << 7)



SRLOCAL void ParXch_BdCtrlWr( DEVINFO *Dev, unsigned char Value ) {

        dbg5_printk( "ParXch_BdCtrlWr 0x%X\n", (Value & 0x000000FF) );


        // Write a value to the PARXCH board control register.

        Dev->ParPortAsWr( Dev, Value );


}
SRLOCAL void ParXch_BdCtrlUnlock( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_BdCtrlUnlock\n" );


        // Unlock the parallel port buffer on the PARXCH board.

        Dev->BdCtrlValue = 0x70;  ParXch_BppAsWr( Dev, Dev->BdCtrlValue );
        Dev->BdCtrlValue = 0x60;  ParXch_BppAsWr( Dev, Dev->BdCtrlValue );
        Dev->BdCtrlValue = 0x30;  ParXch_BppAsWr( Dev, Dev->BdCtrlValue );
        Dev->BdCtrlValue = 0x20;  ParXch_BppAsWr( Dev, Dev->BdCtrlValue );

}
SRLOCAL void ParXch_BdCtrlInit( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_BdCtrlInit\n" );


        // Initialize the PARXCH BdCtrl register.

        Dev->BdCtrlValue =

                    BDCTRL_MODE_HALT_INIT
                  | BDCTRL_DRDYIRQREQ_LOW       // no drdy irq request
                  | BDCTRL_SCLK_LOW             // sclk idles low
                  | BDCTRL_SDI_LOW              // sdi = no care
                  | BDCTRL_DSYNC_HIGH           // modulators running
                  | BDCTRL_FIFORESET_LOW        // leave FIFO alone
                  | BDCTRL_USERLED_OFF          // yellow led off

                  ;


        ParXch_BdCtrlWr( Dev, Dev->BdCtrlValue );


}
SRLOCAL void ParXch_BdCtrlMode( DEVINFO *Dev, unsigned char Value ) {

        dbg4_printk( "ParXch_BdCtrlMode\n" );


        // Set the PARXCH BdCtrlMode bits:
        //
        // These bits control the operating mode of the PARXCH.  When
        // the mode is HALT_INIT, the PC has control of the sclk and sdi
        // pins on the ADS1210s.  All other modes are controlled by the
        // XCR5128.
        //

        Dev->BdCtrlValue &= ~(BDCTRL_MODE0 | BDCTRL_MODE1);
        Dev->BdCtrlValue |= (Value & (BDCTRL_MODE0 | BDCTRL_MODE1));

        ParXch_BdCtrlWr( Dev, Dev->BdCtrlValue );

        return;


}
SRLOCAL void ParXch_BdCtrlDrdyIrqReq( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_BdCtrlDrdyIrqReq\n" );


        // Set the BdCtrl DrdyIrqReq pin high.  This indicates that the
        // next Dready signal is to cause a parallel port interrupt to
        // occur.  This pin is automatically reset by the PARxCH on
        // board controller after causing the interrupt, so the shadow
        // value in Dev->BdCtrlValue is reset here before return.

        Dev->BdCtrlValue |= BDCTRL_DRDYIRQREQ_HIGH;

        ParXch_BdCtrlWr( Dev, Dev->BdCtrlValue);

        Dev->BdCtrlValue &= ~BDCTRL_DRDYIRQREQ;

        return;

}
SRLOCAL void ParXch_BdCtrlFifoReset( DEVINFO *Dev, unsigned char Value ) {

        dbg4_printk( "ParXch_BdCtrlFifoReset\n" );


        // Set the BdCtrl Fifo Reset bit high or low.
        //
        // When the Fifo Reset bit is high, all addresses are held
        // at zero.  Low and the Fifo operates normally.
        //
        //

        Dev->BdCtrlValue &= ~BDCTRL_FIFORESET;
        Dev->BdCtrlValue |= (Value & BDCTRL_FIFORESET);

        ParXch_BdCtrlWr( Dev, Dev->BdCtrlValue );

        return;

}
SRLOCAL void ParXch_BdCtrlUserLed( DEVINFO *Dev, int State ) {

        dbg4_printk( "ParXch_BdCtrlUserLed\n" );


        // Set the BdCtrl User Led bit on or off.

        Dev->BdCtrlValue &= ~BDCTRL_USERLED;

        if ( State )

                Dev->BdCtrlValue |= BDCTRL_USERLED_ON;
        else
                Dev->BdCtrlValue |= BDCTRL_USERLED_OFF;


        ParXch_BdCtrlWr( Dev, Dev->BdCtrlValue );

        return;
}







/* ADS1210 instruction register bit assignments:

These defines give the bit assigments and values of the internal
ADS1210 registers.  They are used in the ADS1210_xxx functions below.

*/

#define ADS1210_WRITE                   (0x00 << 7)
#define ADS1210_READ                    (0x01 << 7)

#define ADS1210_NBYTES_1                (0x00 << 5)
#define ADS1210_NBYTES_2                (0x01 << 5)
#define ADS1210_NBYTES_3                (0x02 << 5)
#define ADS1210_NBYTES_4                (0x03 << 5)

#define ADS1210_ADDR_DATA               (0x00 << 0)
#define ADS1210_ADDR_CONTROL            (0x04 << 0)
#define ADS1210_ADDR_OFFSET             (0x08 << 0)
#define ADS1210_ADDR_FSCAL              (0x0C << 0)

#define ADS1210_ADDR_CONTROL3           (0x04 << 0)
#define ADS1210_ADDR_CONTROL2           (0x05 << 0)
#define ADS1210_ADDR_CONTROL1           (0x06 << 0)
#define ADS1210_ADDR_CONTROL0           (0x07 << 0)


// ADS1210 command register bit assignments:

#define ADS1210_BIAS_GEN_OFF            (0x0L << 31)
#define ADS1210_BIAS_GEN_ON             (0x1L << 31)

#define ADS1210_INT_REF_OFF             (0x0L << 30)
#define ADS1210_INT_REF_ON              (0x1L << 30)

#define ADS1210_DATA_FORMAT_SIGNED      (0x0L << 29)
#define ADS1210_DATA_FORMAT_OFFSET      (0x1L << 29)

#define ADS1210_DATA_FORMAT( VALUE )    ( ((unsigned long)VALUE & 0x1L) << 29 )

#define ADS1210_BIPOLAR                 (0x0L << 28)
#define ADS1210_UNIPOLAR                (0x1L << 28)

#define ADS1210_BYTE_DESCEND            (0x0L << 27)
#define ADS1210_BYTE_ASCEND             (0x1L << 27)

#define ADS1210_MSB_FIRST               (0x0L << 26)
#define ADS1210_MSB_LAST                (0x1L << 26)

#define ADS1210_SDO_DISABLED            (0x0L << 25)
#define ADS1210_SDO_ENABLED             (0x1L << 25)

#define ADS1210_DRDY                    (0x1L << 24)  // << read  only
#define ADS1210_DSYNC_NOCHANGE          (0x0L << 24)  // << write only
#define ADS1210_DSYNC_ASSERT            (0x1L << 24)  // << write only

#define ADS1210_CAL_MODE_NORMAL         (0x0L << 21)
#define ADS1210_CAL_MODE_SELF           (0x1L << 21)
#define ADS1210_CAL_MODE_SYS_OFF        (0x2L << 21)
#define ADS1210_CAL_MODE_SYS_FULL       (0x3L << 21)
#define ADS1210_CAL_MODE_SYS_PSEUDO     (0x4L << 21)
#define ADS1210_CAL_MODE_BACK           (0x5L << 21)
#define ADS1210_CAL_MODE_SLEEP          (0x6L << 21)

#define ADS1210_CAL_MODE( VALUE )       ( ((unsigned long)VALUE & 0x7L) << 21 )

#define ADS1210_GAIN_1                  (0x0L << 18)
#define ADS1210_GAIN_2                  (0x1L << 18)
#define ADS1210_GAIN_4                  (0x2L << 18)
#define ADS1210_GAIN_8                  (0x3L << 18)
#define ADS1210_GAIN_16                 (0x4L << 18)

#define ADS1210_GAIN( VALUE )           ( ((unsigned long)VALUE & 0x7L) << 18 )

#define ADS1210_CHAN_SELECT_0           (0x0L << 16)

#define ADS1210_TURBO_1                 (0x0L << 13)
#define ADS1210_TURBO_2                 (0x1L << 13)
#define ADS1210_TURBO_4                 (0x2L << 13)
#define ADS1210_TURBO_8                 (0x3L << 13)
#define ADS1210_TURBO_16                (0x4L << 13)

#define ADS1210_TURBO( VALUE )          ( ((unsigned long)VALUE & 0x7L) << 13 )
#define ADS1210_DECIMATION( VALUE )     (((unsigned long)VALUE & 0x1FFFL) << 0)






/* ADS1210 SCLK, SDI READ/WRITE BYTE PIN FUNCTIONS:

These helper functions twiddle the ADS1210 sclk, and sdi pins via the
BdCtrl register to read and write single bytes from the ADS1210.

They are designed to be used when the BDCTRL_RUN bit is off.  When the
run bit is on the XCR5128 has control of the A/D converters.

*/

SRLOCAL void ParXch_ADS1210_WriteByte( DEVINFO *Dev, unsigned char Value ) {

        unsigned char CtrlReg;
        int Bit;


        dbg4_printk( "ParXch_ADS1210_WriteByte\n" );

        CtrlReg  = Dev->BdCtrlValue;


        // Write a byte to the ADS1210 via the PARXCH BdCtrl register.
        //
        // A single byte is written MSB first, a bit at a time.  Since
        // the sdi pin is in parallel on all the converters, the same
        // byte is written to each channel.
        //
        // See the ADS1210 spec sheet for the sclk, sdi timing on write
        // operations.
        //

        for ( Bit = 0x80 ; Bit > 0 ; Bit >>= 1 ) {


                // Drive sclk high and assert the data bit.

                CtrlReg |= BDCTRL_SCLK;

                if ( Value & Bit )

                        CtrlReg |= BDCTRL_SDI;
                else
                        CtrlReg &= ~BDCTRL_SDI;

                ParXch_BdCtrlWr( Dev, CtrlReg );



                // Drive sclk low to latch it into the ADS1210.

                CtrlReg &= ~BDCTRL_SCLK;

                ParXch_BdCtrlWr( Dev, CtrlReg );

                }

        Dev->BdCtrlValue = CtrlReg;

        return;

}
SRLOCAL void ParXch_ADS1210_ReadByte( DEVINFO *Dev, unsigned char *Value ) {

        unsigned char CtrlReg, PortValue;
        int Bit;


        dbg4_printk( "ParXch_ADS1210_ReadByte\n" );

        CtrlReg = Dev->BdCtrlValue;


        // Read a byte from the ADS1210 via the PARXCH BdCtrl register.
        //
        // When the BDCTRL_RUN bit is off, the sdo pins from each
        // converter are mapped into the low nibble of EPP AsRd.
        //
        // See the ADS1210 spec sheet for the sclk, sdo timing on read
        // operations.
        //


        *Value = 0;

        for ( Bit = 0x80 ; Bit > 0 ; Bit >>= 1 ) {

                // Toggle the ADS1210 sclk high.

                CtrlReg |= BDCTRL_SCLK;

                ParXch_BdCtrlWr( Dev, CtrlReg );



                // Read the bit.

                Dev->ParPortAsRd( Dev, &PortValue );



                // TEMPORARILY JUST READ BACK A/D ON CHANNEL 0.
                //
                // WHY NOT EXPAND TO GET BACK ALL 4 CHANNELS ????
                // WHY NOT EXPAND TO GET BACK ALL 4 CHANNELS ????
                // WHY NOT EXPAND TO GET BACK ALL 4 CHANNELS ????
                //

                if ( PortValue & 0x01 )

                        *Value |= Bit;
                else
                        *Value &= ~Bit;


                // Toggle the ADS1210 sclk low for next cycle.

                CtrlReg &= ~BDCTRL_SCLK;

                ParXch_BdCtrlWr( Dev, CtrlReg );

                }

        Dev->BdCtrlValue = CtrlReg;

        return;
}





/* ADS1210 INTERNAL REGISTER READ and WRITE:

These functions read and write the ADS1210 internal registers.

They should be used only when the board is not executing and the
BdCtrl RUN bit is off.  Calling them while the ADS1210s are executing
will corrupt board operation.

See the Burr Brown spec sheet for details on the ADS1210 register bit
assigments.

*/

SRLOCAL void ParXch_ADS1210_ReadReg8( DEVINFO *Dev, int Reg, unsigned char *Value ) {

        unsigned char Inst;

        dbg4_printk( "ParXch_ADS1210_ReadReg8\n" );


        // Read an ADS1210 8 bit register.


        Inst = ADS1210_READ | ADS1210_NBYTES_1 | Reg;


        ParXch_ADS1210_WriteByte( Dev, Inst );

                *Value = 0;

                ParXch_ADS1210_ReadByte( Dev, Value );
}

SRLOCAL void ParXch_ADS1210_WriteReg8( DEVINFO *Dev, int Reg, unsigned char Value ) {

        unsigned char Inst;

        dbg4_printk( "ParXch_ADS1210_WriteReg8\n" );


        // Write to an ADS1210 8 bit register.

        Inst = ADS1210_WRITE | ADS1210_NBYTES_1 | Reg;

        ParXch_ADS1210_WriteByte( Dev, Inst );

                ParXch_ADS1210_WriteByte( Dev, Value );
}







/* BASIC DRAM FIFO FUNCTIONS:

These helper functions communicate with the PARXCH DRAM FIFO.

*/

// The FIFO sanity data now contains both the sanity value and
// a GPS mark bit.

#define FIFO_SANITY_MASK    0xFE   // b 1111 1110
#define FIFO_SANITY_0       0x00
#define FIFO_SANITY_1       0xFE

#define FIFO_MARK_MASK      0x01   // b 0000 0001
#define FIFO_MARK_NO        0x00
#define FIFO_MARK_YES       0x01


SRLOCAL void ParXch_DramFifoReset( DEVINFO *Dev ) {

        dbg4_printk( "ParXch_DramFifoReset\n" );

        // Reset the DRAM FIFO read and write addresses and sanity bit:
        //
        // The BdCtrl run bit should be off and ADS1210 drdy deasserted.
        //

        ParXch_BdCtrlFifoReset( Dev, BDCTRL_FIFORESET_HIGH );
        ParXch_BdCtrlFifoReset( Dev, BDCTRL_FIFORESET_LOW );

        Dev->ExpectedFifoSanity = (FIFO_SANITY_0 & Dev->AtodSizeMask);
}

SRLOCAL void ParXch_DramFifoRd( DEVINFO *Dev, unsigned char *Value ) {

        dbg4_printk( "ParXch_DramFifoRd\n" );


        // Read in a byte from the DRAM FIFO.

        Dev->ParPortDsRd( Dev, Value );

        *Value &= Dev->AtodSizeMask; // 0xFF for 8CH, 0x0F for 4CH/1CH


        // Save FIFO history for diagnostic purposes.

        Dev->FifoHistory[Dev->FHindex] = *Value;
        Dev->FHindex += 1;
        Dev->FHindex &= FH_MASK;

}
SRLOCAL void ParXch_DramFifoWr( DEVINFO *Dev, unsigned char Value ) {

        dbg4_printk( "ParXch_DramFifoWr\n" );


        // Write a single nibble (packed in a byte) to the DRAM FIFO.
        //
        // This function only works if the BdCtrlRun bit is off.  If
        // BdCtrlRun is on, the DRAM is connected to the A/D converters,
        // and this call will be ignored.
        //

        Dev->ParPortDsWr( Dev, Value );


}

SRLOCAL int ParXch_DramFifoReady( DEVINFO *Dev ) {

        unsigned char Status, temp;

        dbg4_printk("ParXch_DramFifoReady\n");


        // Test the DRAM FIFO ready/empty flag.
        //
        // The DRAM FIFO ready/empty flag is mapped by the XCR5128 to the PS0
        // signal on the PC parallel port status register.
        //

        ParXch_ParPortStatRd( Dev, &Status );


        if ( (Status & PS0) ) // then empty = 1, data is not ready

                return( 0 );


        else {  // empty = 0, data is ready ...

                // ... but we have to wait for empty flag skew ...

                ParXch_ParPortStatRd( Dev, &temp );
                ParXch_ParPortStatRd( Dev, &temp );
                ParXch_ParPortStatRd( Dev, &temp );
                ParXch_ParPortStatRd( Dev, &temp );

                return( 1 );

                }

}
SRLOCAL int ParXch_DramFifoFull( DEVINFO *Dev ) {

        unsigned char Status;

        dbg4_printk("ParXch_DramFifoFull\n");




        // Test for Dram FIFO overflow.
        //
        // Once overflow has occurred, this bit will stay set until
        // the DramFifoReset function is called by ParXch_Init
        // from the user level ParXchOpen or ParXchFullOpen functions.
        //
        // This code is conditionally compiled in depending on the
        // PARXCH_TEST_OVERFLOW define.


#if PARXCH_TEST_OVERFLOW == 1

        ParXch_ParPortStatRd( Dev, &Status );

        if ( (Status & PS1) )
                return( 1 ); // overflow
        else
                return( 0 ); // ok, no overflow



#elif PARXCH_TEST_OVERFLOW == 0

        // Don't bother testing the overflow signal ... fake success

        return( 0 );


#else
#pragma message( "PARXCH_TEST_OVERFLOW must be defined as 1 or 0" )
#endif


}





/* PARXCH DRAM FIFO READ AND GET DATA FUNCTIONS:

Higher level functions for getting back complete samples from the
DRAM FIFO.

*/

#define NEG24BIT                0x00800000L
#define SIGNEXTEND              0xFF000000L

SRLOCAL int ParXch_DramFifoReadDemux( DEVINFO *Dev,
                                      long *AnalogValues,
                                      long *DigitalValues,
                                      int *Mark,
                                      long *GpsCounter
                                    ) {

        int i, byte, bit;
        unsigned char MuxedBits;
        unsigned char chan0, chan1, chan2, chan3, chan4, chan5, chan6, chan7;



        // Read and verify FIFO sanity bit which toggles between 0 and 1.

        ParXch_DramFifoRd( Dev, &MuxedBits );

        if ((MuxedBits & Dev->AtodFifoSanityMask) != Dev->ExpectedFifoSanity) {
                // Roll index back to failed value instead of next value
                Dev->FHindex -= 1;
                Dev->FHindex &= FH_MASK;
                OsWriteFifoHistory( Dev, MuxedBits, Dev->ExpectedFifoSanity );
                return( -1 );
                }

        Dev->ExpectedFifoSanity ^= Dev->AtodFifoSanityMask;


        // Check for GPS mark bit

        if ((MuxedBits & Dev->AtodFifoMarkMask) == FIFO_MARK_YES)
                *Mark = 1;
        else
                *Mark = 0;



        // Depending on firmware, the A/D may have digital and GPS info too

        if (Dev->AtodCounterChannels > 0) {

                // Read GPS Counter bits, low byte first, and save

                *GpsCounter = 0L;

                for ( i = 0 ; i < 2 ; i++ ) {
                        ParXch_DramFifoRd( Dev, &MuxedBits );
                        *( ((char *)(GpsCounter)) + i ) = MuxedBits;
                        }

                } // end gps counter channels


        if (Dev->AtodDigitalChannels > 0) {

                // Read digital I/O bits and save

                ParXch_DramFifoRd( Dev, &MuxedBits );
                DigitalValues[0] = MuxedBits;

                } // end digital channels > 0





        // Demux bits coming back from the PARXCH FIFO:
        //
        // 3*8 = 24 # bit words are demuxed.  The results are saved in #
        // long ints, one for each channel.  Dimension AnalogValues[] as:
        //
        // long AnalogValues[#]; where # is 4 or 8
        //


        byte = 3;

        *( (char *)(&AnalogValues[0]) + byte ) = 0;
        *( (char *)(&AnalogValues[1]) + byte ) = 0;
        *( (char *)(&AnalogValues[2]) + byte ) = 0;
        *( (char *)(&AnalogValues[3]) + byte ) = 0;
        if (Dev->AtodAnalogChannels == 8) {
                *( (char *)(&AnalogValues[4]) + byte ) = 0;
                *( (char *)(&AnalogValues[5]) + byte ) = 0;
                *( (char *)(&AnalogValues[6]) + byte ) = 0;
                *( (char *)(&AnalogValues[7]) + byte ) = 0;
                }


        for ( byte = 2 ; byte >= 0 ; byte-- ) {

                for ( bit = 0 ; bit < 8 ; bit++ ) {

                        ParXch_DramFifoRd( Dev, &MuxedBits );

#if defined( USE_C_CODE_DEMUX )
                        
                        chan0 <<= 1;
                        chan0 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan1 <<= 1;
                        chan1 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan2 <<= 1;
                        chan2 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan3 <<= 1;
                        chan3 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan4 <<= 1;
                        chan4 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan5 <<= 1;
                        chan5 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan6 <<= 1;
                        chan6 |= (MuxedBits & 0x01);

                        MuxedBits >>= 1;
                        chan7 <<= 1;
                        chan7 |= (MuxedBits & 0x01);


#elif defined( SROS_WINXP ) || defined( SROS_WIN2K ) || defined( SROS_WIN9X ) || defined( SROS_MSDOS )

                        _asm {
                                mov al,BYTE PTR MuxedBits

                                rcr al,1
                                rcl chan0,1

                                rcr al,1
                                rcl chan1,1

                                rcr al,1
                                rcl chan2,1

                                rcr al,1
                                rcl chan3,1

                                rcr al,1
                                rcl chan4,1

                                rcr al,1
                                rcl chan5,1

                                rcr al,1
                                rcl chan6,1

                                rcr al,1
                                rcl chan7,1

                                }

#elif defined( SROS_LINUX )

                        asm (
                               "movb %8,%%al;        \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%0;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%1;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%2;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%3;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%4;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%5;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%6;          \
                                                     \
                                rcrb $1,%%al;        \
                                rclb $1,%7"          \
                                                     \
                               : "=m" (chan0),       \
                                 "=m" (chan1),       \
                                 "=m" (chan2),       \
                                 "=m" (chan3),       \
                                 "=m" (chan4),       \
                                 "=m" (chan5),       \
                                 "=m" (chan6),       \
                                 "=m" (chan7)        \
                               :  "m" (MuxedBits)    \
                               : "%ax"

                                );
#endif // SROS_xxxxx
                        }

                *( (char *)(&AnalogValues[0]) + byte ) = chan0;
                *( (char *)(&AnalogValues[1]) + byte ) = chan1;
                *( (char *)(&AnalogValues[2]) + byte ) = chan2;
                *( (char *)(&AnalogValues[3]) + byte ) = chan3;

                if (Dev->AtodAnalogChannels == 8) {
                        *( (char *)(&AnalogValues[4]) + byte ) = chan4;
                        *( (char *)(&AnalogValues[5]) + byte ) = chan5;
                        *( (char *)(&AnalogValues[6]) + byte ) = chan6;
                        *( (char *)(&AnalogValues[7]) + byte ) = chan7;
                        }

                }


        // If sign extend data format, extend 24 bit values to 32 bit.

        if (Dev->DataFormat == PARXCH_DF_SIGNED) {

                for ( i = 0 ; i < Dev->AtodAnalogChannels ; i++ ) {

                        if ( AnalogValues[i] & NEG24BIT )
                                AnalogValues[i] |= SIGNEXTEND;

                        }
                }


        // Return the number of longs just read.

        return( Dev->AtodAnalogChannels );
}







/* PARXCH FRONT PANEL USER DIGITAL IO:

These functions program the PARXCH digital IO.  ParXchOpen() must be
called before these to initialize the PC parallel port.

*/

SRLOCAL void ParXch_UserIoRd( DEVINFO *Dev, int *Value ) {

        dbg4_printk( "ParXch_UserIoRd\n" );

        // This assumes the PARXCH is in a BdCtrl mode > HALT_INIT ...

        Dev->ParPortAsRd( Dev, (unsigned char *)Value);

        *Value &= 0xF;

}
SRLOCAL void ParXch_UserIoWr( DEVINFO *Dev, int Value ) {

        dbg4_printk( "ParXch_UserIoWr\n" );


        // This assumes the PARXCH is in a BdCtrl mode > HALT_INIT ...

        Value &= 0xF;

        Dev->ParPortDsWr( Dev, (unsigned char)Value);

}






/* PARXCH INITIALIZATION:

This function must be called to initialize the PARXCH board before
starting execution.  It takes care of all initialization.  No other
functions need to be called to initialize the PARXCH.

During initialization the ADS1210 control regsters are programmed
with values setting all board parameters like sampling rate
programmable gain etc.  All ADS1210s are programmed in parallel with
the same values.

See parxch.h for defined constants to use with the parameters passed
to ParXchOpen.  To convert floating point sps (samples per second)
sampling rates to Turbo and Decimation values, use the helper
function ParXchSpsGainToTde.

NOTE:

This function has no return value since it must always be
STATUS_SUCCESS so that the Parm data will get passed.  The true error
information is stored in Parm->Error which IS returned to the user's
calling program from ParXch_DrvIoctl, the only place from which
ParXch_Init is called.

*/

#define DRV_ERROR_RETURN( Dev, err ) {                        \
                                Parm->Error = err;            \
                                ParXch_EcpPortRestore( Dev ); \
                                return;                       \
                                }

SRLOCAL void ParXch_Init( DEVINFO *Dev, int *IOBuffer ) {

        INITXCHPARMS *Parm;
        unsigned char wrdata, rddata, muxbits, expected;
        int           i, fail;
        long          CrValue, ReadbackCrValues[PARXCH_MAX_ANALOG_CHANNELS];
        unsigned char Wbyte[4], Rbyte[4];
        KIRQL         OldIrqlOrFlags;

        dbg2_printk("In ParXch_Init\n");

        Parm = (PINITXCHPARMS)IOBuffer;


        // Clear FIFO history, used for diagnostic purposes.

        Dev->FHindex = 0;
        for ( i = 0 ; i < FH_MASK+1 ; i++ )
                Dev->FifoHistory[i] = 0x00;


        // Set A/D specific values
        Dev->AtodXchModel = Parm->XchModel;

        dbg1_printk( "AtodXchModel is %d\n", Dev->AtodXchModel );

        if (Dev->AtodXchModel == PARXCH_MODEL_PAR8CH) {
                Dev->AtodFifoBits = 8;
                Dev->AtodSizeMask = 0xFF;
                Dev->AtodAnalogChannels  = PAR8CH_ANALOG_CHANNELS;
                Dev->AtodDigitalChannels = PAR8CH_DIGITAL_CHANNELS;
                Dev->AtodGpsChannels     = PAR8CH_GPS_CHANNELS;
                Dev->AtodCounterChannels = PAR8CH_COUNTER_CHANNELS;
                }
        else if (Dev->AtodXchModel == PARXCH_MODEL_PAR4CH) {
                Dev->AtodFifoBits = 4;
                Dev->AtodSizeMask = 0x0F;
                Dev->AtodAnalogChannels  = PAR4CH_ANALOG_CHANNELS;
                Dev->AtodDigitalChannels = PAR4CH_DIGITAL_CHANNELS;
                Dev->AtodGpsChannels     = PAR4CH_GPS_CHANNELS;
                Dev->AtodCounterChannels = PAR4CH_COUNTER_CHANNELS;
                }
        else if (Dev->AtodXchModel == PARXCH_MODEL_PAR1CH) {
                Dev->AtodFifoBits = 1;
                Dev->AtodSizeMask = 0x0F; // 1ch board still has 4ch firmware
                Dev->AtodAnalogChannels  = PAR1CH_ANALOG_CHANNELS;
                Dev->AtodDigitalChannels = PAR1CH_DIGITAL_CHANNELS;
                Dev->AtodGpsChannels     = PAR1CH_GPS_CHANNELS;
                Dev->AtodCounterChannels = PAR1CH_COUNTER_CHANNELS;
                }
        else {
                DRV_ERROR_RETURN( Dev, PARXCH_ERROR_ATOD_UNKNOWN );
                }

        Dev->AtodFifoSanityMask = (FIFO_SANITY_MASK & Dev->AtodSizeMask);
        Dev->AtodFifoMarkMask   = (FIFO_MARK_MASK   & Dev->AtodSizeMask);

        Dev->FHsanityInterval = 25;
        if (Dev->AtodDigitalChannels > 0)
                Dev->FHsanityInterval += 1; // one extra digital byte
        if (Dev->AtodCounterChannels > 0)
                Dev->FHsanityInterval += 2; // two extra counter bytes

        dbg3_printk( "Dev->FHsanityInterval    %d\n", Dev->FHsanityInterval );
        dbg3_printk( "Dev->AtodFifoBits        %d\n", Dev->AtodFifoBits );
        dbg3_printk( "Dev->AtodAnalogChannels  %d\n", Dev->AtodAnalogChannels );
        dbg3_printk( "Dev->AtodDigitalChannels %d\n", Dev->AtodDigitalChannels );
        dbg3_printk( "Dev->AtodGpsChannels     %d\n", Dev->AtodGpsChannels );
        dbg3_printk( "Dev->AtodCounterChannels %d\n", Dev->AtodCounterChannels );
        dbg3_printk( "Dev->AtodSizeMask        0x%X\n", Dev->AtodSizeMask );
        dbg3_printk( "Dev->AtodFifoSanityMask  0x%X\n", Dev->AtodFifoSanityMask );
        dbg3_printk( "Dev->AtodFifoMarkMask    0x%X\n", Dev->AtodFifoMarkMask );


        // Initialize the parallel port mode and drivers.

        Dev->ParPortMode = Parm->ParPortMode;
        ParXch_ParPortInit( Dev );
        ParXch_EcpPortInit( Dev );


        // Initialize various parameters.

        Dev->DataFormat   = Parm->Df;
        Dev->ExtraDecNavg = Parm->ExtraDecimation;

        // Dev->ParPortIrq   = Parm->ParPortIrq; is set elsewhere
        // For example, from the registry in Wdm_DrvAddDevice
        // or from Msdos_NameToIrq in Msdos_DrvOpen
        // or from the insmod command line in the Linux module_param


        // Compute number of PAR8CH OBC counts per sample for later use
        // in the DrvRead function.  The constant 64 in the formula
        // comes from OBC_FREQ * PARXCH_FCONST / PARXCH_XTAL where
        // OBC_FREQ is 1250000 (ie 1/800ns), PARXCH_FCONST is 512, and
        // PARXCH_XTAL is 10.0E+6 (ie 10MHz).  See the library function
        // ParXchSpsGainToTde

        Dev->ExtraDecObcPerSample = (64L * (Parm->Decimation+1)) /
                                                   (1<<Parm->TurboLog);


        // Power down and back up for a board level reset.

        if ( !ParXch_PowerDown( Dev ) )

                DRV_ERROR_RETURN( Dev, PARXCH_ERROR_VOLTAGE_BAD_ON );

        if ( !ParXch_PowerUp( Dev ) )

                DRV_ERROR_RETURN( Dev, PARXCH_ERROR_VOLTAGE_BAD_OFF );





        // Initialize the PARXCH BdCtrl register.

        ParXch_BdCtrlUnlock( Dev );
        ParXch_BdCtrlInit( Dev );




        // Verify that BdCtrlUnlock and BdCtrlInit succeeded by pumping
        // and then reading back the DRAM FIFO.  This is ok to do here
        // since BdCtrlInit leaves the board in the HALT_INIT mode.

        ParXch_DramFifoReset( Dev );

        ParXch_DramFifoWr( Dev, 0x00 );
        wrdata = 0x01;
        for ( i = 0 ; i < Dev->AtodFifoBits ; i++ ) {
                ParXch_DramFifoWr( Dev, wrdata );
                wrdata = (wrdata << 1);
                }

        fail     = 0;
        muxbits  = 0x00;

        ParXch_DramFifoRd( Dev, &rddata );
        if (rddata != 0x00) {
                dbg2_printk("Failed unlock test with 0x%02X, expected 0x00\n",rddata);
                fail     = 1;
                muxbits  = rddata;
                expected = 0x00;
                }

        wrdata   = 0x01;
        expected = wrdata;
        for ( i = 0 ; i < Dev->AtodFifoBits ; i++ ) {
                ParXch_DramFifoRd( Dev, &rddata );
                if (rddata != wrdata) {
                        dbg2_printk("Failed unlock test with 0x%02X, expected 0x%02X\n",rddata,wrdata);
                        fail     = 1;
                        muxbits  = rddata;
                        expected = wrdata;
                        }
                wrdata = (wrdata << 1);
                }
        if (fail) {
                OsWriteFifoHistory( Dev, muxbits, expected );
                DRV_ERROR_RETURN( Dev, PARXCH_ERROR_NOT_UNLOCKED );
                }




        // Program the ADS1210 command register and do a presence test
        // by reading back and checking the ADS1210 command register
        // values one byte at a time.
        //
        // The DRDY/DSYNC bit on the command register is a no care
        // on return for this check.

        CrValue =

                  ADS1210_BIAS_GEN_ON
                | ADS1210_INT_REF_OFF
                | ADS1210_DATA_FORMAT( Parm->Df )
                | ADS1210_BIPOLAR
                | ADS1210_BYTE_DESCEND
                | ADS1210_MSB_FIRST
                | ADS1210_SDO_ENABLED
                | ADS1210_DSYNC_NOCHANGE
                | ADS1210_CAL_MODE_SELF
                | ADS1210_GAIN( Parm->GainLog )
                | ADS1210_CHAN_SELECT_0
                | ADS1210_TURBO( Parm->TurboLog )
                | ADS1210_DECIMATION( Parm->Decimation )

                ;

        Dev->ADS1210SaveCrValue = CrValue;
        ReadbackCrValues[0]     = 0xFFFFFFFF;



        // Prepare individual write bytes from combined CrValue

        Wbyte[0] = *( (unsigned char *)(&CrValue) + 0 );
        Wbyte[1] = *( (unsigned char *)(&CrValue) + 1 );
        Wbyte[2] = *( (unsigned char *)(&CrValue) + 2 );
        Wbyte[3] = *( (unsigned char *)(&CrValue) + 3 );


        // It is important that these bytes get transferred quickly
        // without long gaps so we surround this section with spin
        // locks to make sure we won't get swapped out in the middle.
        // This section typically takes about 1.3 ms to complete.

        OsSpinLockAcquire( &Dev->ADS1210InitLock, OldIrqlOrFlags );


        // Process bias+signed+order byte

        ParXch_ADS1210_WriteReg8( Dev, ADS1210_ADDR_CONTROL3,  Wbyte[3] );
        ParXch_ADS1210_ReadReg8( Dev,  ADS1210_ADDR_CONTROL3, &Rbyte[3] );


        // Process low part of sample rate byte

        ParXch_ADS1210_WriteReg8( Dev, ADS1210_ADDR_CONTROL0,  Wbyte[0] );
        ParXch_ADS1210_ReadReg8( Dev,  ADS1210_ADDR_CONTROL0, &Rbyte[0] );


        // Process high part of sample rate byte

        ParXch_ADS1210_WriteReg8( Dev, ADS1210_ADDR_CONTROL1,  Wbyte[1] );
        ParXch_ADS1210_ReadReg8( Dev,  ADS1210_ADDR_CONTROL1, &Rbyte[1] );


        // Process self calibration request byte

        ParXch_ADS1210_WriteReg8( Dev, ADS1210_ADDR_CONTROL2,  Wbyte[2] );
        ParXch_ADS1210_ReadReg8( Dev,  ADS1210_ADDR_CONTROL2, &Rbyte[2] );

        OsSpinLockRelease( &Dev->ADS1210InitLock, OldIrqlOrFlags );




        // Prepare combined readback value from individually read bytes

        *( (unsigned char *)(&ReadbackCrValues[0]) + 0 ) = Rbyte[0];
        *( (unsigned char *)(&ReadbackCrValues[0]) + 1 ) = Rbyte[1];
        *( (unsigned char *)(&ReadbackCrValues[0]) + 2 ) = Rbyte[2];
        *( (unsigned char *)(&ReadbackCrValues[0]) + 3 ) = Rbyte[3];





        // While the A/Ds finish completing self calibration, reset the
        // DRAM FIFO, and enable the XCR5128 for run mode.

        ParXch_DramFifoReset( Dev );



        // Check the ADS1210 command register readback value.

        Parm->ADS1210_CrValue = ReadbackCrValues[0];

        if ( (ReadbackCrValues[0] & ~ADS1210_DRDY) != (CrValue  & ~ADS1210_DRDY) ) {
                dbg2_printk("Failed readback test got 0x%08lX expected 0x%08lX\n",(ReadbackCrValues[0] & ~ADS1210_DRDY), (CrValue & ~ADS1210_DRDY) );
                DRV_ERROR_RETURN( Dev, PARXCH_ERROR_ADS1210_CR_READBACK );
                }
        dbg2_printk("Passed readback test got 0x%08lX expected 0x%08lX\n",(ReadbackCrValues[0] & ~ADS1210_DRDY), (CrValue & ~ADS1210_DRDY) );



        // All is ok ... !
        //
        // Put the BdCtrl register in HALT_UDIO mode so the user can
        // use the digital io before starting A/D execution.
        //
        //

        ParXch_BdCtrlMode( Dev, BDCTRL_MODE_HALT_UDIO );


        Parm->Error = PARXCH_ERROR_NONE;
        return;

}




/* PARXCH EXECUTION CONTROL FUNCTIONS:
 *
 *
 *
 *
 */

SRLOCAL void ParXch_Start( DEVINFO *Dev ) {

        dbg2_printk("In ParXch_Start\n");


        // Reset the hardware FIFO.

        ParXch_DramFifoReset( Dev );


        // Reset count of acquired data.

        Dev->AcqDataCount = 0;


        // Put the Xilinx CoolRunner controller in run mode.

        ParXch_BdCtrlMode( Dev, BDCTRL_MODE_RUN_EF );


        // Enable Interrupts when driver acquired an interrupt and is using GPS

        if ( (Dev->ParPortIrq != 0) && (Dev->SrCommonArea != NULL) )
                ParXch_InterruptEnable( Dev );

}
SRLOCAL void ParXch_Stop( DEVINFO *Dev ) {

        dbg2_printk("In ParXch_Stop\n");


        // Disable Interrupts
        // To be safe, we always disable them, even though they are only
        // enabled in the start function if GPS is being used.

        ParXch_InterruptDisable( Dev );


        // Reset the ADS1210, stopping execution.

        ParXch_BdCtrlInit( Dev );

}





/* PARXCH PARALLEL PORT PC INTERRUPT ENABLE/DISABLE:

These functions enable or disable the parallel port PC interrupt.
Note that ParXchOpen() must be called before these to initialize the
PC parallel port.

*/

SRLOCAL void ParXch_InterruptEnable( DEVINFO *Dev ) {

        dbg5_printk( "ParXch_InterruptEnable\n" );


        // Enable the parallel port PC interrupt.

        ParXch_ParPortCtrlIrqen( Dev, PIRQ_ENABLE );

}
SRLOCAL void ParXch_InterruptDisable( DEVINFO *Dev ){

        dbg5_printk( "ParXch_InterruptDisable\n" );


        // Disable the parallel port PC interrupt.

        ParXch_ParPortCtrlIrqen( Dev, PIRQ_DISABLE );

}










/* PARXCH HELPER FUNCTION:

*/


SRLOCAL void ParXch_FillDevInfo( DEVINFO *Dev, char *DriverName,
                                 int ParPortBase, int ParPortEcpOffset,
                                 int ParPortIrq ) {

        int i;

        dbg0_printk("In FillDevInfo\n");


        // Initialize Device Information structure

        strncpy( Dev->DriverNameBuffer, DriverName, MAX_STR_LENGTH-1 );
        Dev->DriverNameBuffer[MAX_STR_LENGTH-1] = '\0';


        Dev->AtodXchModel        = 0;     // These set correctly in ParXch_Init
        Dev->AtodFifoBits        = 0;
        Dev->AtodAnalogChannels  = 0;
        Dev->AtodDigitalChannels = 0;
        Dev->AtodGpsChannels     = 0;
        Dev->AtodCounterChannels = 0;
        Dev->AtodSizeMask        = 0xFF;
        Dev->AtodFifoSanityMask  = 0x00;
        Dev->AtodFifoMarkMask    = 0x00;

        Dev->ParPortBase         = ParPortBase;
        Dev->ParPortMode         = PARXCH_PORT_MODE_BPP;
        Dev->ParPortIrq          = ParPortIrq;
        Dev->ParPortFound        = 0;

        // Set ECP offset from standard port base
        // If user has specified, use that value
        // otherwise standard addresses use the standard 0x400
        // and non-standard ones use 0x100 since that seems
        // typical for PCI cards

        if (ParPortEcpOffset > 0)
                Dev->ParPortEcpOffset = ParPortEcpOffset;
        else if (ParPortBase == 0x378 || ParPortBase == 0x278)
                Dev->ParPortEcpOffset = 0x400;
        else
                Dev->ParPortEcpOffset = 0x100;

        ParXch_ParPortInit( Dev );

        Dev->ParPortCtrlValue    = 0x00;
        Dev->EcpPortSaveEcrValue = ECP_SVCINTR_DISABLE
                                 | ECP_DMA_DISABLE
                                 | ECP_ERRINTR_DISABLE
                                 | ECP_SUBMODE_PS2;

        Dev->VoltageIsGood       = (PARXCH_VOLTAGE_IS_GOOD *
                                    PARXCH_VOLTAGE_GOOD_MASK);

        Dev->BdCtrlValue         = 0x00;
        Dev->ADS1210SaveCrValue  = 0x00000000L;

        Dev->DataFormat          = PARXCH_DF_SIGNED;
        Dev->ExpectedFifoSanity  = (FIFO_SANITY_0 & Dev->AtodSizeMask);

        Dev->ExtraDecNavg        = PARXCH_EXTRADECIMATION_MIN;
        Dev->ExtraDecNcount      = 0;

        for ( i = 0 ; i < PARXCH_MAX_ANALOG_CHANNELS ; i++ )
                Dev->ExtraDecSum[i] = 0L;

        Dev->ExtraDecSumGpsCount  = 0L;
        Dev->ExtraDecObcPerSample = 0L;

        Dev->SrCommonArea        = NULL;
        Dev->AcqDataCount        = 0;
        Dev->IncludeMark         = 0;
        Dev->IncludeDigital      = 0;
        Dev->IncludeCounter      = 0;

        Dev->StartDone           = 0;
        Dev->ParPortAcquired     = 0;
        Dev->ParPortDevObj       = NULL;
        Dev->ParPortFileObj      = NULL;
        Dev->IrqObj              = NULL;

        Dev->DriverObj           = NULL;
        Dev->ThisFdoDevObj       = NULL;
        Dev->LowerDevObj         = NULL;
        Dev->PdoDevObj           = NULL;
        Dev->NotificationHandle  = NULL;

        Dev->ParPortDriver       = NULL;
        Dev->ParPortClass        = NULL;
        Dev->ParPortDevice       = NULL;
        Dev->ParPortPort         = NULL;

        Dev->PrevDevState        = RUNSTATE_STOPPED;
        Dev->CurrDevState        = RUNSTATE_STOPPED;
        Dev->CurrDevPower        = PowerDeviceD0;
        Dev->CurrSysPower        = PowerSystemWorking;

        Dev->WorkItemCount       = 0;

//        Dev->DevCaps;
//        Dev->RemoveLock;

        Dev->PowerDownOnExitFlag = 1;
        Dev->NumberOpensLock     = 0;
        Dev->MsgBuffer[0]        = '\0';


        Dev->FHsanityInterval = 25;

        Dev->FHindex = 0;
        for ( i = 0 ; i < FH_MASK+1 ; i++ )
                Dev->FifoHistory[i] = 0x00;

        ParXch_ResetMarkCount( Dev );
}

SRLOCAL int ParXch_CheckHealth( DEVINFO *Dev ) {

        int Health;

        dbg4_printk("In ParXch_CheckHealth\n" );


        Health = 0x00;

#if PARXCH_TEST_VOLTAGE_GOOD == 1
        if ( !ParXch_VoltageGood( Dev ) )
                Health |= HEALTH_VOLTAGE_BAD;
#endif

        if ( ParXch_DramFifoFull( Dev ) )
                Health |= HEALTH_OVERFLOW;

        return( Health );
}


SRLOCAL int ParXch_AttachGps( DEVINFO *Dev, void *CommonArea ) {

        dbg0_printk("In ParXch_AttachGps, incoming CommonArea is 0x%p\n",CommonArea);

        Dev->SrCommonArea = (GPSCKA*)CommonArea;

#ifdef GPS_AVAILABLE

        dbg0_printk("In ParXch_AttachGps, GPS_AVAILABLE section\n");

        // Do some error checks

        if (Dev->SrCommonArea == NULL) {
                dbg_printk("WARNING: Common area not set !!!\n");
                return( PARXCH_ERROR_GPS_NOT_ATTACHED );
                }

        if (Dev->SrCommonArea->CkaRev != CKA_REV) {
                dbg_printk("WARNING: Common area mismatch !!!  Recompile drivers\n");
                dbg_printk("Gps thinks %d, ParXch thinks %d\n",Dev->SrCommonArea->CkaRev,CKA_REV);
                return( PARXCH_ERROR_GPS_VERSION_MISMATCH );
                }



        // Store ParXch info in the common area

        Dev->SrCommonArea->AtodDev          = (void*)Dev;
        Dev->SrCommonArea->AtodFuncControl  = (PKFN_CONTROL)ParXch_BdCtrlDrdyIrqReq;
        Dev->SrCommonArea->AtodFuncRelease  = (PKFN_RELEASE)ParXch_ReleaseGps;
        Dev->SrCommonArea->AtodFuncResetGps = (PKFN_RESETGPS)ParXch_ResetMarkCount;
        Dev->SrCommonArea->InterruptObject  = (void*)Dev->IrqObj;


        ParXch_ResetMarkCount( Dev );


        return( PARXCH_ERROR_NONE );
#else
        return( PARXCH_ERROR_GPS_NOT_AVAILABLE );
#endif
}

SRLOCAL void ParXch_ReleaseGps( DEVINFO *Dev ) {

        dbg0_printk("In ParXch_ReleaseGps\n" );

        if (Dev->SrCommonArea == NULL) {
                dbg_printk("WARNING: Common area not set !!!\n");
                return;
                }

#ifdef GPS_AVAILABLE
        Dev->SrCommonArea->AtodDev          = NULL;
        Dev->SrCommonArea->AtodFuncControl  = NULL;
        Dev->SrCommonArea->AtodFuncRelease  = NULL;
        Dev->SrCommonArea->AtodFuncResetGps = NULL;
#endif
        Dev->SrCommonArea                   = NULL;
}

SRLOCAL void ParXch_ResetMarkCount( DEVINFO *Dev ) {

        dbg0_printk("In ParXch_ResetMarkCount\n" );

        // Reset GPS mark count

        Dev->MarkCount   = 1L;
        Dev->MarkCurrent = 0L;

}

int ParXch_DrvInterrupt( DEVINFO *Dev ) {

#ifdef GPS_AVAILABLE
        PKFN_INTERRUPT InterruptFn;

//        dbg5_printk( "ParXch_DrvInterrupt\n" );


        // No pass-through interrupt installed

        if ( (Dev == NULL) || (Dev->SrCommonArea == NULL) )
                return( 1 );

        // Process pass-through GPS interrupt

        InterruptFn = (PKFN_INTERRUPT)Dev->SrCommonArea->GpsFuncInterrupt;

        if ((void*)InterruptFn == NULL)
                return( 0 );

        InterruptFn( (void*)Dev->SrCommonArea->GpsDev );
#endif

        return( 1 );
}



/* DRIVER DISPATCH FUNCTIONS:
 *
 *
 *
 *
 *
 *
 *
 */

SRLOCAL long ParXch_DrvOpen( DEVINFO *Dev, IRP *Irp ) {

        dbg1_printk("In ParXch_DrvOpen\n");

        /* Allow only one device open at a time */

        if (Dev->NumberOpensLock) {
                dbg_printk( "Multiple Opens! 0x%lX\n", STATUS_TOKEN_ALREADY_IN_USE);
                OsIrpFinish( Irp, STATUS_TOKEN_ALREADY_IN_USE, 0 );
                return( STATUS_TOKEN_ALREADY_IN_USE );
                }
        else
                Dev->NumberOpensLock++;


        OsModuleUseCountInc;


        // Set up the default parallel port mode specific functions.

        Dev->ParPortMode = PARXCH_PORT_MODE_BPP;
        ParXch_ParPortInit( Dev );
        ParXch_EcpPortInit( Dev );


        OsIrpFinish( Irp, STATUS_SUCCESS, 0 );

        return( STATUS_SUCCESS );
}

SRLOCAL long ParXch_DrvClose( DEVINFO *Dev, IRP *Irp ) {

        dbg1_printk("In ParXch_DrvClose\n");


        // Restore the saved ECP parallel port ECR value.

        ParXch_EcpPortRestore( Dev );


        // Ensure that execution and interrupts are really stopped.

        ParXch_Stop( Dev );


        // Release common kernel area.

        if (Dev->SrCommonArea != NULL)
                ParXch_ReleaseGps( Dev );



        // Turn off power.

        if ( Dev->PowerDownOnExitFlag )
                ParXch_PowerDown( Dev );


        // Release device for others to open.

        Dev->NumberOpensLock--;

        OsModuleUseCountDec;


        OsIrpFinish( Irp, STATUS_SUCCESS, 0 );

        return( STATUS_SUCCESS );
}

SRLOCAL long ParXch_DrvRead( DEVINFO *Dev, IRP *Irp ) {

        long               Status;
        unsigned long      OutBufferSize;      // Max data caller can accept
        unsigned long      DataBufferSize;
        long              *IOBuffer;           // Pointer to transfer buffer
        unsigned long     *UserBuffer;
        int                AnalogChannels, DigitalChannels, GpsChannels;
        int                CounterChannels, TotalChannels;
        int                MarkChannelIndex, CounterChannelIndex;
        long               TmpVal[PARXCH_MAX_ANALOG_CHANNELS];
        long               TmpDigVal[PARXCH_MAX_DIGITAL_CHANNELS], GpsCounter;
        int                Num, Nval, ipt, ichan, Mark;

        dbg1_printk("In ParXch_DrvRead for Port 0x%X\n", Dev->ParPortData);


        // Set and verify buffer parameters.

        Status = OsSetReadParms( Irp, &OutBufferSize, &IOBuffer );

        if (Status != STATUS_SUCCESS) {
                OsIrpFinish( Irp, Status, OutBufferSize );
                return( Status );
                }


        // Determine how many channels we're working with

        AnalogChannels = Dev->AtodAnalogChannels;

        if (Dev->IncludeDigital == 1)
                DigitalChannels = Dev->AtodDigitalChannels;
        else
                DigitalChannels = 0;

        if (Dev->IncludeMark == 1)
                GpsChannels     = Dev->AtodGpsChannels;
        else
                GpsChannels     = 0;

        if (Dev->IncludeCounter == 1)
                CounterChannels = Dev->AtodCounterChannels;
        else
                CounterChannels = 0;


        TotalChannels       = AnalogChannels + DigitalChannels +
                              GpsChannels    + CounterChannels;
        MarkChannelIndex    = AnalogChannels + DigitalChannels;
        CounterChannelIndex = MarkChannelIndex + GpsChannels;

        // Fill user buffer with data from hardware FIFO

        Nval = (int)(OutBufferSize / sizeof(long));
        Num  = Nval / TotalChannels;      // Must read an even number of
        Nval = Num  * TotalChannels;      // points for each channel (+mark)
        Status = STATUS_SUCCESS;
        UserBuffer = (unsigned long *)IOBuffer;
        Num = ipt = 0;


        while ( ipt < Nval ) {

            // Check for no more raw points ready to read

            if ( !ParXch_DramFifoReady( Dev ) ) {
                    Num = ipt;
                    break;
                    }


            // Read next raw value from hardware FIFO,
            // checking for sanity bit error

            if (ParXch_DramFifoReadDemux( Dev, TmpVal, TmpDigVal,
                                          &Mark, &GpsCounter ) == -1) {
                    Status = STATUS_DEVICE_DATA_ERROR;
                    OsIrpFinish( Irp, Status, (Num*sizeof(long)) );
                    return( Status );
                    }



            // If ExtraDecimation is in effect (eg ExtraDecNavg > 0),
            // several A/D points are acquired and processed into a
            // single user point so the analog data values are added
            // into an accumulating sum.  This is divided by the sum
            // count to get an average user point value once enough raw
            // A/D points have been acquired.

            for ( ichan = 0 ; ichan < AnalogChannels ; ichan++ )
                    Dev->ExtraDecSum[ichan] += TmpVal[ichan];

            Dev->ExtraDecNcount++;

            // The PAR8CH GPS on-board counter (OBC) value measures the
            // number of 800ns counts between the arrival of a PPS
            // signal and the next A/D point.  If ExtraDecimation is in
            // effect, once an OBC value is detected, it must be incremented
            // for each subsequent A/D point that is acquired until the
            // user point is ready.

            if ( GpsCounter != 0 )                   // OBC value found + saved
                    Dev->ExtraDecSumGpsCount = GpsCounter;

            else if ( Dev->ExtraDecSumGpsCount != 0) // incremented for A/D pt
                    Dev->ExtraDecSumGpsCount += Dev->ExtraDecObcPerSample;


            // Keep track of the current Gps mark count (ie PPS event num)

            if (Mark)
                    Dev->MarkCurrent = Dev->MarkCount++;




            // If Ncount reaches Nvag, we are ready to generate a
            // user point from the raw hardware points we have read
            // and summed.  Otherwise, just return and wait to be
            // called for another raw point.

            if ( Dev->ExtraDecNcount >= Dev->ExtraDecNavg ) {

                // Divide by Navg, save to DataBuffer and
                // reset sum and count

                for ( ichan = 0 ; ichan < AnalogChannels ; ichan++ ) {
                    Dev->ExtraDecSum[ichan] /= Dev->ExtraDecNavg;

                    OsCopyToUser( &UserBuffer[ichan],
                                  &Dev->ExtraDecSum[ichan],
                                  sizeof(long) );

                    Dev->ExtraDecSum[ichan] = 0L;
                    }



                // If requested, save digital channels

                // Note: When decimating, we are just saving the
                // last hardware point acquired and are throwing
                // the others away to produce one decimated
                // output point.

                if (Dev->IncludeDigital == 1) {
                    for ( ichan = 0; ichan < DigitalChannels ; ichan++ ) {
                        OsCopyToUser( &UserBuffer[AnalogChannels+ichan],
                                      &TmpDigVal[ichan],
                                      sizeof( long ) );
                            } // end for
                        } // end if include digital


                // If requested, save mark info

                // Note: When decimating, some uncertainty in
                // the timing of the mark is now introduced since
                // we aren't recording which of the hardware points
                // that are being averaged to produce one decimated
                // output point is the one being marked.

                if (Dev->IncludeMark == 1) {
                    ichan = MarkChannelIndex;
                    OsCopyToUser( &UserBuffer[ichan],
                                  &Dev->MarkCurrent,
                                  sizeof( long ) );
                    }

                if (Dev->IncludeCounter == 1) {
                    ichan = CounterChannelIndex;
                    OsCopyToUser( &UserBuffer[ichan],
                                  &Dev->ExtraDecSumGpsCount,
                                  sizeof(long) );
                    }


                Dev->ExtraDecNcount      = 0;
                Dev->ExtraDecSumGpsCount = 0L;
                Dev->MarkCurrent         = 0L;
                Dev->AcqDataCount++;


                // Move to next point

                UserBuffer += TotalChannels;
                ipt        += TotalChannels;

                }  // End of if Ncount == Navg

            Num = ipt;

            } // End of for ipt


        DataBufferSize = Num * sizeof(long);

        OsIrpFinish( Irp, Status, DataBufferSize );

        return( Status );
}

SRLOCAL long ParXch_DrvIoctl( DEVINFO *Dev, IRP *Irp ) {

        long           Status;
        unsigned long  ReturnedBytes, Command;
        int           *InBuffer, *OutBuffer, Nin, Nout, Tmp;
        unsigned char  ucharvalue;
        PORTVALUEXCH   *Pv;

        dbg4_printk("In ParXch_DrvIoctl\n");


        // Set and check command

        OsSetIoctlCommand( Irp, &Command );

        if ( ( IOCTL_TYPE(Command) != PARXCH_DEVICE_TYPE ) ||
             ( IOCTL_FUNC(Command) >= IOCTL_MAXFUNC ) ) {

                dbg4_printk("In ParXch_DrvIoctl fail device type and maxfunc check\n");
                OsIrpFinish( Irp, STATUS_INVALID_DEVICE_REQUEST, 0 );
                return( STATUS_INVALID_DEVICE_REQUEST );
                }



        // Set, check, and fill the data buffers

        Status = OsSetIoctlBuffers( Irp, Command, &InBuffer, &OutBuffer,
                                    &Nin, &Nout );

        if (Status != STATUS_SUCCESS) {
                dbg4_printk("In ParXch_DrvIoctl fail set buffers with status 0x%X\n", Status);
                dbg4_printk("Command 0x%X, Nin %d, Nout %d\n", Command, Nin, Nout);
                OsIrpFinish( Irp, Status, 0 );
                return( Status );
                }




        // Call specified IOCTL request function

        Status        = STATUS_SUCCESS;
        ReturnedBytes = 0;


        switch (Command) {

                case IOCTL_PARXCH_INIT:
                        ParXch_Init( Dev, InBuffer );
                        ReturnedBytes = sizeof(INITXCHPARMS);
                        break;

                case IOCTL_PARXCH_START:
                        ParXch_Start( Dev );
                        break;

                case IOCTL_PARXCH_STOP:
                        ParXch_Stop( Dev );
                        break;

                case IOCTL_PARXCH_READY:
                        *OutBuffer    = ParXch_DramFifoReady( Dev );
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_OVERFLOW:
                        *OutBuffer    = ParXch_DramFifoFull( Dev );
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_USER_LED:
                        ParXch_BdCtrlUserLed( Dev, *InBuffer );
                        break;

                case IOCTL_PARXCH_DIO_READ:
                        ParXch_UserIoRd( Dev, OutBuffer );
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_DIO_WRITE:
                        ParXch_UserIoWr( Dev, *InBuffer );
                        break;

                case IOCTL_PARXCH_GET_INTR:
                        *OutBuffer    = Dev->ParPortIrq;
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_INTR_ENABLE:
                        *OutBuffer = 1;
                        if ( Dev->ParPortIrq > 0 )
                                ParXch_InterruptEnable( Dev );
                        else
                                *OutBuffer = 0;
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_INTR_DISABLE:
                        *OutBuffer = 1;
                        if ( Dev->ParPortIrq > 0 )
                                ParXch_InterruptDisable( Dev );
                        else
                                *OutBuffer = 0;
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_ARM_DREADY_INTR:
                        ParXch_BdCtrlDrdyIrqReq( Dev );
                        break;

                case IOCTL_PARXCH_VOLTAGE_GOOD:
                        *OutBuffer    = ParXch_VoltageGood( Dev );
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_CHECK_HEALTH:
                        *OutBuffer = ParXch_CheckHealth( Dev );
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_PREPARE_TO_READ:
                        Tmp = *InBuffer;
                        Dev->IncludeMark    = (Tmp & PARXCH_REQUEST_GPS)     ? 1 : 0;
                        Dev->IncludeDigital = (Tmp & PARXCH_REQUEST_DIGITAL) ? 1 : 0;
                        Dev->IncludeCounter = (Tmp & PARXCH_REQUEST_COUNTER) ? 1 : 0;
                        *OutBuffer    = ParXch_CheckHealth( Dev );
                        ReturnedBytes = sizeof(int);
                        break;




                // These cases are for GPS support

                case IOCTL_PARXCH_ATTACH_GPS:
                        if ((void*)*InBuffer == NULL )
                                *OutBuffer = PARXCH_ERROR_GPS_NOT_ATTACHED;
                        else
                                *OutBuffer = ParXch_AttachGps( Dev,
                                                        (long *)*InBuffer );
                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_RELEASE_GPS:
                        if (Dev->SrCommonArea != NULL)
                                ParXch_ReleaseGps( Dev );
                        break;






                // The next cases are for diagnostic and debugging
                // use only and should not be called by regular users.

                case IOCTL_PARXCH_DIAG_SET_POWER_DOWN:
                        Dev->PowerDownOnExitFlag = *InBuffer;
                        break;

                case IOCTL_PARXCH_DIAG_BDCTRL_MODE:
                        ucharvalue = (unsigned char)(*InBuffer&DIAG_FLAG_VALUE);
                        ParXch_BdCtrlMode( Dev, ucharvalue );
                        break;

                case IOCTL_PARXCH_DIAG_FIFO_RESET:

                        // This is the only mode where the PC can talk
                        // to the FIFO.  Bad contention would happen if
                        // this call were done when the Xilinx part was
                        // in control of the FIFO.

                        if ( ( Dev->BdCtrlValue & (BDCTRL_MODE0|BDCTRL_MODE1) )
                              == BDCTRL_MODE_HALT_INIT ) {

                                ParXch_DramFifoReset( Dev );

                                *OutBuffer = 0x00;
                                if ( ParXch_DramFifoReady(Dev) )
                                        *OutBuffer |= DIAG_FLAG_READY;
                                if ( ParXch_DramFifoFull(Dev) )
                                        *OutBuffer |= DIAG_FLAG_FULL;

                                ReturnedBytes = sizeof(int);
                                }
                        else
                            Status = STATUS_INVALID_DEVICE_REQUEST;
                        break;

                case IOCTL_PARXCH_DIAG_FIFO_READ:

                        ParXch_DramFifoRd( Dev, &ucharvalue );

                        *OutBuffer = ucharvalue & DIAG_FLAG_VALUE;
                        if ( ParXch_DramFifoReady(Dev) )
                                *OutBuffer |= DIAG_FLAG_READY;
                        if ( ParXch_DramFifoFull(Dev) )
                                *OutBuffer |= DIAG_FLAG_FULL;

                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_DIAG_FIFO_WRITE:

                        ucharvalue = *InBuffer & DIAG_FLAG_VALUE;
                        ParXch_DramFifoWr( Dev, ucharvalue );

                        *OutBuffer = 0x00;
                        if ( ParXch_DramFifoReady(Dev) )
                                *OutBuffer |= DIAG_FLAG_READY;
                        if ( ParXch_DramFifoFull(Dev) )
                                *OutBuffer |= DIAG_FLAG_FULL;

                        ReturnedBytes = sizeof(int);
                        break;

                case IOCTL_PARXCH_DIAG_TURN_ON:
                        ParXch_TurnOn( Dev );
                        break;

                case IOCTL_PARXCH_DIAG_TURN_OFF:
                        ParXch_TurnOff( Dev );
                        break;

                case IOCTL_PARXCH_DIAG_POWER_UP:
                        ParXch_PowerUp( Dev );
                        break;

                case IOCTL_PARXCH_DIAG_POWER_DOWN:
                        ParXch_PowerDown( Dev );
                        break;


                // Raw in and out from port -- for hardware debugging only

                case IOCTL_PARXCH_DIAG_PORT_IN:
                        Pv = (PPORTVALUEXCH)InBuffer;
                        OsPortIn( Pv->Port, &ucharvalue );
                        Pv->Value = (int)ucharvalue;
                        ReturnedBytes = sizeof(PORTVALUEXCH);
                        break;

                case IOCTL_PARXCH_DIAG_PORT_OUT:
                        Pv = (PPORTVALUEXCH)InBuffer;
                        ucharvalue = (unsigned char)Pv->Value;
                        OsPortOut( Pv->Port, ucharvalue );
                        break;



                default:
                    ReturnedBytes = 0;
                    Status        = STATUS_INVALID_DEVICE_REQUEST;
                }



        // Return OutBuffer data to user space

        OsReturnIoctlBuffer( Irp, Command, OutBuffer, ReturnedBytes );

        OsIrpFinish( Irp, Status, ReturnedBytes );

        return( Status );
}

SRLOCAL void ParXch_DrvUnload( DEVINFO *Dev ) {

        dbg0_printk("In ParXch_DrvUnload\n");

        // Release common kernel area (should already be done)

        if (Dev->SrCommonArea != NULL)
                ParXch_ReleaseGps( Dev );
}








////////////////////////////////////////////////////////////////////////
/* Group 3: OS DEPENDENT INITIALIZATION AND DISPATCH FUNCTIONS:

The functions below are extremely system dependent as they involve very
OS specific initialization code.

For programmers interested in the low level mechanics of how it all
works, a brief overview follows:

When the indriver utility is run, the OS calls a driver
initialization function with a known name.  This function gets any
user supplied information, sets up pointers to the "dispatch"
functions for read and I/O control, and fills the Device Information
structure which holds global data for the driver.  For Win95 and DOS,
the Device Information structure is filled when the driver open is
called.

When the user calls a parxch.c library function, it calls an OS
function that in turn calls a driver function from parxchkd.c using
the previously set "dispatch" pointers.  For Win95 and DOS, there is
no intermediate OS function and the library functions call the driver
functions directly.

Most of the library function calls eventually get sent through
ParXch_DrvIoctl which has a big switch statement to select from the
many available functions.


A short comparison between some driver items for NT and Linux:

item description     Win2K/NT          Linux                 95/DOS
----------------     -----------       -----------           ------
init known name      DriverEntry       init_module           none
src of user info     registry          insmod command line   argument list
OS read function     ReadFile          read                  none
OS I/O ctrl func     DeviceIoControl   ioctl                 none

Note: For Linux 2.6, the module entry point name is no longer forced
to be init_module.  So, we choose DriverEntry to match Windows more
closely.

*/


// *************************************************************************

#if defined( SROS_WINXP ) || defined( SROS_WIN2K )

NTSTATUS DriverEntry( DRIVER_OBJECT *DriverObject,
                      UNICODE_STRING *RegistryPath ) {

        dbg0_printk("\nIn Wdm DriverEntry\n");


        // STEP 1: Set up dispatch table

        DriverObject->DriverExtension->AddDevice           = Wdm_DrvAddDevice;
        DriverObject->MajorFunction[IRP_MJ_PNP]            = Wdm_DrvPnp;
        DriverObject->MajorFunction[IRP_MJ_POWER]          = Wdm_DrvPower;
        DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = Wdm_DrvWmi;
        DriverObject->MajorFunction[IRP_MJ_CREATE]         = Wdm_DrvOpen;
        DriverObject->MajorFunction[IRP_MJ_CLOSE]          = Wdm_DrvClose;
        DriverObject->MajorFunction[IRP_MJ_READ]           = Wdm_DrvRead;
        DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = Wdm_DrvIoctl;
        DriverObject->DriverUnload                         = Wdm_DrvUnload;




        // STEP 2: Save registry path to service key


        //WCT: Is a PagedPool call like Oney did sufficient here ?

        DriverRegPath.MaximumLength = RegistryPath->Length + sizeof(WCHAR);
        DriverRegPath.Buffer = (PWSTR)ExAllocatePoolWithTag(
                                    PagedPool, DriverRegPath.MaximumLength, '1myS' );
        if (!DriverRegPath.Buffer)
                return( STATUS_INSUFFICIENT_RESOURCES );
        RtlCopyUnicodeString( &DriverRegPath, RegistryPath );


        return( STATUS_SUCCESS );
}

void Wdm_DrvUnload( DRIVER_OBJECT *DriverObject ) {

        dbg0_printk("In Wdm_DrvUnload\n");


        // This function undoes everything done in DriverEntry

        RtlFreeUnicodeString( &DriverRegPath ); // Does ExFreePoolWithTag for '1myS'

}

#define DEVICE_NAME_FORMAT_DOS  "\\??\\%s"
#define DEVICE_NAME_FORMAT_NT   "\\Device\\%s"
#define DEVICE_NAME_FORMAT_LINK "\\DosDevices\\%s"

NTSTATUS Wdm_DrvAddDevice( DRIVER_OBJECT *DriverObject, DEVICE_OBJECT *PDO ) {

        NTSTATUS          Status;
        WCHAR             ParXchNameBuffer[MAX_STR_LENGTH];
        unsigned char     AnsiDriverBuffer[MAX_STR_LENGTH];
        char              DeviceNameBuffer[MAX_STR_LENGTH];
        ANSI_STRING       AnsiDriverString;
        UNICODE_STRING    NtDeviceName, LinkDeviceName, ParXchName;
        DEVICE_OBJECT    *FDO;
        DEVINFO          *Dev;
        POWER_STATE       InitialPowerState;
        int               DeviceBase, DevicePortNum, DeviceIrq, DeviceFound;
        char             *DeviceName;

        dbg0_printk("In Wdm_DrvAddDevice\n");

        // DriverObject is same one as passed to DriverEntry
        //
        // PDO is address of the physical device object at the bottom
        // of the device stack (even though some lower filter driver
        // may already be on top of it).


        // Get driver name and port parameters from the registry

        DevicePortNum = 0;

        ParXchName.Length        = 0;
        ParXchName.MaximumLength = MAX_STR_LENGTH * sizeof(WCHAR);
        ParXchName.Buffer        = ParXchNameBuffer;

        Status = Wdm_GetParameters( &DriverRegPath,
                                    &ParXchName,
                                    &DeviceBase,
                                    &DeviceIrq,
                                    &DeviceFound );


        AnsiDriverString.Length        = 0;
        AnsiDriverString.MaximumLength = MAX_STR_LENGTH * sizeof(char);
        AnsiDriverString.Buffer        = DeviceNameBuffer;

        DeviceName = DeviceNameBuffer;

        RtlUnicodeStringToAnsiString( &AnsiDriverString, &ParXchName, FALSE );



        // STEP 1a: Create device (and device extension)

        // Note: type, characteristics and exclusive can be overridden
        // by values in the hardware key

        // Consider adding FILE_DEVICE_SECURE_OPEN to characteristics.pg56

        AnsiDriverString.Length        = 0;
        AnsiDriverString.MaximumLength = MAX_STR_LENGTH * sizeof(char);
        AnsiDriverString.Buffer        = AnsiDriverBuffer;
        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_NT, DeviceName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &NtDeviceName, &AnsiDriverString, TRUE);

        Status = IoCreateDevice(
                                DriverObject,
                                sizeof(DEVINFO),
                                &NtDeviceName,
                                FILE_DEVICE_UNKNOWN,
                                0,                      // Characteristics
//WCT: Allow for running more than 2 apps at one time in the future
//                              TRUE,                   // Exclusive (only 1)
                                FALSE,                  // NOT Exclusive (only 1)
                                &FDO
                               );

        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk("IoCreateDevice failed with status 0x%X\n", Status);
                RtlFreeUnicodeString( &NtDeviceName );
                return( Status );
                }



        // Hook DPC for ISR - not currently needed
        //        IoInitializeDpcRequest( FDO, DpcForIsr );




        // STEP 2: Register device interfaces OR
        //         Create name and symbolic link


        // Create symbolic link

        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_LINK, DeviceName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &LinkDeviceName, &AnsiDriverString, TRUE );

        Status = IoCreateSymbolicLink( &LinkDeviceName, &NtDeviceName );

        RtlFreeUnicodeString( &NtDeviceName );
        RtlFreeUnicodeString( &LinkDeviceName );

        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk("IoCreateSymbolicLink failed with status 0x%X\n", Status);
                Wdm_RemoveDevice( FDO ); // this calls IoDeleteDevice(FDO)
                return( Status );
                }






        // STEP 3: Initialize device extension

        Dev = (DEVINFO *)FDO->DeviceExtension;

        RtlZeroMemory( Dev, sizeof(DEVINFO) );
        ParXch_FillDevInfo( Dev, DeviceName, DeviceBase, 0, DeviceIrq );
        Dev->ParPortFound = DeviceFound;


        // Initialize spin lock for protecting port


        // Set up synchronization items like mutex and spinlock

        KeInitializeMutex( &Dev->ParPortAcquireMutex, 0 );

        KeInitializeSpinLock( &Dev->PortLock );
        KeInitializeSpinLock( &Dev->ADS1210InitLock );


        Dev->DriverObj     = DriverObject;
        Dev->ThisFdoDevObj = FDO;
        Dev->PdoDevObj     = PDO;

        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_LINK, DeviceName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &Dev->LinkName, &AnsiDriverString, TRUE );


        IoInitializeRemoveLock( &Dev->RemoveLock, '_MYS', 0, 999 );

        // WCT: Add device interface support later
        //        IoRegisterDeviceInterface( ..., &Dev->InterfaceName );


        // Inform power manager of initial power state

        InitialPowerState.DeviceState = Dev->CurrDevPower;
        PoSetPowerState( FDO, DevicePowerState, InitialPowerState );




        // STEP 4: Put device object into the stack

        Dev->LowerDevObj = IoAttachDeviceToDeviceStack( FDO, PDO );

        if (Dev->LowerDevObj == NULL) {
                Wdm_RemoveDevice( FDO );
                return( STATUS_NO_SUCH_DEVICE );
                }


        // Set device options for Read/Write data alignment and passing
        // and indicate done initializing

        FDO->AlignmentRequirement = FILE_BYTE_ALIGNMENT;
        FDO->Flags               |= DO_DIRECT_IO;
        FDO->Flags               &= ~DO_DEVICE_INITIALIZING;

        return( Status );
}

SRLOCAL NTSTATUS Wdm_GetParameters( UNICODE_STRING *RegistryPath,
                                    UNICODE_STRING *pParXchName,
                                    unsigned long  *pPortBase,
                                    unsigned long  *pPortIrq,
                                    unsigned long  *pPortFound ) {

        // Get PARXCH port parameters from the registry.

        NTSTATUS                 Status;
        WCHAR                   *queryPath;
        RTL_QUERY_REGISTRY_TABLE queryTable[5];  // Leave room for all 0's
        unsigned long            defaultBase  = DEFAULT_XCH_ADDRESS;
        unsigned long            defaultIrq   = DEFAULT_XCH_IRQ;
        unsigned long            defaultFound = 0;
        UNICODE_STRING           defaultName;


        RtlInitUnicodeString( &defaultName, L"SrParXch0" );


        // Generate wide character string having the \Parameters
        // subkey registry path to query

        queryPath = ExAllocatePoolWithTag( PagedPool, MAX_STR_LENGTH, '2myS' );

        if (queryPath == NULL) {
                dbg_printk("No memory in Wdm_GetParameters\n");
                return( STATUS_NO_MEMORY );
                }

        RtlZeroMemory( queryPath, MAX_STR_LENGTH );
        RtlMoveMemory( queryPath, RegistryPath->Buffer, RegistryPath->Length );
        wcscat( queryPath, L"\\Parameters" );



        // Set up table of parameters to query

        RtlZeroMemory( &queryTable[0], sizeof(queryTable) );

        queryTable[0].Flags         = RTL_QUERY_REGISTRY_DIRECT;
        queryTable[0].Name          = LREG_PARM_DRIVERNAME;
        queryTable[0].EntryContext  = pParXchName;
        queryTable[0].DefaultType   = REG_SZ;
        queryTable[0].DefaultData   = &defaultName;
        queryTable[0].DefaultLength = 0; // compute value based on default

        queryTable[1].Flags         = RTL_QUERY_REGISTRY_DIRECT;
        queryTable[1].Name          = LREG_PARM_PORTADDR;
        queryTable[1].EntryContext  = pPortBase;
        queryTable[1].DefaultType   = REG_DWORD;
        queryTable[1].DefaultData   = &defaultBase;
        queryTable[1].DefaultLength = sizeof(unsigned long);

        queryTable[2].Flags         = RTL_QUERY_REGISTRY_DIRECT;
        queryTable[2].Name          = LREG_PARM_PORTIRQ;
        queryTable[2].EntryContext  = pPortIrq;
        queryTable[2].DefaultType   = REG_DWORD;
        queryTable[2].DefaultData   = &defaultIrq;
        queryTable[2].DefaultLength = sizeof(unsigned long);

        queryTable[3].Flags         = RTL_QUERY_REGISTRY_DIRECT;
        queryTable[3].Name          = LREG_PARM_PORTFOUND;
        queryTable[3].EntryContext  = pPortFound;
        queryTable[3].DefaultType   = REG_DWORD;
        queryTable[3].DefaultData   = &defaultFound;
        queryTable[3].DefaultLength = sizeof(unsigned long);

        queryTable[4].QueryRoutine  = NULL; // This [3] entry is already zeroed
        queryTable[4].Name          = NULL; // and marks the end of the table


        // Do the query

        Status = RtlQueryRegistryValues(
                      RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
                      queryPath, &queryTable[0], NULL, NULL );


        // Set defaults if the query failed

        if ( !NT_SUCCESS(Status) ) {
                *pPortBase = defaultBase;
                *pPortIrq  = defaultIrq;
                *pPortFound = defaultFound;

                dbg_printk("Parameter query failed, using defaults\n");

//                Wdm_ErrorLog( NULL, NULL, Status, PARXCH_LOG_INFO,
//                              L"using default port parameters" );
                }

        ExFreePoolWithTag( queryPath, '2myS' );


        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_ErrorLog( void *IoObject,
                               IRP *Irp,
                               NTSTATUS CurrStatus,
                               unsigned long ErrCode,
                               const WCHAR *Msg ) {

        IO_ERROR_LOG_PACKET *errorLogEntry;
        IO_STACK_LOCATION   *IrpStack;
        unsigned long        CtrlCode;
        unsigned char        MajorFunction, len1, len2;


        // Initialize entry variables

        CtrlCode      = 0;
        MajorFunction = 0;

        if (Irp) {
                IrpStack      = IoGetCurrentIrpStackLocation( Irp );
                MajorFunction = IrpStack->MajorFunction;
                if (MajorFunction == IRP_MJ_DEVICE_CONTROL)
                        CtrlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
                }



        // Calculate packet length including a NULL terminator for the string

        len1 = (unsigned char)( (wcslen( Msg ) + 1) * sizeof(WCHAR) );
        len2 = (unsigned char)( sizeof(IO_ERROR_LOG_PACKET) + len1 );

        if (len2 > ERROR_LOG_MAXIMUM_SIZE) {
                len2 = ERROR_LOG_MAXIMUM_SIZE;
                len1 = len2 - sizeof(IO_ERROR_LOG_PACKET);
                if (len1 <= 0)
                        return( STATUS_NO_MEMORY );
                }




        // Allocate the packet

        errorLogEntry = IoAllocateErrorLogEntry( IoObject, len2 );

        if (!errorLogEntry)
                return( STATUS_NO_MEMORY );



        // Fill in the packet

        errorLogEntry->ErrorCode         = ErrCode;
        errorLogEntry->NumberOfStrings   = 1 ;
        errorLogEntry->StringOffset      = FIELD_OFFSET(
                                           IO_ERROR_LOG_PACKET, DumpData );
        errorLogEntry->MajorFunctionCode = MajorFunction;
        errorLogEntry->IoControlCode     = CtrlCode;
        errorLogEntry->UniqueErrorValue  = 0;
        errorLogEntry->SequenceNumber    = 0;
        errorLogEntry->FinalStatus       = CurrStatus;




        // Copy the message string to the packet

        RtlCopyMemory( &errorLogEntry->DumpData[0], Msg, len1 );



        // Write the packet to the event log

        IoWriteErrorLogEntry( errorLogEntry );



        return( STATUS_SUCCESS );
}


SRLOCAL NTSTATUS Wdm_RemoveDevice( DEVICE_OBJECT *FDO ) {

        DEVINFO *Dev;

        dbg0_printk("In Wdm_RemoveDevice\n");

        // This function undoes everything done by DrvAddDevice


        if (!FDO)
                return( STATUS_UNSUCCESSFUL );

        Dev = (DEVINFO *)FDO->DeviceExtension;





        // Detach the current device from the driver stack

        if ( Dev->LowerDevObj ) {
                IoDetachDevice( Dev->LowerDevObj );
                Dev->LowerDevObj = NULL;
                }


        // Remove symbolic link name

        if ( Dev->LinkName.Length > 0 ) {
                IoDeleteSymbolicLink( &Dev->LinkName );
                RtlFreeUnicodeString( &Dev->LinkName );
                Dev->LinkName.Length = 0;
                }

        // Delete the current device
        // Careful, in Win98 IoDeleteDevice calls DriverUnload !!

        IoDeleteDevice( FDO );


        return( STATUS_SUCCESS );

}

// Interrupt Function

BOOLEAN Wdm_DrvInterrupt( PKINTERRUPT pInterrupt, void *Context ) {

        DEVINFO *Dev;

        // If this device not interrupting, return FALSE

        if (Context == NULL)
                return( FALSE );

        Dev = (DEVINFO *)Context;

        return( (BOOLEAN)ParXch_DrvInterrupt( Dev ) );
}


// Dispatch Functions

NTSTATUS Wdm_DrvOpen( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;

        // Verify that an underlying Parport has truely been found
        // before we allow the user to work with this device.

        Dev = (DEVINFO *)FDO->DeviceExtension;

        if ( Dev->ParPortAcquired == 0 ) {
                dbg_printk( "Missing a parallel port! 0x%X\n", STATUS_PORT_CONNECTION_REFUSED);
                OsIrpFinish( Irp, STATUS_PORT_CONNECTION_REFUSED, 0 );
                return( STATUS_PORT_CONNECTION_REFUSED );
                }

        if ( Dev->StartDone == 0 ) {
                dbg_printk( "Driver not started! 0x%X\n", STATUS_PORT_CONNECTION_REFUSED);
                OsIrpFinish( Irp, STATUS_PORT_CONNECTION_REFUSED, 0 );
                return( STATUS_PORT_CONNECTION_REFUSED );
                }


        return( (NTSTATUS)ParXch_DrvOpen(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}

NTSTATUS Wdm_DrvClose( DEVICE_OBJECT *FDO, IRP *Irp ) {
        return( (NTSTATUS)ParXch_DrvClose(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}

NTSTATUS Wdm_DrvRead( DEVICE_OBJECT *FDO, IRP *Irp ) {
        return( (NTSTATUS)ParXch_DrvRead(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}

NTSTATUS Wdm_DrvIoctl( DEVICE_OBJECT *FDO, IRP *Irp ) {
        return( (NTSTATUS)ParXch_DrvIoctl(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}





// Plug and Play Functions

NTSTATUS Wdm_DrvPnp( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;
        IO_STACK_LOCATION *IrpStack;
        ULONG MinorFn;

        dbg2_printk("In Wdm_DrvPnp\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        MinorFn  = IrpStack->MinorFunction;


        switch( MinorFn ) {
                case IRP_MN_START_DEVICE:
                        Status = Wdm_PnpStartDevice( FDO, Irp );
                        break;
                case IRP_MN_STOP_DEVICE:
                        Status = Wdm_PnpStopDevice( FDO, Irp );
                        break;
                case IRP_MN_REMOVE_DEVICE:
                        Status = Wdm_PnpRemoveDevice( FDO, Irp );
                        break;
                case IRP_MN_QUERY_CAPABILITIES:
                        Status = Wdm_PnpQueryCapabilities( FDO, Irp );
                        break;
                case IRP_MN_CANCEL_STOP_DEVICE:
                        Status = Wdm_PnpCancelStop( FDO, Irp );
                        break;
                case IRP_MN_CANCEL_REMOVE_DEVICE:
                        Status = Wdm_PnpCancelRemove( FDO, Irp );
                        break;
                case IRP_MN_QUERY_STOP_DEVICE:
                        Status = Wdm_PnpQueryStop( FDO, Irp );
                        break;
                case IRP_MN_QUERY_REMOVE_DEVICE:
                        Status = Wdm_PnpQueryRemove( FDO, Irp );
                        break;
                case IRP_MN_QUERY_DEVICE_RELATIONS:
                        Status = Wdm_PnpQueryDevRelations( FDO, Irp );
                        break;
                case IRP_MN_QUERY_INTERFACE:
                        Status = Wdm_PnpQueryInterface( FDO, Irp );
                        break;
                case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:
                        Status = Wdm_PnpFilterResourceReq( FDO, Irp );
                        break;
                case IRP_MN_SURPRISE_REMOVAL:
                        Status = Wdm_PnpSurpriseRemoval( FDO, Irp );
                        break;
                default:
                        if (MinorFn == IRP_MN_QUERY_PNP_DEVICE_STATE)
                                dbg2_printk("PnP QUERY_PNP_DEVICE_STATE\n");
//                        else if (MinorFn == IRP_MN_QUERY_LEGACY_BUS_INFORMATION)
//                                dbg2_printk("PnP QUERY_LEGACY_BUS_INFORMATION\n");
                        else
                                dbg2_printk("Pnp MinorFn 0x%X\n", MinorFn);
                        Status = Wdm_IrpForward( FDO, Irp );
                }

        dbg2_printk("Exit Wdm_DrvPnp with Status 0x%lX\n", Status);
        return( Status );
}

NTSTATUS Wdm_PnpStartDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;

        dbg2_printk("In Wdm_PnpStartDevice\n");


        // Let lower drivers do their starting stuff first.

        Irp->IoStatus.Status = STATUS_SUCCESS;
        Status = Wdm_IrpForwardAndWait( FDO, Irp );
        if ( !NT_SUCCESS( Status )) {
                OsIrpFinish( Irp, Status, Irp->IoStatus.Information );
                dbg2_printk("Exit Wdm_PnpStartDevice lower drivers failed\n");
                return( Status );
                }


        // We have now been assigned resources by the PnP Manager.  So
        // now do our starting work of allocating them and starting the
        // device.

        Wdm_ResourceSet( FDO, Irp );



        Dev = (DEVINFO *)FDO->DeviceExtension;



        // We would like to exclusively allocate the parallel port in
        // the PnpStartDevice routine so we can proceed to do our
        // startup work, but on XP reboots there is no guarantee of the
        // start order between this driver and the parallel port it
        // depends on.  So, here we register for parallel port PnP
        // interface changes and do all our allocating and starting work
        // in the PnpNotify function that gets called when a Parport
        // arrives or leaves.

        Status = IoRegisterPlugPlayNotification(
             EventCategoryDeviceInterfaceChange,                       // category type
             PNPNOTIFY_DEVICE_INTERFACE_INCLUDE_EXISTING_INTERFACES,   // flags
             (PVOID)&GUID_PARALLEL_DEVICE,                             // interface GUID
             FDO->DriverObject,                                        // driver
             (PDRIVER_NOTIFICATION_CALLBACK_ROUTINE)Wdm_PnpNotifyInterfaceChange, // callback
             (PVOID)FDO,                                               // context
             &Dev->NotificationHandle                                  // notification
              );

        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk("IoRegisterPlugPlayNotification InterfaceChange FAILED, status= %x\n", Status );
                swprintf( Dev->MsgBuffer, L"IoRegPnpNotify failure 0x%lX", Status );
                Wdm_ErrorLog( FDO, NULL, Status, PARXCH_LOG_INFO, Dev->MsgBuffer );
                OsIrpFinish( Irp, Status, Irp->IoStatus.Information );
                return( Status );
                }


        // Handle the case where NO parport is currently ready.

        if ( Dev->ParPortAcquired == 0 ) {

                // If the registry indicates that we have not found a valid
                // parallel port before, we expect we are being installed
                // by the user on a system with all parallel ports already
                // started.  Therefore, this indicates a problem (probably
                // a user typo).  But there is a chance that a queued work
                // item getting the port has not finished yet, so don't fail
                // if there are any work items in progress.

                if ( !Dev->ParPortFound ) {
                        dbg0_printk("No parport found on user install\n");
                        swprintf( Dev->MsgBuffer, L"No parport found on user install\n" );
                        Wdm_ErrorLog( FDO, NULL, Status, PARXCH_LOG_INFO, Dev->MsgBuffer );

                        if (Dev->WorkItemCount == 0) {
                                Status = STATUS_PORT_CONNECTION_REFUSED;
                                OsIrpFinish( Irp, Status, Irp->IoStatus.Information );
                                dbg0_printk("Exit Wdm_PnpStartDevice no port found and no work items pending\n");
                                return( Status );
                                }
                        else {
                                dbg0_printk("Exit Wdm_PnpStartDevice no port found BUT %ld work items still pending\n", Dev->WorkItemCount);
                                }
                        }

                // Otherwise, the registry indicates that a parallel port
                // was found at least once, so we expect we are in a reboot
                // situation.  Therefore, we are willing to put off our
                // actual hardware starting behavior until the underlying
                // Parport actually arrives and is started.

                else { // ( Dev->ParPortFound ) {
                        dbg0_printk("No parport found yet on reboot\n");
                        swprintf( Dev->MsgBuffer, L"No parport found yet on reboot\n" );
                        Wdm_ErrorLog( FDO, NULL, Status, PARXCH_LOG_INFO, Dev->MsgBuffer );
                        }

                } // end if no parport ready


        // Done

        OsIrpFinish( Irp, Status, Irp->IoStatus.Information );
        dbg2_printk("Exit Wdm_PnpStartDevice with Status 0x%lX\n", Status);
        return( Status );
}
SRLOCAL NTSTATUS Wdm_ResourceSet( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;
        ULONG i, nr, vector;
        PHYSICAL_ADDRESS portbase;
        KIRQL irql;
        KINTERRUPT_MODE mode;
        KAFFINITY affinity;
        BOOLEAN irqshare;
        CM_PARTIAL_RESOURCE_DESCRIPTOR *resource;
        IO_STACK_LOCATION *IrpStack;
        CM_RESOURCE_LIST *rlist;
        CM_PARTIAL_RESOURCE_LIST *translated;

        dbg0_printk("In Wdm_ResourceSet\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;



        // Set defaults

        Dev->ParPortMapped = 0;
        Dev->ParPortCount  = 0;
        portbase.HighPart  = 0L;
        portbase.LowPart   = Dev->ParPortBase;


//WCT: We don't expect to get ANY resources since we use ParPort
//     and it has the resources allocated.


        // Get translated resources passed by PnP manager

        IrpStack = IoGetCurrentIrpStackLocation( Irp );

        rlist = IrpStack->Parameters.StartDevice.AllocatedResourcesTranslated;
        if (rlist == NULL || rlist->Count == 0) {
                dbg0_printk("No translated resources\n");
                translated = NULL;
                nr         = 0;
                resource   = NULL;
                }
        else {
                translated = &rlist->List[0].PartialResourceList;
                nr         = translated->Count;
                resource   = translated->PartialDescriptors;
                }



        // Extract the resources that have been allocated to this device.


        for ( i = 0 ; i < nr ; i++ ) {

            switch( resource->Type ) {

                case CmResourceTypePort:
                        portbase          = resource->u.Port.Start;
                        Dev->ParPortCount = resource->u.Port.Length;
                        if ( (resource->Flags & CM_RESOURCE_PORT_IO) == 0 )
                                Dev->ParPortMapped = 1;
                        else
                                Dev->ParPortMapped = 0;
                        break;

                case CmResourceTypeInterrupt:
                        irql     = (KIRQL)resource->u.Interrupt.Level;
                        vector   =        resource->u.Interrupt.Vector;
                        affinity =        resource->u.Interrupt.Affinity;
                        if (resource->Flags == CM_RESOURCE_INTERRUPT_LATCHED)
                                mode = Latched;
                        else
                                mode = LevelSensitive;
                        if (resource->ShareDisposition == CmResourceShareShared)
                                irqshare = 1;
                        else
                                irqshare = 0;
                        break;

                case CmResourceTypeMemory:
                case CmResourceTypeDma:
                default:
                        break;

                } // end switch

            resource++;

            } // end for i



        // Map I/O port if needed

        if (Dev->ParPortMapped) {
                dbg1_printk("Port is Mapped\n");
                Dev->ParPortBase = (int)MmMapIoSpace( portbase,
                                                      Dev->ParPortCount,
                                                      MmNonCached );
                if (!Dev->ParPortBase)
                        return( STATUS_NO_MEMORY );
                }
        else {
                if (portbase.HighPart != 0L)
                        dbg0_printk("WARNING: Ignoring high part of port base\n");
                Dev->ParPortBase = (int)portbase.LowPart;
                }


        return( STATUS_SUCCESS );
}

NTSTATUS Wdm_PnpNotifyInterfaceChange(
                PDEVICE_INTERFACE_CHANGE_NOTIFICATION NotificationStruct,
                PDEVICE_OBJECT                        FDO
                ) {

        DEVINFO            *Dev;
        PNP_NOTIFY_CONTEXT *PnpNotifyContext;
        int                 nchar;

        dbg0_printk( "Wdm_PnpNotifyInterfaceChange Enter\n" );



        // Ignore any interface class that is not a ParPort device interface.

        if ( !IsEqualGUID( (LPGUID)&(NotificationStruct->InterfaceClassGuid),
                           (LPGUID)&GUID_PARALLEL_DEVICE) ) {
                dbg0_printk("PnpNotifyInterfaceChange ERROR - Bad InterfaceClassGuid\n");
                return( STATUS_SUCCESS );
                }


        Dev = (DEVINFO *)FDO->DeviceExtension;


//        dbg2_printk("GUID_PARALLEL_DEVICE interface notification\n");
//        dbg2_printk( "Notification for %ls\n", NotificationStruct->SymbolicLinkName->Buffer);
//
//        dbg2_printk( "Context is 0x%X, Extension is 0x%X\n", FDO, Dev );
//        dbg2_printk( "Link name is %ls\n", Dev->LinkName.Buffer );




        // Process Arrivals
        //
        // We want to acquire exclusive access to the arriving port, but to
        // do this we need to call IoGetObjectPointer which might block and
        // cause a deadlock.  So we queue up a work item to do this job for
        // us.

        if (IsEqualGUID( (LPGUID)&(NotificationStruct->Event),
                         (LPGUID)&GUID_DEVICE_INTERFACE_ARRIVAL )) {


                dbg0_printk( "Wdm_PnpNotifyInterfaceChange Notify Type = Arrival\n" );


                // Just ignore arrivals once we have found a port we like

                if ( Dev->ParPortAcquired == 1 ) {
                        dbg0_printk( "Wdm_PnpNotifyInterfaceChange Exit since have port\n" );
                        return( STATUS_SUCCESS );
                        }



                // Prepare context info to be passed to work item callback
                // It includes the work item and a string with the symbolic
                // link name of the notification structure (ie the name of
                // the parallel port that just arrived).

                PnpNotifyContext = (PNP_NOTIFY_CONTEXT *) ExAllocatePoolWithTag(NonPagedPool, sizeof(PNP_NOTIFY_CONTEXT), '3myS');
                if (!PnpNotifyContext) {
                        dbg0_printk( "Wdm_PnpNotifyInterfaceChange Exit Error Context\n" );
                        return( STATUS_INSUFFICIENT_RESOURCES );
                        }

                PnpNotifyContext->WorkItem = IoAllocateWorkItem( Dev->ThisFdoDevObj );
                if (!PnpNotifyContext->WorkItem) {
                        ExFreePoolWithTag( PnpNotifyContext, '4myS' );
                        dbg0_printk( "Wdm_PnpNotifyInterfaceChange Exit Error WorkItem\n" );
                        return( STATUS_INSUFFICIENT_RESOURCES );
                        }

                nchar = NotificationStruct->SymbolicLinkName->Length;
                if (nchar > MAX_STR_LENGTH) {
                        IoFreeWorkItem( PnpNotifyContext->WorkItem );
                        ExFreePoolWithTag( PnpNotifyContext, '3myS' );
                        dbg0_printk( "Wdm_PnpNotifyInterfaceChange Exit Error Name\n" );
                        return( STATUS_INSUFFICIENT_RESOURCES );
                        }

                RtlZeroMemory( PnpNotifyContext->NotifySymLink, MAX_STR_LENGTH * sizeof(WCHAR) );
                RtlCopyMemory( PnpNotifyContext->NotifySymLink,
                               NotificationStruct->SymbolicLinkName->Buffer,
                               nchar );


                // Count and queue the work item

                InterlockedIncrement( &Dev->WorkItemCount );

                dbg0_printk( "Wdm_PnpNotifyInterfaceChange Queueing work item %d with context 0x%X\n",
                            Dev->WorkItemCount, PnpNotifyContext );

                IoQueueWorkItem( PnpNotifyContext->WorkItem,
                                 Wdm_WorkItemNotifyCallback,
                                 DelayedWorkQueue,              // FIX was CriticalWorkQueue,
                                 PnpNotifyContext
                               );


                }


        // Process Removals

        else if (IsEqualGUID( (LPGUID)&(NotificationStruct->Event),
                         (LPGUID)&GUID_DEVICE_INTERFACE_REMOVAL )) {
                dbg0_printk( "Type = Interface Removal\n" );
                }


        else {
                dbg0_printk( "Type = Interface Unknown\n" );
                }

        dbg0_printk( "Wdm_PnpNotifyInterfaceChange Exit\n" );
        return( STATUS_SUCCESS );
}


// PIO_WORKITEM_ROUTINE

VOID Wdm_WorkItemNotifyCallback( PDEVICE_OBJECT FDO, PVOID Context ) {

        PNP_NOTIFY_CONTEXT *PnpNotifyContext;
        DEVINFO            *Dev;
        UNICODE_STRING      DriverName;
        NTSTATUS            Status;

        dbg0_printk( "Wdm_WorkItemNotifyCallback Enter with Context 0x%X\n", Context );

        // Verify arguments

        if ( !FDO || !Context ) {
                dbg_printk("Notify work item called with bad parameters\n" );
                if (FDO) {
                        Wdm_ErrorLog( FDO, NULL, STATUS_INVALID_PARAMETER, PARXCH_LOG_INFO,
                                      L"Notify work item with bad parameters" );
                        }
                Wdm_WorkItemNotifyCleanup( (DEVINFO *)NULL,
                                           (PNP_NOTIFY_CONTEXT *)Context,
                                           FALSE );
                return;
                }


        Dev              = (DEVINFO *)FDO->DeviceExtension;
        PnpNotifyContext = (PNP_NOTIFY_CONTEXT *)Context;



        // If we've already found a port we like, do cleanup and exit

        if ( Dev->ParPortAcquired == 1 ) {
                Wdm_WorkItemNotifyCleanup( Dev, PnpNotifyContext, FALSE );
                dbg0_printk( "Wdm_WorkItemNotifyCallback Exit Port already found Context 0x%X\n", Context );
                return;
                }


        // If several parallel ports are available in the PC, it is important
        // to work with only one port at a time when trying to acquire it.
        // So we wait until the mutex is signaled before continuing.

        dbg0_printk( "Wdm_WorkItemNotifyCallback Begin wait Context 0x%X\n", Context );

        Status = KeWaitForMutexObject(
                                      &Dev->ParPortAcquireMutex, // mutex
                                      Executive,                 // wait reason
                                      KernelMode,                // wait mode
                                      FALSE,                     // not alertable
                                      NULL                       // Timeout  OPTIONAL
                                     );

        if ( Status != STATUS_SUCCESS ) {
                dbg0_printk("Wdm_WorkItemNotifyCallback Exit Error Mutex Status 0x%X Context 0x%X\n", Status, Context );
                Wdm_WorkItemNotifyCleanup( Dev, PnpNotifyContext, TRUE );
                return;
                }

        dbg0_printk( "Wdm_WorkItemNotifyCallback Past wait Context 0x%X\n", Context );


        // Now that it's our turn, check again to be sure we don't have a port already

        if ( Dev->ParPortAcquired == 1 ) {
                dbg0_printk("Wdm_WorkItemNotifyCallback Exit Port already found after wait Context 0x%X\n", Context );
                Wdm_WorkItemNotifyCleanup( Dev, PnpNotifyContext, TRUE );
                return;
                }



        // Try to acquire exclusive access to the parallel port

        RtlInitUnicodeString( &DriverName, PnpNotifyContext->NotifySymLink );

        Status = Wdm_ParPortAcquire( Dev, Dev->ParPortBase,
                                     Dev->ParPortIrq, &DriverName, Context );


        // Got it, now start the port

        if ( NT_SUCCESS( Status ) ) {
                dbg0_printk("Found parport 0x%X in notify Context 0x%X\n", Dev->ParPortBase, Context );
                Wdm_ErrorLog( FDO, NULL, Status, PARXCH_LOG_INFO,
                              L"Found parport in notify workitem" );

                Wdm_StartDevice( FDO );
                }

        // Failed to acquire port

        else {
                dbg1_printk( "Wdm_WorkItemNotifyCallback failed to get valid device obj ptr 0x%X for port 0x%X Context 0x%X\n", Status, Dev->ParPortBase, Context );
                }


        // Cleanup

        Wdm_WorkItemNotifyCleanup( Dev, PnpNotifyContext, TRUE );
        dbg0_printk( "Wdm_WorkItemNotifyCallback Exit Context 0x%X\n", Context );
        return;
}

VOID Wdm_WorkItemNotifyCleanup( DEVINFO            *Dev,
                                PNP_NOTIFY_CONTEXT *PnpNotifyContext,
                                BOOLEAN             FreeMutex ) {

        LONG ReleaseStatus;


        if ( FreeMutex && Dev )
                ReleaseStatus = KeReleaseMutex( &Dev->ParPortAcquireMutex, FALSE );

        if ( Dev )
                InterlockedDecrement( &Dev->WorkItemCount );

        if ( PnpNotifyContext ) {
                IoFreeWorkItem( PnpNotifyContext->WorkItem );
                ExFreePoolWithTag( PnpNotifyContext, '3myS' );
                }
}

SRLOCAL NTSTATUS Wdm_ParPortAcquire( DEVINFO *Dev,
                                     unsigned long DeviceBase,
                                     unsigned long DeviceIrq,
                                     UNICODE_STRING *PptName,
                                     VOID *Context ) {

        NTSTATUS                            Status;
        HANDLE                              hKey;
        UNICODE_STRING                      valueName;
        ULONG                               YesConnect, PptAddress;
        PARALLEL_PORT_INFORMATION           ParallelPortInfo;
        PARALLEL_PNP_INFORMATION            ParallelPnpInfo;
        PARALLEL_CHIP_MODE                  ChipMode;
        PARALLEL_INTERRUPT_SERVICE_ROUTINE  ParallelIsr;
        PARALLEL_INTERRUPT_INFORMATION      ParallelIsrInfo;
        DEVICE_OBJECT                      *PPDevObj, *ParportPDO;
        FILE_OBJECT                        *PPFileObj;
        KINTERRUPT                         *PPIrqObj;


        dbg0_printk( "Wdm_ParPortAcquire Enter Port name %ls, Context 0x%X\n", PptName->Buffer, Context );

        // Initialize variables

        PPDevObj  = NULL;
        PPFileObj = NULL;
        PPIrqObj  = NULL;


        // Get a pointer to parport, the parallel port device.
        // We choose the one which matches our address.

        Status = IoGetDeviceObjectPointer(
                                          PptName,
                                          STANDARD_RIGHTS_ALL,
                                          &PPFileObj,
                                          &PPDevObj
                                         );

        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk("Wdm_ParPortAcquire Exit Error DevObjPtr Context 0x%X\n", Context );
                return( STATUS_PORT_CONNECTION_REFUSED );
                }




        // Use pointer to get connection with port info

        Status = Wdm_IrpBuildSendAndWait(
                    IOCTL_INTERNAL,
                    IOCTL_INTERNAL_GET_PARALLEL_PORT_INFO,
                    PPDevObj,
                    NULL,
                    0,
                    &ParallelPortInfo,
                    sizeof( PARALLEL_PORT_INFORMATION ),
                    NULL
                    );

        if ( !NT_SUCCESS( Status ) ) {
                Wdm_ParPortRelease( Dev, PPDevObj, PPFileObj, PPIrqObj );
                dbg0_printk("Wdm_ParPortAcquire Exit Error no Parallel port info 0x%lX Context 0x%X\n", Status, Context);
                return( STATUS_PORT_CONNECTION_REFUSED );
                }



        // Save actual port address and clean up connection

        PptAddress = (ULONG)ParallelPortInfo.Controller;

        Status = Wdm_IrpBuildSendAndWait(
                    IOCTL_INTERNAL,
                    IOCTL_INTERNAL_RELEASE_PARALLEL_PORT_INFO,
                    PPDevObj,
                    NULL,
                    0,
                    NULL,
                    0,
                    NULL
                    );
        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk("Wdm_ParPortAcquire Failed to release Parallel port info 0x%lX Context 0x%X\n", Status, Context);
                }


        // We found a match!

        if ( PptAddress == DeviceBase ) {
                dbg0_printk("Wdm_ParPortAcquire match at 0x%lX, Context 0x%X\n", DeviceBase, Context);
                }

        // Otherwise, wrong port clean up

        else {
                Wdm_ParPortRelease( Dev, PPDevObj, PPFileObj, PPIrqObj );
                dbg0_printk("Wdm_ParPortAcquire Exit 0x%lX does NOT match 0x%lX\n", PptAddress, DeviceBase);
                swprintf( Dev->MsgBuffer,
                          L"Failed to find a parallel port at 0x%lX (it is at 0x%lX)",
                          DeviceBase, PptAddress);
                dbg0_printk("%ls  (Status = 0x%lX)\n", Dev->MsgBuffer, Status );
                Wdm_ErrorLog( Dev->ThisFdoDevObj, NULL, Status, PARXCH_LOG_INFO,
                              Dev->MsgBuffer );
                dbg0_printk( "Wdm_ParPortAcquire Exit Error Match 0x%lX NOT 0x%lX Context 0x%X\n", PptAddress, DeviceBase, Context);
                return( STATUS_PORT_CONNECTION_REFUSED );
                }




        // Reserve the parport device for our EXCLUSIVE access

        Status = Wdm_IrpBuildSendAndWait(
                    IOCTL_INTERNAL,
                    IOCTL_INTERNAL_PARALLEL_PORT_ALLOCATE,
                    PPDevObj,
                    NULL,
                    0,
                    NULL,
                    0,
                    NULL
                    );

        if ( !NT_SUCCESS( Status ) ) {
                Wdm_ErrorLog( Dev->ThisFdoDevObj, NULL, Status, PARXCH_LOG_INFO,
                              L"Can't allocate parallel port.  Already in use" );
                Wdm_ParPortRelease( Dev, PPDevObj, PPFileObj, PPIrqObj );
                dbg0_printk( "Wdm_ParPortAcquire Exit Error Allocate 0x%lX, Context 0x%X\n", Status, Context);
                return( Status );
                }



        // Get current mode info

        Status = Wdm_IrpBuildSendAndWait(
                    IOCTL_INTERNAL,
                    IOCTL_INTERNAL_GET_PARALLEL_PNP_INFO,
                    PPDevObj,
                    NULL,
                    0,
                    &ParallelPnpInfo,
                    sizeof(PARALLEL_PNP_INFORMATION),
                    NULL
                    );

        if ( !NT_SUCCESS( Status ) ) {
                dbg2_printk("Wdm_ParPortAcquire Failed to get pnp info 0x%lX Context 0x%X\n", Status, Context);
                }



        // Using the current mode, clear the chip mode to compatibility

        ChipMode.ModeFlags = (UCHAR)ParallelPnpInfo.CurrentMode;

        Status = Wdm_IrpBuildSendAndWait(
                    IOCTL_INTERNAL,
                    IOCTL_INTERNAL_PARALLEL_CLEAR_CHIP_MODE,
                    PPDevObj,
                    &ChipMode,
                    sizeof(PARALLEL_CHIP_MODE),
                    NULL,
                    0,
                    NULL
                    );

        if ( !NT_SUCCESS( Status ) ) {
                dbg2_printk("Wdm_ParPortAcquire Failed to clear chip mode 0x%lX Context 0x%X\n", Status, Context);
                }



        // Set chip mode to BPP (or EPP or ECP)

        if (ParallelPnpInfo.HardwareCapabilities & PPT_BYTE_PRESENT)
                ChipMode.ModeFlags = ECR_BYTE_MODE;

        else if (ParallelPnpInfo.HardwareCapabilities & PPT_EPP_PRESENT)
                ChipMode.ModeFlags = ECR_EPP_MODE;

        else if (ParallelPnpInfo.HardwareCapabilities & PPT_ECP_PRESENT)
                ChipMode.ModeFlags = ECR_ECP_MODE;
        else
                ChipMode.ModeFlags = ECR_BYTE_MODE;


        Status = Wdm_IrpBuildSendAndWait(
                    IOCTL_INTERNAL,
                    IOCTL_INTERNAL_PARALLEL_SET_CHIP_MODE,
                    PPDevObj,
                    &ChipMode,
                    sizeof(PARALLEL_CHIP_MODE),
                    NULL,
                    0,
                    NULL
                    );

        if ( !NT_SUCCESS( Status ) ) {
                dbg2_printk("Wdm_ParPortAcquire Failed to set 0x%x chip mode 0x%lX Context 0x%X\n",ChipMode.ModeFlags, Status, Context);
                }




        // Connect parport to our interrupt service routine.

        if ( DeviceIrq != 0 ) {     // Interrupt requested

                // Get the ParPort device PDO so we can find the location
                // for the registry key we must set so ParPort will allow
                // interrupt connections.

                Status = Wdm_ParPortGetPdo( Dev, PPDevObj, PPFileObj, &ParportPDO );
                if ( !NT_SUCCESS( Status ) ) {
                        dbg0_printk("Failed to get parport pdo 0x%lX\n", Status);
                        Wdm_ParPortRelease( Dev, PPDevObj, PPFileObj, PPIrqObj );
                        return( Status );
                        }
                dbg2_printk("Parport PDO is 0x%lX\n", ParportPDO );



                // Set the ParPort's EnableConnectInterruptIoctl key

                dbg2_printk("OpenDeviceRegistryKey DEVICE for Parport PDO 0x%lX\n", ParportPDO);
                Status = IoOpenDeviceRegistryKey(
                                                 ParportPDO,
                                                 PLUGPLAY_REGKEY_DEVICE,
                                                 KEY_WRITE,
                                                 &hKey
                                                );
                if ( ParportPDO )
                        ObDereferenceObject( ParportPDO );

                if ( !NT_SUCCESS( Status ) ) {
                        dbg0_printk( "PLUGPLAY_REGKEY_DEVICE openKey FAILED 0x%X\n", Status );
                        }
                else {
                        RtlInitUnicodeString( &valueName,
                                              (PWSTR)L"EnableConnectInterruptIoctl" );

                        YesConnect = 1;
                        Status = ZwSetValueKey( hKey,
                                                &valueName,
                                                0,
                                                REG_DWORD,
                                                &YesConnect,
                                                sizeof(ULONG)
                                              );
                        if ( !NT_SUCCESS( Status ) ) {
                                dbg0_printk( "Can't set PDO device value 0x%X\n", Status );
                                }

                        ZwClose( hKey );
                        }



                // Setup the ParPort ISR to call our interrupt service routine

                ParallelIsr.InterruptServiceRoutine = Wdm_DrvInterrupt; // PKSERVICE_ROUTINE
                ParallelIsr.InterruptServiceContext = (VOID*)Dev;       // PVOID
                ParallelIsr.DeferredPortCheckRoutine = NULL;
                ParallelIsr.DeferredPortCheckContext = NULL;

                Status = Wdm_IrpBuildSendAndWait(
                            IOCTL_INTERNAL,
                            IOCTL_INTERNAL_PARALLEL_CONNECT_INTERRUPT,
                            PPDevObj,
                            &ParallelIsr,
                            sizeof( PARALLEL_INTERRUPT_SERVICE_ROUTINE ),
                            &ParallelIsrInfo,
                            sizeof( PARALLEL_INTERRUPT_INFORMATION ),
                            NULL
                            );
                if ( !NT_SUCCESS( Status ) ) {
                        dbg0_printk("Failed to connect parallel ISR 0x%lX\n", Status);
                        Wdm_ErrorLog( Dev->ThisFdoDevObj, NULL, Status, PARXCH_LOG_INFO,
                                      L"Can't connect to parallel port interrupt" );
                        Wdm_ParPortRelease( Dev, PPDevObj, PPFileObj, PPIrqObj );
                        return( Status );
                        }
                else {
                        PPIrqObj = ParallelIsrInfo.InterruptObject;
                        }

                } // end if DeviceIrq != 0




        // Port acquired OK, so update these ParPort parameters in the DevInfo structure

        Dev->ParPortAcquired = 1;
        Dev->ParPortDevObj   = PPDevObj;
        Dev->ParPortFileObj  = PPFileObj;
        Dev->IrqObj          = PPIrqObj;


        dbg0_printk( "Wdm_ParPortAcquire Exit Context 0x%X\n", Context );

        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_ParPortRelease( DEVINFO *Dev,
                                     DEVICE_OBJECT *PPDevObj,
                                     FILE_OBJECT *PPFileObj,
                                     KINTERRUPT *PPIrqObj ) {

        NTSTATUS                            Status;
        PARALLEL_INTERRUPT_SERVICE_ROUTINE  ParallelIsr;

        dbg0_printk("Wdm_ParPortRelease Enter\n");


        // Disconnect the ISR

        if ( PPIrqObj && PPDevObj ) {

                ParallelIsr.InterruptServiceRoutine = Wdm_DrvInterrupt; // PKSERVICE_ROUTINE
                ParallelIsr.InterruptServiceContext = (VOID*)Dev;       // PVOID
                ParallelIsr.DeferredPortCheckRoutine = NULL;
                ParallelIsr.DeferredPortCheckContext = NULL;

                Status = Wdm_IrpBuildSendAndWait(
                            IOCTL_INTERNAL,
                            IOCTL_INTERNAL_PARALLEL_DISCONNECT_INTERRUPT,
                            PPDevObj,
                            &ParallelIsr,
                            sizeof( PARALLEL_INTERRUPT_SERVICE_ROUTINE ),
                            NULL,
                            0,
                            NULL
                            );

                if ( !NT_SUCCESS( Status ) ) {
                        dbg0_printk("Failed to disconnect parallel ISR 0x%lX\n", Status);
                        }
                }




        // Free the parport device from our EXCLUSIVE access

        if ( PPDevObj ) {
                Status = Wdm_IrpBuildSendAndWait(
                            IOCTL_INTERNAL,
                            IOCTL_INTERNAL_PARALLEL_PORT_FREE,
                            PPDevObj,
                            NULL,
                            0,
                            NULL,
                            0,
                            NULL
                            );

                if ( !NT_SUCCESS( Status ) ) {
                        dbg0_printk("Failed to exclusive free Parallel port 0x%lX\n", Status);
                        }
                }



        // Release our connection to the Parport device's file object

        if ( PPFileObj )
                ObDereferenceObject( PPFileObj );


        // Port release OK, so update these ParPort parameters in the DevInfo structure

        Dev->ParPortAcquired = 0;
        Dev->ParPortDevObj   = NULL;
        Dev->ParPortFileObj  = NULL;
        Dev->IrqObj          = NULL;

        dbg0_printk("Wdm_ParPortRelease Exit\n");

        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_ParPortGetPdo( DEVINFO *Dev,
                                    DEVICE_OBJECT *TargetDeviceObject,
                                    FILE_OBJECT *TargetFileObject,
                                    DEVICE_OBJECT **ParportPDO ) {

        //    This routine builds and sends a IRP_MN_QUERY_DEVICE_RELATIONS
        //    to the TargetDeviceObject, waits for the IRP to complete, and
        //    returns status to the caller.

        KEVENT             Event;
        NTSTATUS           Status;
        PIRP               Irp;
        LARGE_INTEGER      Timeout;
        PIO_STACK_LOCATION IrpSp;
        PDEVICE_RELATIONS  pDevRel;

//        PAGED_CODE();

        dbg0_printk("In Wdm_ParPortGetPdo\n");

        *ParportPDO = NULL;

        if ( TargetDeviceObject == NULL || TargetFileObject == NULL )
                return( STATUS_PORT_CONNECTION_REFUSED );


        // Allocate and initialize IRP

        Irp = IoAllocateIrp( (CCHAR)(TargetDeviceObject->StackSize + 1), FALSE );
        if ( !Irp )
                return( STATUS_INSUFFICIENT_RESOURCES );

        Irp->IoStatus.Status = STATUS_NOT_SUPPORTED;

        IrpSp = IoGetNextIrpStackLocation( Irp );

        IrpSp->MajorFunction = IRP_MJ_PNP;
        IrpSp->MinorFunction = IRP_MN_QUERY_DEVICE_RELATIONS;
        IrpSp->FileObject    = TargetFileObject;
        IrpSp->Parameters.QueryDeviceRelations.Type = TargetDeviceRelation;




        // Set completion routine and send IRP

        KeInitializeEvent( &Event, NotificationEvent, FALSE );

        IoSetCompletionRoutine( Irp,
                                (PIO_COMPLETION_ROUTINE)Wdm_OnRequestComplete,
                                (PVOID) &Event,
                                TRUE,
                                TRUE,
                                TRUE
                              );


        Status = IoCallDriver( TargetDeviceObject, Irp );
        if ( !NT_SUCCESS( Status ) ) {
                IoFreeIrp( Irp );
                return( Status );
                }



        // Set timeout in 100ns units and wait

        Timeout.QuadPart = -((LONGLONG) 10*1000*500); // default 500 ms

        Status = KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, &Timeout );


        // If we timed out, cancel the IRP

        if ( Status == STATUS_TIMEOUT ) {
                dbg2_printk("Timeout on ParPortGetPdo\n");
                IoCancelIrp( Irp );
                KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, NULL );
                }


        // Else the IRP is complete (or canceled).

        Status = Irp->IoStatus.Status;

        if ( !NT_SUCCESS( Status ) ) {
                dbg2_printk("ParPortGetPdo IRP completed but failed\n");
                IoFreeIrp( Irp );
                return( Status );
                }


        // Extract the PDO.

        pDevRel = (PDEVICE_RELATIONS)Irp->IoStatus.Information;

        if (pDevRel != NULL) {
                dbg2_printk( "Dev Rel Count is %d\n", pDevRel->Count );
                if (pDevRel->Count >= 1) {
                        dbg2_printk( "Count 0 obj is 0x%lX\n", pDevRel->Objects[0] );
                        *ParportPDO = pDevRel->Objects[0];
                        }
                ExFreePool( pDevRel );
                }

        IoFreeIrp( Irp );
        Status = STATUS_SUCCESS;

        dbg0_printk("Exit Wdm_ParPortGetPdo\n");
        return( Status );
}


SRLOCAL NTSTATUS Wdm_StartDevice( DEVICE_OBJECT *FDO ) {

        NTSTATUS     Status;
        DEVINFO     *Dev;
        POWER_STATE  PowerState;
        INITXCHPARMS Parm;

        dbg0_printk("Wdm_StartDevice Enter\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;



        // Initialize the PARXCH


//WCT: Why are we doing a full init here, don't we just want to
//     get the parport function pointers set

        Parm.XchModel =        PARXCH_MODEL_PAR4CH;
        Parm.ParPortIrq =      PARXCH_IRQ_7;
        Parm.ParPortMode =     PARXCH_PORT_MODE_BPP;
        Parm.Df =              PARXCH_DF_SIGNED;
        Parm.GainLog =         PARXCH_GAIN_1;
        Parm.TurboLog =        PARXCH_TURBO_16;
        Parm.Decimation =      PARXCH_DECIMATION_100Hz_T16;
        Parm.ExtraDecimation = PARXCH_EXTRADECIMATION_MIN;
        Parm.ADS1210_CrValue = 0L;
        Parm.Error =           PARXCH_ERROR_NONE;


        ParXch_Init( Dev, (int*)&Parm );



        // Turn on device and inform the power manager.

        if ( ParXch_PowerUp( Dev ) ) {
                Dev->CurrDevState = RUNSTATE_RUNNING;
                PowerState.DeviceState = PowerDeviceD0;
                }
        else {
                Dev->CurrDevState = RUNSTATE_STOPPED;
                PowerState.DeviceState = PowerDeviceD3;
                }

        PoSetPowerState( FDO, DevicePowerState, PowerState );



        // Set start indicators

        Status = Wdm_SetRegPortFound( Dev );
        if ( !NT_SUCCESS( Status ) ) {
                dbg_printk( "Can't write port found value to registry\n" );
                Wdm_ErrorLog( Dev->ThisFdoDevObj, NULL, Status, PARXCH_LOG_INFO,
                              L"Can't write port found value to registry" );
                }

        Dev->StartDone = 1;


//        IoSetDeviceInterfaceState( &Dev->InterfaceName, TRUE ); //enable

        dbg0_printk("Wdm_StartDevice Exit\n");

        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_SetRegPortFound( DEVINFO *Dev ) {

        NTSTATUS           Status;
        HANDLE             hKey;
        OBJECT_ATTRIBUTES  ObjAttr;
        UNICODE_STRING     ObjName;
        UNICODE_STRING     ValueName;
        ULONG              PortFound;

        dbg0_printk( "In Wdm_SetRegPortFound\n" );

        RtlZeroMemory( Dev->MsgBuffer, MAX_STR_LENGTH );
        RtlMoveMemory( Dev->MsgBuffer,
                       DriverRegPath.Buffer,
                       DriverRegPath.Length );
        wcscat( Dev->MsgBuffer, L"\\Parameters" );
        RtlInitUnicodeString( &ObjName, Dev->MsgBuffer );
        dbg0_printk( "Made parm key name of %ls\n", Dev->MsgBuffer );

        InitializeObjectAttributes(
                                   &ObjAttr,
                                   &ObjName,
                                   OBJ_KERNEL_HANDLE, // Attributes
                                   NULL,              // RootDirectory
                                   NULL               // SecurityDescriptor
                                  );

        Status = ZwOpenKey( &hKey, KEY_WRITE, &ObjAttr );
        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk( "Can't open driver parameters key Status = 0x%X\n", Status );
                return( Status );
                }



        RtlInitUnicodeString( &ValueName, (PWSTR)LREG_PARM_PORTFOUND );
        PortFound = 1;
        Status = ZwSetValueKey( hKey,
                                &ValueName,
                                0,
                                REG_DWORD,
                                &PortFound,
                                sizeof(ULONG)
                              );
        if ( !NT_SUCCESS( Status ) ) {
                dbg0_printk( "Can't set PortFound value Status = 0x%X\n", Status );
                }

        ZwClose( hKey );

        return( Status );
}

NTSTATUS Wdm_PnpStopDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;

        dbg0_printk("In Wdm_PnpStopDevice\n");

        // Handle queueing and cancellation.


        // Do our stopping work.

        Wdm_StopDevice( FDO );
        Wdm_ResourceClear( FDO );



        // Release exclusive access of the parallel port and disconnect
        // from its interrupt

        Dev = (DEVINFO *)FDO->DeviceExtension;
        Status = Wdm_ParPortRelease( Dev,
                                     Dev->ParPortDevObj,
                                     Dev->ParPortFileObj,
                                     Dev->IrqObj
                                   );




        if (Dev->NotificationHandle) {
                IoUnregisterPlugPlayNotification( Dev->NotificationHandle );
                Dev->NotificationHandle = NULL;
                }


        // Let lower drivers do their stopping work.

        Irp->IoStatus.Status = STATUS_SUCCESS;
        return( Wdm_IrpForward( FDO, Irp ) );
}

SRLOCAL NTSTATUS Wdm_StopDevice( DEVICE_OBJECT *FDO ) {

        DEVINFO *Dev;
        POWER_STATE PowerState;

        dbg0_printk("Wdm_StopDevice Enter\n");


        Dev = (DEVINFO *)FDO->DeviceExtension;


        // This function undoes everything done in StartDevice

        Dev->StartDone = 0;



        // Turn off device and inform the power manager.

        if ( ParXch_PowerDown( Dev ) ) {
                Dev->CurrDevState = RUNSTATE_STOPPED;
                PowerState.DeviceState = PowerDeviceD3;
                }
        else {
                Dev->CurrDevState = RUNSTATE_RUNNING;
                PowerState.DeviceState = PowerDeviceD0;
                }

        PoSetPowerState( FDO, DevicePowerState, PowerState );


//        IoSetDeviceInterfaceState( &Dev->InterfaceName, FALSE ); //disable

        dbg0_printk("Wdm_StopDevice Exit\n");
        return( STATUS_SUCCESS );
}


SRLOCAL NTSTATUS Wdm_ResourceClear( DEVICE_OBJECT *FDO ) {

        DEVINFO *Dev;

        Dev = (DEVINFO *)FDO->DeviceExtension;

        // Unmap port if needed

        if ( Dev->ParPortBase && Dev->ParPortMapped) {
                MmUnmapIoSpace( (UCHAR*)Dev->ParPortBase,
                                Dev->ParPortCount );
                Dev->ParPortMapped = 0;
                }

        return( STATUS_SUCCESS );
}

NTSTATUS Wdm_PnpRemoveDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;

        dbg0_printk("In Wdm_PnpRemoveDevice\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;


        // Deregister all interfaces.



        // We must first do all the stopping things

        Status = Wdm_PnpStopDevice( FDO, Irp );



        // RemoveDevice undoes things done in AddDevice

        Wdm_RemoveDevice( FDO );

        return( Status );
}
NTSTATUS Wdm_PnpQueryCapabilities( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryCapabilities\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpCancelStop( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpCancelStop\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpCancelRemove( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpCancelRemove\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryStop( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryStop\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryRemove( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryRemove\n");
        Irp->IoStatus.Status = STATUS_SUCCESS;
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryDevRelations( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryDevRelations\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryInterface( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryInterface\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpFilterResourceReq( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpFilterResourceReq\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpSurpriseRemoval( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpSurpriseRemoval\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}




// Power Functions

NTSTATUS Wdm_DrvPower( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS           Status;
        DEVINFO           *Dev;
        IO_STACK_LOCATION *IrpStack;
        ULONG              MinorFn;
        POWER_STATE_TYPE   PowerType;

        dbg0_printk("Wdm_DrvPower Enter\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        MinorFn    = IrpStack->MinorFunction;
        PowerType  = IrpStack->Parameters.Power.Type;


//        Status = IoAcquireRemoveLock();
//        if ( !NT_SUCCESS( Status ) ) {
//            Wdm_PowerFail( Irp, Status );
//            return( Status );
//            }



        switch( MinorFn ) {

            case IRP_MN_SET_POWER:

                if ( PowerType == SystemPowerState )
                        Status = Wdm_PowerSetSystem( FDO, Irp );

                else if ( PowerType == DevicePowerState )
                        Status = Wdm_PowerSetDevice( FDO, Irp );

                else
                        Status = Wdm_PowerForward( FDO, Irp );

                break;

            case IRP_MN_QUERY_POWER:
                dbg0_printk("Power IRP_MN_QUERY_POWER always forward\n");
                Status = Wdm_PowerForward( FDO, Irp );
                break;

            case IRP_MN_WAIT_WAKE:
            case IRP_MN_POWER_SEQUENCE:
            default:
                if (MinorFn == IRP_MN_WAIT_WAKE)
                        dbg0_printk("Power IRP_MN_WAIT_WAKE\n");
                else if (MinorFn == IRP_MN_POWER_SEQUENCE)
                        dbg0_printk("Power IRP_MN_POWER_SEQUENCE\n");
                else
                        dbg0_printk("Power minor fn 0x%X\n", MinorFn);
                Status = Wdm_PowerForward( FDO, Irp );
            }

//        IoReleaseRemoveLock();


        dbg0_printk("Wdm_DrvPower Exit\n");
        return( Status );
}

NTSTATUS Wdm_PowerSetSystem( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;

        dbg0_printk("Wdm_PowerSetSystem Enter\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;


        // We pass ALL system power requests down ...

        // Mark the incoming system set power (S-IRP) as pending.  The
        // function driver continues to process the S-IRP when the
        // system calls the Wdm_OnCompletionPowerSystem I/O completion
        // routine after the underlying bus driver completes the S-IRP.

        IoMarkIrpPending( Irp );
        IoCopyCurrentIrpStackLocationToNext( Irp );
        IoSetCompletionRoutine( Irp,
                                Wdm_OnCompletePowerSystem,
                                Dev,
                                TRUE,
                                TRUE,
                                TRUE
                              );
        PoCallDriver( Dev->LowerDevObj, Irp );

        Status = STATUS_PENDING;  // processing continues in Wdm_OnCompletePowerSystem

        dbg0_printk("Wdm_PowerSetSystem Exit\n");
        return( Status );
}

NTSTATUS Wdm_OnCompletePowerSystem( DEVICE_OBJECT *FDO, IRP *Irp, void *Context ) {

        NTSTATUS           Status;
        DEVINFO           *Dev;
        IO_STACK_LOCATION *IrpStack;
        POWER_STATE        DevPowerState, SysPowerState;

        dbg0_printk("Wdm_OnCompletePowerSystem Enter\n");


        // System Irp has been passed down the driver stack and we are now
        // on the way back up

        // If a lower driver failed, start the next IRP.
        // The IO manager will complete this one on the success return.

        if ( !NT_SUCCESS( Irp->IoStatus.Status ) ) {
                PoStartNextPowerIrp( Irp );
                return( STATUS_SUCCESS );  // was Status, but must not fail power irp
                }


        // Get device and irp info

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        SysPowerState = IrpStack->Parameters.Power.State;


        // Determine desired device state from requested system state
        // Device ON for full system ON, otherwise device OFF

        if ( SysPowerState.SystemState == PowerSystemWorking )
                DevPowerState.DeviceState = PowerDeviceD0;
        else
                DevPowerState.DeviceState = PowerDeviceD3;


        // Save this system IRP so it can be finished later
        // and request a device IRP to set the device power.

        Dev->SystemIrp = Irp;

        Status = PoRequestPowerIrp( Dev->PdoDevObj,
                                    IrpStack->MinorFunction,
                                    DevPowerState,
                                    Wdm_OnCompletePowerFinish,
                                    Dev,
                                    NULL
                                  );


        // Inform the IO subsystem that we are not done with this IRP.
        // It will be finished by Wdm_OnCompletePowerFinish

        Status = STATUS_MORE_PROCESSING_REQUIRED;

        dbg0_printk("Wdm_OnCompletePowerSystem Exit\n");
        return( Status );
}

void Wdm_OnCompletePowerFinish( DEVICE_OBJECT *FDO,
                                UCHAR MinorFunction,
                                POWER_STATE PowerState,
                                void *Context,
                                IO_STATUS_BLOCK *IoStatus ) {

        DEVINFO           *Dev;
        IRP               *SystemStateIrp;
        IO_STACK_LOCATION *IrpStack;
        POWER_STATE        SysPowerState;

        dbg0_printk("Wdm_OnCompletePowerFinish Enter\n");

//      This callback executes when all devices in the stack have
//      completed the DevicePowerState _SET_POWER Irp that we requested.
//      Its purpose is to start the next SystemPower IRP and complete the
//      current SystemPower IRP


        Dev = (DEVINFO *)Context;

        SystemStateIrp = Dev->SystemIrp;
        Dev->SystemIrp = NULL;          // clean up as we don't need this any more

        IrpStack = IoGetCurrentIrpStackLocation( SystemStateIrp );
        SysPowerState = IrpStack->Parameters.Power.State;

        // Record the new system power state saved in the device info area if all was successful

        if ( NT_SUCCESS(IoStatus->Status) ) {
                dbg3_printk("Changing system power from %d to %d (passed state is %d)\n", Dev->CurrSysPower, SysPowerState.SystemState, PowerState.SystemState );
                Dev->CurrSysPower = SysPowerState.SystemState;
                }


        SystemStateIrp->IoStatus.Status = IoStatus->Status;
        PoStartNextPowerIrp( SystemStateIrp );
        IoCompleteRequest( SystemStateIrp, IO_NO_INCREMENT );

        dbg0_printk("Wdm_OnCompletePowerFinish Exit\n");
        return;
}

NTSTATUS Wdm_PowerSetDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS           Status;
        DEVINFO           *Dev;
        IO_STACK_LOCATION *IrpStack;
        POWER_STATE        PowerState;

        dbg0_printk("Wdm_PowerSetDevice Enter\n");


        // Note: HIGHER state numbers mean LOWER power

        // Get device and irp info

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        PowerState = IrpStack->Parameters.Power.State;


        // We always set success for SET_POWER

        Irp->IoStatus.Status = STATUS_SUCCESS;


        // Powering down (current state < new state )
        // Do hardware power down work, update and record state, send IRP down the stack

        if ( Dev->CurrDevPower < PowerState.DeviceState ) {

                dbg3_printk("Powering down device from %d to %d\n",Dev->CurrDevPower, PowerState.DeviceState);

                // WCT: Actually do hardware power down here

                PoSetPowerState( Dev->ThisFdoDevObj, DevicePowerState, PowerState );
                Dev->CurrDevPower = PowerState.DeviceState;

                Status = Wdm_PowerForward( FDO, Irp );
                }


        // Powering up, we'll do our work in the completion routine on the
        // way back up the driver stack

        else if ( Dev->CurrDevPower > PowerState.DeviceState ) {

                dbg3_printk("Powering up device from %d to %d\n",Dev->CurrDevPower, PowerState.DeviceState);

                IoCopyCurrentIrpStackLocationToNext( Irp );
                IoSetCompletionRoutine( Irp,
                                        Wdm_OnCompletePowerDeviceUp,
                                        Dev,
                                        TRUE,
                                        TRUE,
                                        TRUE
                                      );
                Status = PoCallDriver( Dev->LowerDevObj, Irp );
                }


        // No change, just pass the IRP on down the driver stack

        else
                Status = Wdm_PowerForward( FDO, Irp );


        dbg0_printk("Wdm_PowerSetDevice Exit\n");
        return( Status );
}

NTSTATUS Wdm_OnCompletePowerDeviceUp( DEVICE_OBJECT *FDO, IRP *Irp, void *Context ) {

        NTSTATUS           Status;
        DEVINFO           *Dev;
        IO_STACK_LOCATION *IrpStack;
        POWER_STATE        PowerState;

        dbg0_printk("Wdm_OnCompletePowerUpDevice Enter\n");

        // Check if a lower driver failed

        if ( !NT_SUCCESS( Irp->IoStatus.Status ) ) {
                PoStartNextPowerIrp( Irp );
                return( STATUS_SUCCESS );  // was Status, but can't fail power IRPs
                }



        // Remember if a lower driver pended the Irp
        // (as we did not mark it pending in the dispatch routine)

        if ( Irp->PendingReturned )
                IoMarkIrpPending( Irp );


        // Get device and irp info

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        PowerState = IrpStack->Parameters.Power.State;


        // Finish powering up

        // WCT: Actually do hardware power on here

        PoSetPowerState( Dev->ThisFdoDevObj, DevicePowerState, PowerState );
        Dev->CurrDevPower = PowerState.DeviceState;


        // We're done processing this power Irp, start the next one

        Status = STATUS_SUCCESS;
        PoStartNextPowerIrp( Irp );


        dbg0_printk("Wdm_OnCompletePowerUpDevice Exit\n");
        return( Status );
}

NTSTATUS Wdm_PowerForward( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;

        dbg0_printk("Wdm_PowerForward Enter\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;

        PoStartNextPowerIrp( Irp );
        IoSkipCurrentIrpStackLocation( Irp );
        Status = PoCallDriver( Dev->LowerDevObj, Irp );

        dbg0_printk("Wdm_PowerForward Exit\n");
        return( Status );
}

NTSTATUS Wdm_PowerFail( IRP *Irp, NTSTATUS Status ) {

        dbg0_printk("Wdm_PowerFail Enter\n");

        PoStartNextPowerIrp( Irp );
        OsIrpFinish( Irp, Status, 0 );

        dbg0_printk("Wdm_PowerFail Exit\n");
        return( Status );
}




// Windows Management Instrumentation (WMI) Functions

// WCT: Add support for WMI here

NTSTATUS Wdm_DrvWmi( DEVICE_OBJECT *FDO, IRP *Irp ) {

        IO_STACK_LOCATION *IrpStack;
        ULONG MinorFn;

        dbg1_printk("In Wdm_DrvWmi\n");

        // Get info about the current minor function
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        MinorFn  = IrpStack->MinorFunction;
        dbg1_printk("WMI minor fn 0x%X\n", MinorFn);

        // Let lower drivers do their WMI work.
        return( Wdm_IrpForward( FDO, Irp ) );
}

// **************************** WMI *********************************




SRLOCAL NTSTATUS Wdm_OnRequestComplete( DEVICE_OBJECT *FDO,
                                        IRP *Irp,
                                        void *Context ) {

        dbg0_printk("In Wdm_OnRequestComplete\n");

        KeSetEvent( (KEVENT *)Context, 0, FALSE );
        return( STATUS_MORE_PROCESSING_REQUIRED );
}

SRLOCAL NTSTATUS Wdm_IrpBuildSendAndWait(
                        ULONG           Internal,
                        ULONG           IoControlCode,
                        PDEVICE_OBJECT  TargetDeviceObject,
                        PVOID           InputBuffer,
                        ULONG           InputBufferLength,
                        PVOID           OutputBuffer,
                        ULONG           OutputBufferLength,
                        PLARGE_INTEGER  RequestedTimeout
                        ) {

/*
    This routine builds and sends an Internal IOCTL to the TargetDeviceObject, waits
    for the IOCTL to complete, and returns status to the caller.

    *** This function does not support Input and Output in the same IOCTL

    Internal            - flag indicating if the IOCTL is internal or not
    IoControlCode       - the IOCTL to send
    TargetDeviceObject  - who to send the IOCTL to
    InputBuffer         - pointer to input buffer, if any
    InputBufferLength,  - length of input buffer
    OutputBuffer        - pointer to output buffer, if any
    OutputBufferLength, - length of output buffer
    Timeout             - how long to wait for request to complete,
                          NULL==use driver global AcquirePortTimeout

    All parameters are inputs except for OutputBuffer.
    InputBuffer, OutputBuffer, and RequestedTimeout are OPTIONAL.

    Returns    Status
*/

        KEVENT             Event;
        NTSTATUS           Status;
        PIRP               Irp;
        LARGE_INTEGER      Timeout;
        PIO_STACK_LOCATION IrpSp;
        BOOLEAN            NeedToCopyOutputBuffer = FALSE;
        ULONG              i;
        char              *pBuffIn, *pBuffOut;


//        PAGED_CODE();

        dbg0_printk("In Wdm_IrpBuildSendAndWait IOCTL 0x%X\n", IoControlCode);



        // Allocate and initialize IRP

        Irp = IoAllocateIrp( (CCHAR)(TargetDeviceObject->StackSize + 1), FALSE );
        if ( !Irp )
                return( STATUS_INSUFFICIENT_RESOURCES );

        IrpSp = IoGetNextIrpStackLocation( Irp );

        if (Internal)
                IrpSp->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        else
                IrpSp->MajorFunction = IRP_MJ_DEVICE_CONTROL;

        IrpSp->Parameters.DeviceIoControl.OutputBufferLength = OutputBufferLength;
        IrpSp->Parameters.DeviceIoControl.InputBufferLength  = InputBufferLength;
        IrpSp->Parameters.DeviceIoControl.IoControlCode      = IoControlCode;

// WCT: Oney recommends setting the FileObject too
//        if (TargetFileObject)
//                IrpSp->FileObject = TargetFileObject;


        // Set the system buffer to be the larger of the two buffers
        // If not using the input buffer, then copy the input data.

        if ( InputBufferLength >= OutputBufferLength ) {
                Irp->AssociatedIrp.SystemBuffer = InputBuffer;
                }
        else {
                Irp->AssociatedIrp.SystemBuffer = OutputBuffer;
                pBuffIn  = (char*)InputBuffer;
                pBuffOut = (char*)OutputBuffer;
                for ( i = 0 ; i < InputBufferLength ; i++ )
                        *pBuffOut++ = *pBuffIn++;
                }


        // Set completion routine and send IRP

        KeInitializeEvent( &Event, NotificationEvent, FALSE );

        IoSetCompletionRoutine( Irp,
                                (PIO_COMPLETION_ROUTINE)Wdm_OnRequestComplete,
                                (PVOID) &Event,
                                TRUE,
                                TRUE,
                                TRUE
                              );


        Status = IoCallDriver( TargetDeviceObject, Irp );
        if ( !NT_SUCCESS( Status ) ) {
                IoFreeIrp( Irp );
                return( Status );
                }



        // Set timeout in 100ns units and wait

        if (RequestedTimeout != NULL)
                Timeout = *RequestedTimeout;
        else
                Timeout.QuadPart = -((LONGLONG) 10*1000*500); // default 500 ms

        Status = KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, &Timeout );


        // If we timed out, cancel the IRP

        if ( Status == STATUS_TIMEOUT ) {
                IoCancelIrp( Irp );
                KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, NULL );
                }


        // Else the IRP is complete (or canceled).

        Status = Irp->IoStatus.Status;
        IoFreeIrp( Irp );


        // If using the input buffer as the system buffer,
        // then copy the output data.

        if ( InputBufferLength >= OutputBufferLength ) {

                pBuffIn  = (char*)InputBuffer;
                pBuffOut = (char*)OutputBuffer;
                for ( i = 0 ; i < OutputBufferLength ; i++ )
                        *pBuffOut++ = *pBuffIn++ ;
                }


        return( Status );
}

SRLOCAL NTSTATUS Wdm_IrpForwardAndWait( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;
        KEVENT   Event;

        dbg0_printk("In Wdm_IrpForwardAndWait\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;

        // Setup event, Irp stack, and completion routine

        KeInitializeEvent( &Event, NotificationEvent, FALSE );

        IoCopyCurrentIrpStackLocationToNext( Irp );
        IoSetCompletionRoutine( Irp,
                                (PIO_COMPLETION_ROUTINE)Wdm_OnRequestComplete,
                                (PVOID) &Event,
                                TRUE,
                                TRUE,
                                TRUE
                              );


        // Pass Irp down and wait until lower object is done

        IoCallDriver( Dev->LowerDevObj, Irp );
        KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, NULL );

        return( Irp->IoStatus.Status );
}

SRLOCAL NTSTATUS Wdm_IrpForward( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;

        dbg0_printk("In Wdm_IrpForward\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IoSkipCurrentIrpStackLocation( Irp );
        return( IoCallDriver( Dev->LowerDevObj, Irp ) );
}


SRLOCAL void Wdm_ResourceFree( DEVICE_OBJECT *DeviceObject,
                               UNICODE_STRING *LinkDeviceName
                                ) {

        DEVINFO *Dev;

        dbg0_printk("In Wdm_ResourceFree\n");

        if (LinkDeviceName != NULL)
                IoDeleteSymbolicLink( LinkDeviceName );

        if (DeviceObject) {

                Dev = (DEVINFO *)DeviceObject->DeviceExtension;

                if (Dev->LowerDevObj)
                        IoDetachDevice(Dev->LowerDevObj);

                // Careful, in Win98 IoDeleteDevice calls DriverUnload !!
                IoDeleteDevice( DeviceObject );  // This is the FDO
                }

}




// *************************************************************************

#elif defined( SROS_WIN9X ) || defined( SROS_MSDOS )

// Global variables for Win95/98 and MsDos "driver" code

SRLOCAL DEVINFO *Dev;
SRLOCAL int      isOpen = 0;
SRLOCAL long     Status = STATUS_SUCCESS;


// WCT: In the future, indriver.h may be needed for correctly choosing the
// address and irq based on the specified driver name.  In the meantime,
// the current method works ok.
//
//#include "indriver.h"

SRLOCAL int Msdos_NameToAddress( char *DriverName ) {

        int ParPortAddress;

        if ( strcmp( DriverName, PARXCH_0 ) == 0 )
                ParPortAddress = 0x378;

        else if ( strcmp( DriverName, PARXCH_1 ) == 0 )
                ParPortAddress = 0x278;

        else if ( strcmp( DriverName, PARXCH_0x378 ) == 0 )
                ParPortAddress = 0x378;

        else if ( strcmp( DriverName, PARXCH_0x278 ) == 0 )
                ParPortAddress = 0x278;

        else {
                if (sscanf( DriverName, "SrParXch%3x", &ParPortAddress ) != 1)
                        ParPortAddress = 0x378;
                }

        return( ParPortAddress );
}


SRLOCAL int Msdos_NameToIrq( char *DriverName ) {

        int ParPortIrq;

        if ( strcmp( DriverName, PARXCH_0 ) == 0 )
                ParPortIrq = PARXCH_IRQ_0x378;

        else if ( strcmp( DriverName, PARXCH_1 ) == 0 )
                ParPortIrq = PARXCH_IRQ_0x278;

        else if ( strcmp( DriverName, PARXCH_0x378 ) == 0 )
                ParPortIrq = PARXCH_IRQ_0x378;

        else if ( strcmp( DriverName, PARXCH_0x278 ) == 0 )
                ParPortIrq = PARXCH_IRQ_0x278;

        else
                ParPortIrq = PARXCH_IRQ_0x378;

        return( ParPortIrq );
}


DEVHANDLE Msdos_DrvOpen( char *DriverName ) {

        int ParPortAddress, ParPortIrq;

        dbg0_printk("In Msdos_DrvOpen\n");


        // This part performs driver initialization ....

        // Prevent multiple opens

        if (isOpen)
                return( BAD_DEVHANDLE );


        // Set up dispatch table
        // None needed since user library can call driver functions directly.



        // Get parameters from argument list
        // Convert from DriverName to ParPortAddress

        ParPortAddress = Msdos_NameToAddress( DriverName );
        ParPortIrq     = Msdos_NameToIrq( DriverName );


        // Fill device information

        Dev = &ParXchDeviceInfo;
        ParXch_FillDevInfo( Dev, DriverName, ParPortAddress, 0, ParPortIrq );




        // This part opens the driver ....

        // Open driver

        Status = ParXch_DrvOpen( Dev, NULL );

        if (Status == STATUS_SUCCESS) {
                isOpen = 1;
                return( (DEVHANDLE)Dev );
                }
        else {
                Msdos_DrvClose( (DEVHANDLE)Dev );
                return( BAD_DEVHANDLE );
                }
}

int Msdos_DrvRead( DEVHANDLE hDevice,
                   void *lpBuffer,
                   unsigned long nNumberOfBytesToRead,
                   unsigned long *lpNumberOfBytesRead ) {

        IRP Irp;

        dbg4_printk("In Msdos_DrvRead\n");

        // Check for wrong device
        if (hDevice != (DEVHANDLE)Dev)
                return( 0 );

        Irp.Command   = IOCTL_MAXFUNC;
        Irp.InBuffer  = NULL;
        Irp.InSize    = 0;
        Irp.OutBuffer = lpBuffer;
        Irp.OutSize   = nNumberOfBytesToRead;

        Status = ParXch_DrvRead( Dev, &Irp );

        *lpNumberOfBytesRead = Irp.ReturnedBytes;

        if (Status == STATUS_SUCCESS)
                return( 1 );
        else
                return( 0 );
}


int Msdos_DrvIoctl( DEVHANDLE hDevice,
                    unsigned long dwIoControlCode,
                    void *lpInBuffer,
                    unsigned long nInBufferSize,
                    void *lpOutBuffer,
                    unsigned long nOutBufferSize,
                    unsigned long *lpBytesReturned ) {

        IRP Irp;

        dbg4_printk("In Msdos_DrvIoctl\n");

        // Check for wrong device
        if (hDevice != (DEVHANDLE)Dev)
                return( 0 );


        Irp.Command    = dwIoControlCode;
        Irp.InBuffer   = lpInBuffer;
        Irp.InSize     = nInBufferSize;
        Irp.OutBuffer  = lpOutBuffer;
        Irp.OutSize    = nOutBufferSize;
        Irp.DataMethod = METHOD_BUFFERED;

        // Check in parxchkd.h for which control codes do not
        // use the more common buffered method of passing data

//        if ( (Irp.Command == IOCTL_PARXCH_XXX) ||
//             (Irp.Command == IOCTL_PARXCH_YYY) )
//                Irp.DataMethod = METHOD_OUT_DIRECT;
//        else
//                Irp.DataMethod = METHOD_BUFFERED;


        Status = ParXch_DrvIoctl( Dev, &Irp );

        *lpBytesReturned = Irp.ReturnedBytes;

        if (Status == STATUS_SUCCESS)
                return( 1 );
        else
                return( 0 );
}


int Msdos_DrvClose( DEVHANDLE hDevice ) {

        dbg0_printk("In Msdos_DrvClose\n");

        // This part closes the driver ...

        // Check for device already closed or wrong device
        if ((!isOpen) || (hDevice != (DEVHANDLE)Dev) )
                return( 0 );


        // Close it
        isOpen = 0;
        ParXch_DrvClose( Dev, NULL );



        // This part cleans up (unloads) the driver ...

        ParXch_DrvUnload( Dev );

        return( 1 );
}

long Msdos_DrvGetLastError( void ) {
        // Note: Error number locations vary by type:
        //       NT     in \ddk\inc\ntstatus.h
        //       Win32  in \MFC\include\winerror.h
        // where MFC stands for the visual C directory \msdevstd\include
        // or Program Files\Microsoft Visual Studio\VC98\Include


        // Mimic NT to Win32 error translation for specific errors.

        if (Status == STATUS_DEVICE_DATA_ERROR)
                return( ERROR_CRC );


        // Return the untranslated error for all the rest.

        else
                return( Status );
}






// *************************************************************************

#elif defined( SROS_LINUX )


MODULE_AUTHOR("Symmetric Research");
MODULE_DESCRIPTION("Symmetric Research PARxCH 24 bit A/D Linux Device Driver");
MODULE_LICENSE("GPL"); // Other Options: GPL v2, Dual BSD/GPL, Proprietary, etc


// Allow these variables to be overridden from the insmod command line.

SRLOCAL char  xch_name_buf[50];
SRLOCAL char  xch_class_buf[50];
SRLOCAL char  xch_suffix_buf[50];
SRLOCAL char *xch_name    = xch_name_buf;
SRLOCAL char *xch_class   = xch_class_buf;
SRLOCAL char *xch_suffix  = xch_suffix_buf;
SRLOCAL int   xch_address = -1;
SRLOCAL int   xch_irq     = -1;
SRLOCAL int   xch_major   = -1;
SRLOCAL int   xch_cdev    = -1;


#define MODPERM 0  // A future option is S_IRUGO

module_param_string( xch_name,   xch_name_buf,   sizeof(xch_name_buf),   MODPERM );
module_param_string( xch_class,  xch_class_buf,  sizeof(xch_class_buf),  MODPERM );
module_param_string( xch_suffix, xch_suffix_buf, sizeof(xch_suffix_buf), MODPERM );
module_param( xch_address, int, MODPERM );
module_param( xch_irq,     int, MODPERM );
module_param( xch_major,   int, MODPERM );




// Dispatch function table

SRLOCAL struct file_operations MajorFunction = {
        .owner   = THIS_MODULE,
        .open    = Linux_DrvOpen,
        .release = Linux_DrvClose,
        .read    = Linux_DrvRead,
        .ioctl   = Linux_DrvIoctl,
        };


// ParPort sysfs class device name

#define SYSFSDEV "SrParXch%X" // no longer needed in class_device_create call

static PARPORT_DRIVER DefaultParPortDriver = {
        .name = "SrParXch0", // This will be replaced if needed
        .attach = Linux_ParPortAttachPort,
        .detach = Linux_ParPortDetachPort,
        };


SRLOCAL void Linux_GetParameters( void ) {

        // Set global parameters from defaults if not already
        // set from command line values

        if (xch_major == -1)
                xch_major = DEFAULT_XCH_MAJOR;

        if (xch_address == -1)
                xch_address = DEFAULT_XCH_ADDRESS;

        if (xch_name == NULL)
                xch_name = DEFAULT_XCH_NAME;

        if (xch_irq == -1)
                xch_irq = DEFAULT_XCH_IRQ;

        if (xch_cdev == -1)
                xch_cdev = 0;

        if (xch_class == NULL)
                xch_class = DEFAULT_XCH_CLASS;

        if (xch_suffix == NULL)
                xch_suffix = DEFAULT_XCH_SUFFIX;
}

SRLOCAL void Linux_ResourceFree( DEVINFO *Dev,
                                 int DeviceCdev,
                                 int PortIrq,
                                 int PortAddress,
                                 int DevMajor,
                                 char *DriverName ) {

        dbg0_printk( "Linux_ResourceFree\n" );


        // Remove device from system

        if (DeviceCdev == 1) {
                cdev_del( &Dev->Cdev );
                dbg0_printk( "Deleting Cdev\n" );
                }


        // Clean up parport at the DRIVER, class and DEVICE levels

        Linux_ParPortUnregisterRelease( Dev );



        // Unregister the character device

        if (DriverName) {
                unregister_chrdev_region( Dev->DevNum, 1 );
                dbg0_printk( "Unregistering char device = %d devnum = 0x%lX\n", DevMajor, (long)Dev->DevNum );
                }
}

int Linux_ResourceReserve( DEVINFO *Dev ) {

        int Status;


        // Fill device information

        ParXch_FillDevInfo( Dev, xch_name, xch_address, 0, xch_irq );



        // Register character device

        Dev->DevNum = MKDEV( xch_major, 0 );
        dbg0_printk( "Register major %d, devnum 0x%lX, name %s\n", xch_major, (unsigned long)Dev->DevNum, xch_name );

        if (xch_major == 0) {             // Allow dynamic allocation

                Status = alloc_chrdev_region( &Dev->DevNum, 0, 1, xch_name );

                if (Status < 0) {
                        Linux_ResourceFree( Dev, xch_cdev, -1, xch_address, xch_major, NULL );
                        dbg0_printk( "PARXCH: Can't get dynamic major\n" );
                        return( Status );
                }
                xch_major = MAJOR( Dev->DevNum );
                dbg0_printk( "Register major %d, devnum 0x%lX, name %s\n", xch_major, (unsigned long)Dev->DevNum, xch_name );

        }

        else { // (xch_major != 0)        // Request fixed major number
                Status = register_chrdev_region( Dev->DevNum, 1, xch_name );

                if (Status < 0) {
                        Linux_ResourceFree( Dev, xch_cdev, -1, xch_address, xch_major, NULL );
                        dbg0_printk( "PARXCH: Can't get fixed major %d\n", xch_major );
                        return( Status );
                }
        }

        dbg0_printk( "Registered as char device %d\n", xch_major );




        // Register with the parport system

        Status = Linux_ParPortRegisterClaim( Dev );
        if (Status != STATUS_SUCCESS ) {
                Linux_ResourceFree( Dev, xch_cdev, -1, xch_address, xch_major, xch_name );
                return( Status );
                }

        // Initialize spin lock for protecting port

        spin_lock_init( &Dev->PortLock );
        spin_lock_init( &Dev->ADS1210InitLock );




        // Initialize character device and add it to the system

        cdev_init( &Dev->Cdev, &MajorFunction );
        Dev->Cdev.owner = THIS_MODULE;  // WCT26 - Is this required or done by default in init call above?
        Dev->Cdev.ops   = &MajorFunction;

        Status = cdev_add( &Dev->Cdev, Dev->DevNum, 1 ); // device now live
        if (Status) {
                Linux_ResourceFree( Dev, xch_cdev, xch_irq, xch_address, xch_major, xch_name );
                printk(KERN_NOTICE "parxch: Failed cdev add (err %d)\n", Status );
                return( STATUS_NO_TOKEN );
                }
        else {
                xch_cdev = 1;
                }

        return( STATUS_SUCCESS );
}

int Linux_ParPortPreempt( void *handle ) {

        return( 1 ); // We refuse to give it up
}

void Linux_ParPortIrq( void *dev_id ) {

        dbg5_printk("Linux_ParPortIrq - Enter\n");

        ParXch_DrvInterrupt( (DEVINFO *)dev_id );

        dbg5_printk("Linux_ParPortIrq - Exit\n");
}

int Linux_ParPortRegisterClaim( DEVINFO *Dev ) {

        dbg0_printk( "Linux_ParPortRegisterClaim - Enter\n" );

        // Register with the parport system.


        // Prepare entries in the newer sysfs systems at the upper or
        // dir/class level.

        sprintf( Dev->SysFsClassName, "%s%s",      xch_class, xch_suffix );

        dbg2_printk("Trying sysfs class name %s\n", Dev->SysFsClassName );

        Dev->ParPortClass = class_create( THIS_MODULE, Dev->SysFsClassName );

        if ( IS_ERR( Dev->ParPortClass ) ) {
                dbg_printk("Unable to prepare sysfs class entries for parport\n");
                return( STATUS_PORT_CONNECTION_REFUSED );
                }


        // Register our driver with the parport subsystem.  This says
        // tell me about all current ports and those that come and go.
        // This is done by the parport subsystem calling the
        // attach/detach routines we provide in parxch_parport_driver
        // which is defined above as a struct parport_driver *.

        DefaultParPortDriver.name = xch_name;
        Dev->ParPortDriver        = &DefaultParPortDriver;

        if ( parport_register_driver( Dev->ParPortDriver ) != STATUS_SUCCESS ) {
                dbg_printk("Unable to register driver with parport\n");
                Dev->ParPortDriver = NULL; // Indicate it is not available.
                return( STATUS_PORT_CONNECTION_REFUSED );
                }


        // Linux_ParPortAttachPort is called for all exisiting ports by
        // the parport_register_driver function above.  If our attach
        // function sees a port we like, we register our device which
        // sets parxch_parport_device ...



        // We insist on being attached to an existing parport and refuse
        // to wait and hope one shows up later.  We can tell this has
        // occurred by checking parxch_parport_device.

        if ( !Dev->ParPortDevice ) {
                dbg_printk("Did not get attached to the desired parport\n");
                return( STATUS_PORT_CONNECTION_REFUSED );
                }




        // If all went well and we are attached to a parport, indicate
        // we want to start using it now.  We don't want to deal with
        // interrupts until after we've opened and started the PARxCH
        // driver.  We tried to disable them in ParPortAttach after
        // registering the device, but do it again here just to be sure
        // they are really diabled.

        if ( parport_claim( Dev->ParPortDevice ) == STATUS_SUCCESS ) {
                ParXch_InterruptDisable( Dev );
                Dev->ParPortPort  = Dev->ParPortDevice->port;
                }
        else {
                dbg_printk("Failed to claim the desired parport at 0x%X\n", xch_address);
                return( STATUS_PORT_CONNECTION_REFUSED );
                }


        dbg0_printk( "Linux_ParPortRegisterClaim - Exit\n" );

        return( STATUS_SUCCESS );
}

void Linux_ParPortAttachPort( struct parport *port ) {

        DEVINFO *Dev;

        dbg0_printk( "Linux_ParPortAttachPort - Enter\n" );


        // Error check

        if (!port)
                return;


        dbg1_printk( "Port name    = %s\n",    port->name );
        dbg1_printk( "Port num     = %d\n",    port->portnum );
        dbg1_printk( "Port index   = %d\n",    port->number );
        dbg1_printk( "Port base    = 0x%lX\n", port->base );
        dbg1_printk( "Port base hi = 0x%lX\n", port->base_hi );
        dbg1_printk( "Port modes   = 0x%X\n",  port->modes );
        dbg1_printk( "Port irq     = %d\n",    port->irq );


        Dev = &ParXchDeviceInfo;


        // Attach to a parport.

        // This routine is called when a parport is "found".  There are
        // two times when this may occur.  The first is at DriverEntry
        // time when ParPortRegisterClaim calls parport_register_driver.
        // It goes through the list of all ports it already knows about
        // calling this attach routine.  The second is if a port arrives
        // at some later time after we've started.  In our case, this
        // should never happen since we will refuse to succeed in
        // DriverEntry without a port.

        // xch_address contains the desired port address (such as
        // 0x378) that was specified by the user.  This is the only
        // port we're interested in.  So just ignore any others that
        // are presented to us.

        if ( xch_address == (int)port->base ) {

                // Prepare entries in the newer sysfs systems at the
                // lower or cdev/device level.  DevNum contains our
                // device number.  It was constructed out of our major
                // and minor numbers (xch_major and 0 respectively).  We
                // expect the class device name we are creating here
                // with SrParXch%X and xch_address matches xch_name.
                //
                // Typically device_create is called with a SYSFSDEV
                // format string like "SrParXch%X" as the next to last
                // argument and the last argument is used to fill in the
                // variable part.  But we go ahead and pass xch_name which
                // has the full driver name "SrParXch0", so the xch_address
                // argument is effectively ignored.

                device_create( Dev->ParPortClass, NULL, Dev->DevNum, NULL, xch_name, xch_address );


                // Now register as a parport device.  This means we want
                // to be listed as one of the devices that are allowed
                // to use this port.

                if ( xch_irq == 0 ) { // Register WITHOUT interrupt

                        Dev->ParPortDevice = parport_register_device(
                                port,                   // ParPort we're registering with
                                xch_name,               // Our device name
                                Linux_ParPortPreempt,   // Routine to say NO when others want to use port
                                NULL,                   // Routine to grab port when it comes free
                                NULL,                   // NO routine to process interrupt
                                0,                      // flags usually 0, PARPORT_DEV_EXCL for nasty exclusive request
                                (void *)Dev );          // We'll pass our DEVINFO to the callback routines
                        }

                else {                 // Register WITH interrupt
                        Dev->ParPortDevice = parport_register_device(
                                port,                   // ParPort we're registering with
                                xch_name,               // Our device name
                                Linux_ParPortPreempt,   // Routine to say NO when others want to use port
                                NULL,                   // Routine to grab port when it comes free
                                Linux_ParPortIrq,       // Routine to process interrupt
                                0,                      // flags usually 0, PARPORT_DEV_EXCL for nasty exclusive request
                                (void *)Dev );          // We'll pass our DEVINFO to the callback routines
                        }

                // WCT - It looks like the register device code may
                // automatically enable interrupts if an irq callback is
                // provided.  Perhaps we should register without the irq
                // callback and add it to the parxch_parport_device
                // structure manually here instead to avoid having
                // interrupts start enabled.
                // parxch_parport_device->irq_func = Linux_ParPortIrq;

                // Try disabling interrupts as soon as possible, even
                // though we haven't claimed the port yet.

                if (Dev->ParPortDevice)
                        ParXch_InterruptDisable( Dev );

                if (xch_irq == 0) {
                        dbg1_printk("SUCCESS: Attaching port at 0x%lX 0x%lX (NO irq 0x%d)\n", port->base, port->base_hi, port->irq );
                        }
                else {
                        dbg1_printk("SUCCESS: Attaching port at 0x%lX 0x%lX (irq 0x%d)\n", port->base, port->base_hi, port->irq );
                        }
                } // end if xch_address

        else {
                dbg1_printk("Ignoring port at 0x%lX since we want the one at 0x%X\n", port->base, xch_address );
                }

        dbg0_printk( "Linux_ParPortAttachPort - Exit\n" );
}

void Linux_ParPortUnregisterRelease( DEVINFO *Dev ) {

        dbg0_printk( "Linux_ParPortUnregisterRelease - Enter\n" );


        // Unregister with the parport system.
        // This reverses anything done in Linux_ParPortRegister.



        // Release the port we claimed in ParPortRegisterClaim, since we
        // are done using it now.  This lets other devices use it.

        if ( Dev->ParPortPort ) {
                parport_release( Dev->ParPortDevice );
                Dev->ParPortPort = NULL;        // Indicate it is gone.
                }
        else {
                dbg_printk( "No release since ParPortPort **already** NULL\n" );
                }


        // Unregister our driver with the parport subsystem.  This says
        // I am no longer interested in hearing when ports come and go.

        if ( Dev->ParPortDriver ) {
                parport_unregister_driver( Dev->ParPortDriver );
                Dev->ParPortDriver = NULL;  // Indicate it is gone.
                }
        else {
                dbg_printk("ParPort NOT registered, so parport_unregister_driver not called\n");
                }


        // Linux_ParPortDetachPort is called for all exisiting ports by
        // the parport_unregister_driver function above.  If our detach
        // function sees the port we are using, we unregister our device ...


        // Remove entries in the newer sysfs systems at the upper or
        // dir/class level.

        if (Dev->ParPortClass) {
                class_destroy( Dev->ParPortClass );
                Dev->ParPortClass = NULL; // Indicate it is gone.
                }
        else {
                dbg_printk("ParPortClass is ** already** NULL\n");
                }

        dbg0_printk( "Linux_ParPortUnregisterRelease - Exit\n" );
}

void Linux_ParPortDetachPort( struct parport *port ) {

        DEVINFO *Dev;

        dbg0_printk( "Linux_ParPortDetachPort - Enter\n" );

        // Error check

        if (!port)
                return;


        dbg1_printk( "Port name = %s\n",    port->name );
        dbg1_printk( "Port num  = %d\n",    port->number );
        dbg1_printk( "Port base = 0x%lX\n", port->base );
        dbg1_printk( "Port modes = 0x%X\n", port->modes );

        Dev = &ParXchDeviceInfo;



        // Detach from a parport.
        // This reverses anything done in Linux_ParPortAtttachPort.

        // This routine is called by the parport subsystem when a
        // parport "goes away".  This should never happen to the port of
        // interest because we put a claim on that port in our
        // DriverEntry function and do not release it until DrvUnload
        // time when ParPortUnregisterRelease is called from the
        // ResourceFree function.


        // xch_address contains the desired port address (such as
        // 0x378) that was specified by the user.  This is the only
        // port we're interested in.  So just ignore any others that
        // are presented to us.

        if ( xch_address == (int)port->base ) {

                // Take us off the list of devices allowed to use it.

                if (Dev->ParPortDevice) {
                        parport_unregister_device( Dev->ParPortDevice );
                        Dev->ParPortDevice = NULL; // Indicate it is gone.
                        }
                else {
                        dbg_printk("ParPortDevice is **already** NULL\n");
                        }



                // Remove entries in the newer sysfs systems at the
                // lower or cdev/device level.  DevNum contains our
                // device number.  It was constructed out of our major
                // and minor numbers (xch_major and 0 respectively).

                if (Dev->ParPortClass) { // We should really test against the DevNum class device
                        device_destroy( Dev->ParPortClass, Dev->DevNum );
                        // Don't have a pointer to the DevNum ClassDevice to NULL
                        }
                else {
                        dbg_printk("ParPortClass is **already** NULL\n");
                        }

                // Don't set parxch_parport_class = NULL yet since we
                // just destroyed the device below the class and not
                // the class itself.

                } // end if xch_address

        else {
                dbg1_printk("Ignore port at 0x%lX since we're using the one at 0x%X\n", port->base, xch_address );
                }

        dbg0_printk( "Linux_ParPortDetachPort - Exit\n" );
}

irqreturn_t Linux_DrvInterrupt( int irq, void *dev_id ) {

        if ( ParXch_DrvInterrupt( (DEVINFO *)dev_id ) == 1 )
                return( IRQ_HANDLED );
        else
                return( IRQ_NONE );
}

int Linux_DrvOpen(  struct inode *inode, struct file *filp ) {

        filp->private_data = &ParXchDeviceInfo;

        return( ParXch_DrvOpen( filp->private_data, NULL ) );
}

int Linux_DrvClose( struct inode *inode, struct file *filp ) {

        ParXch_DrvClose( filp->private_data, NULL );

        return( 0 );
}

int Linux_DrvIoctl( struct inode *inode, struct file *filp,
                    unsigned int cmd, unsigned long arg ) {

        int      retval;
        DEVINFO *dev;

        dbg5_printk( "In Linux_DrvIoctl with cmd 0x%X, arg 0x%lX\n", cmd, arg );

        if (filp)
                dev = (DEVINFO*)filp->private_data;
        else
                return( STATUS_DEVICE_CONFIGURATION_ERROR );


        // Starting with Linux 2.6, we can NEVER access the user space IRP
        // directly.  So, we check it here and make ourselves a kernel
        // space copy we can use.

        if ( access_ok( VERIFY_WRITE, (void*)arg, sizeof( IRP ) ) ) {
                retval = copy_from_user( &(dev->DriverIrp), (IRP __user *)arg, sizeof( IRP ) );
                if (retval < 0) {
                        dbg_printk("Failed copy_from_user on Irp err = %d\n", retval );
                        return( STATUS_INVALID_PARAMETER );
                        }
                }
        else {
                dbg_printk( "Access is NOT OK for write\n" );
                return( STATUS_INVALID_USER_BUFFER );
                }

        return( ParXch_DrvIoctl( filp->private_data, (IRP*)&(dev->DriverIrp) ) );
}

ssize_t Linux_DrvRead(struct file *filp, char *buf, size_t count, loff_t *offset){

        long status;
        IRP  Irp;

        Irp.Command    = IOCTL_MAXFUNC;
        Irp.InBuffer   = NULL;
        Irp.InSize     = 0;
        Irp.OutBuffer  = buf;
        Irp.OutSize    = count;
        Irp.DataMethod = METHOD_OUT_DIRECT;
        Irp.UserIrp    = NULL;

        status = ParXch_DrvRead( filp->private_data, &Irp );

        if (status == STATUS_SUCCESS) {
                filp->f_pos += Irp.ReturnedBytes;
                return( Irp.ReturnedBytes );
                }
        else
                return( status );
}

static int __init DriverEntry( void ) {

        int Status;

        dbg0_printk( "DriverEntry (ie init_module)\n" );

        // EXPORT_NO_SYMBOLS; is now the default in 2.6

        // Set up dispatch table
        // This is done above by initializing the MajorFunction structure.



        // Get parameters from command line values or defaults

        Linux_GetParameters( );



        // Allocate and register resources

        Status = Linux_ResourceReserve( &ParXchDeviceInfo );


        return( Status );
}
module_init( DriverEntry ); // This macro relates new name to init_module

static void __exit Linux_DrvUnload( void ) {

        DEVINFO *Dev;

        dbg0_printk( "Linux_DrvUnload (ie cleanup_module)\n" );


        Dev = &ParXchDeviceInfo;


        // Free common kernel area

        ParXch_DrvUnload( Dev );



        // Clean up device and port

        Linux_ResourceFree( Dev, xch_cdev, xch_irq, xch_address, xch_major, xch_name );

}
module_exit( Linux_DrvUnload );  // This macro relates new name to cleanup_module

#endif // SROS_xxxxx
